// =============================================================================
//
// ztd.text
// Copyright © 2021 JeanHeyd "ThePhD" Meneide and Shepherd's Oasis, LLC
// Contact: opensource@soasis.org
//
// Commercial License Usage
// Licensees holding valid commercial ztd.text licenses may use this file in
// accordance with the commercial license agreement provided with the
// Software or, alternatively, in accordance with the terms contained in
// a written agreement between you and Shepherd's Oasis, LLC.
// For licensing terms and conditions see your agreement. For
// further information contact opensource@soasis.org.
//
// Apache License Version 2 Usage
// Alternatively, this file may be used under the terms of Apache License
// Version 2.0 (the "License") for non-commercial use; you may not use this
// file except in compliance with the License. You may obtain a copy of the
// License at
//
//		http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// ============================================================================>

#include <cstdint>
#include <array>
#include <utility>
#include <algorithm>
#include <optional>

namespace {

	std::array<std::pair<std::uint_least32_t, std::uint_least32_t>, 7724> index_codepoint_map { { { 0, 0x3000 },
		{ 1, 0x3001 }, { 2, 0x3002 }, { 3, 0xFF0C }, { 4, 0xFF0E }, { 5, 0x30FB }, { 6, 0xFF1A }, { 7, 0xFF1B },
		{ 8, 0xFF1F }, { 9, 0xFF01 }, { 10, 0x309B }, { 11, 0x309C }, { 12, 0x00B4 }, { 13, 0xFF40 }, { 14, 0x00A8 },
		{ 15, 0xFF3E }, { 16, 0xFFE3 }, { 17, 0xFF3F }, { 18, 0x30FD }, { 19, 0x30FE }, { 20, 0x309D },
		{ 21, 0x309E }, { 22, 0x3003 }, { 23, 0x4EDD }, { 24, 0x3005 }, { 25, 0x3006 }, { 26, 0x3007 },
		{ 27, 0x30FC }, { 28, 0x2015 }, { 29, 0x2010 }, { 30, 0xFF0F }, { 31, 0xFF3C }, { 32, 0xFF5E },
		{ 33, 0x2225 }, { 34, 0xFF5C }, { 35, 0x2026 }, { 36, 0x2025 }, { 37, 0x2018 }, { 38, 0x2019 },
		{ 39, 0x201C }, { 40, 0x201D }, { 41, 0xFF08 }, { 42, 0xFF09 }, { 43, 0x3014 }, { 44, 0x3015 },
		{ 45, 0xFF3B }, { 46, 0xFF3D }, { 47, 0xFF5B }, { 48, 0xFF5D }, { 49, 0x3008 }, { 50, 0x3009 },
		{ 51, 0x300A }, { 52, 0x300B }, { 53, 0x300C }, { 54, 0x300D }, { 55, 0x300E }, { 56, 0x300F },
		{ 57, 0x3010 }, { 58, 0x3011 }, { 59, 0xFF0B }, { 60, 0xFF0D }, { 61, 0x00B1 }, { 62, 0x00D7 },
		{ 63, 0x00F7 }, { 64, 0xFF1D }, { 65, 0x2260 }, { 66, 0xFF1C }, { 67, 0xFF1E }, { 68, 0x2266 },
		{ 69, 0x2267 }, { 70, 0x221E }, { 71, 0x2234 }, { 72, 0x2642 }, { 73, 0x2640 }, { 74, 0x00B0 },
		{ 75, 0x2032 }, { 76, 0x2033 }, { 77, 0x2103 }, { 78, 0xFFE5 }, { 79, 0xFF04 }, { 80, 0xFFE0 },
		{ 81, 0xFFE1 }, { 82, 0xFF05 }, { 83, 0xFF03 }, { 84, 0xFF06 }, { 85, 0xFF0A }, { 86, 0xFF20 },
		{ 87, 0x00A7 }, { 88, 0x2606 }, { 89, 0x2605 }, { 90, 0x25CB }, { 91, 0x25CF }, { 92, 0x25CE },
		{ 93, 0x25C7 }, { 94, 0x25C6 }, { 95, 0x25A1 }, { 96, 0x25A0 }, { 97, 0x25B3 }, { 98, 0x25B2 },
		{ 99, 0x25BD }, { 100, 0x25BC }, { 101, 0x203B }, { 102, 0x3012 }, { 103, 0x2192 }, { 104, 0x2190 },
		{ 105, 0x2191 }, { 106, 0x2193 }, { 107, 0x3013 }, { 119, 0x2208 }, { 120, 0x220B }, { 121, 0x2286 },
		{ 122, 0x2287 }, { 123, 0x2282 }, { 124, 0x2283 }, { 125, 0x222A }, { 126, 0x2229 }, { 135, 0x2227 },
		{ 136, 0x2228 }, { 137, 0xFFE2 }, { 138, 0x21D2 }, { 139, 0x21D4 }, { 140, 0x2200 }, { 141, 0x2203 },
		{ 153, 0x2220 }, { 154, 0x22A5 }, { 155, 0x2312 }, { 156, 0x2202 }, { 157, 0x2207 }, { 158, 0x2261 },
		{ 159, 0x2252 }, { 160, 0x226A }, { 161, 0x226B }, { 162, 0x221A }, { 163, 0x223D }, { 164, 0x221D },
		{ 165, 0x2235 }, { 166, 0x222B }, { 167, 0x222C }, { 175, 0x212B }, { 176, 0x2030 }, { 177, 0x266F },
		{ 178, 0x266D }, { 179, 0x266A }, { 180, 0x2020 }, { 181, 0x2021 }, { 182, 0x00B6 }, { 187, 0x25EF },
		{ 203, 0xFF10 }, { 204, 0xFF11 }, { 205, 0xFF12 }, { 206, 0xFF13 }, { 207, 0xFF14 }, { 208, 0xFF15 },
		{ 209, 0xFF16 }, { 210, 0xFF17 }, { 211, 0xFF18 }, { 212, 0xFF19 }, { 220, 0xFF21 }, { 221, 0xFF22 },
		{ 222, 0xFF23 }, { 223, 0xFF24 }, { 224, 0xFF25 }, { 225, 0xFF26 }, { 226, 0xFF27 }, { 227, 0xFF28 },
		{ 228, 0xFF29 }, { 229, 0xFF2A }, { 230, 0xFF2B }, { 231, 0xFF2C }, { 232, 0xFF2D }, { 233, 0xFF2E },
		{ 234, 0xFF2F }, { 235, 0xFF30 }, { 236, 0xFF31 }, { 237, 0xFF32 }, { 238, 0xFF33 }, { 239, 0xFF34 },
		{ 240, 0xFF35 }, { 241, 0xFF36 }, { 242, 0xFF37 }, { 243, 0xFF38 }, { 244, 0xFF39 }, { 245, 0xFF3A },
		{ 252, 0xFF41 }, { 253, 0xFF42 }, { 254, 0xFF43 }, { 255, 0xFF44 }, { 256, 0xFF45 }, { 257, 0xFF46 },
		{ 258, 0xFF47 }, { 259, 0xFF48 }, { 260, 0xFF49 }, { 261, 0xFF4A }, { 262, 0xFF4B }, { 263, 0xFF4C },
		{ 264, 0xFF4D }, { 265, 0xFF4E }, { 266, 0xFF4F }, { 267, 0xFF50 }, { 268, 0xFF51 }, { 269, 0xFF52 },
		{ 270, 0xFF53 }, { 271, 0xFF54 }, { 272, 0xFF55 }, { 273, 0xFF56 }, { 274, 0xFF57 }, { 275, 0xFF58 },
		{ 276, 0xFF59 }, { 277, 0xFF5A }, { 282, 0x3041 }, { 283, 0x3042 }, { 284, 0x3043 }, { 285, 0x3044 },
		{ 286, 0x3045 }, { 287, 0x3046 }, { 288, 0x3047 }, { 289, 0x3048 }, { 290, 0x3049 }, { 291, 0x304A },
		{ 292, 0x304B }, { 293, 0x304C }, { 294, 0x304D }, { 295, 0x304E }, { 296, 0x304F }, { 297, 0x3050 },
		{ 298, 0x3051 }, { 299, 0x3052 }, { 300, 0x3053 }, { 301, 0x3054 }, { 302, 0x3055 }, { 303, 0x3056 },
		{ 304, 0x3057 }, { 305, 0x3058 }, { 306, 0x3059 }, { 307, 0x305A }, { 308, 0x305B }, { 309, 0x305C },
		{ 310, 0x305D }, { 311, 0x305E }, { 312, 0x305F }, { 313, 0x3060 }, { 314, 0x3061 }, { 315, 0x3062 },
		{ 316, 0x3063 }, { 317, 0x3064 }, { 318, 0x3065 }, { 319, 0x3066 }, { 320, 0x3067 }, { 321, 0x3068 },
		{ 322, 0x3069 }, { 323, 0x306A }, { 324, 0x306B }, { 325, 0x306C }, { 326, 0x306D }, { 327, 0x306E },
		{ 328, 0x306F }, { 329, 0x3070 }, { 330, 0x3071 }, { 331, 0x3072 }, { 332, 0x3073 }, { 333, 0x3074 },
		{ 334, 0x3075 }, { 335, 0x3076 }, { 336, 0x3077 }, { 337, 0x3078 }, { 338, 0x3079 }, { 339, 0x307A },
		{ 340, 0x307B }, { 341, 0x307C }, { 342, 0x307D }, { 343, 0x307E }, { 344, 0x307F }, { 345, 0x3080 },
		{ 346, 0x3081 }, { 347, 0x3082 }, { 348, 0x3083 }, { 349, 0x3084 }, { 350, 0x3085 }, { 351, 0x3086 },
		{ 352, 0x3087 }, { 353, 0x3088 }, { 354, 0x3089 }, { 355, 0x308A }, { 356, 0x308B }, { 357, 0x308C },
		{ 358, 0x308D }, { 359, 0x308E }, { 360, 0x308F }, { 361, 0x3090 }, { 362, 0x3091 }, { 363, 0x3092 },
		{ 364, 0x3093 }, { 376, 0x30A1 }, { 377, 0x30A2 }, { 378, 0x30A3 }, { 379, 0x30A4 }, { 380, 0x30A5 },
		{ 381, 0x30A6 }, { 382, 0x30A7 }, { 383, 0x30A8 }, { 384, 0x30A9 }, { 385, 0x30AA }, { 386, 0x30AB },
		{ 387, 0x30AC }, { 388, 0x30AD }, { 389, 0x30AE }, { 390, 0x30AF }, { 391, 0x30B0 }, { 392, 0x30B1 },
		{ 393, 0x30B2 }, { 394, 0x30B3 }, { 395, 0x30B4 }, { 396, 0x30B5 }, { 397, 0x30B6 }, { 398, 0x30B7 },
		{ 399, 0x30B8 }, { 400, 0x30B9 }, { 401, 0x30BA }, { 402, 0x30BB }, { 403, 0x30BC }, { 404, 0x30BD },
		{ 405, 0x30BE }, { 406, 0x30BF }, { 407, 0x30C0 }, { 408, 0x30C1 }, { 409, 0x30C2 }, { 410, 0x30C3 },
		{ 411, 0x30C4 }, { 412, 0x30C5 }, { 413, 0x30C6 }, { 414, 0x30C7 }, { 415, 0x30C8 }, { 416, 0x30C9 },
		{ 417, 0x30CA }, { 418, 0x30CB }, { 419, 0x30CC }, { 420, 0x30CD }, { 421, 0x30CE }, { 422, 0x30CF },
		{ 423, 0x30D0 }, { 424, 0x30D1 }, { 425, 0x30D2 }, { 426, 0x30D3 }, { 427, 0x30D4 }, { 428, 0x30D5 },
		{ 429, 0x30D6 }, { 430, 0x30D7 }, { 431, 0x30D8 }, { 432, 0x30D9 }, { 433, 0x30DA }, { 434, 0x30DB },
		{ 435, 0x30DC }, { 436, 0x30DD }, { 437, 0x30DE }, { 438, 0x30DF }, { 439, 0x30E0 }, { 440, 0x30E1 },
		{ 441, 0x30E2 }, { 442, 0x30E3 }, { 443, 0x30E4 }, { 444, 0x30E5 }, { 445, 0x30E6 }, { 446, 0x30E7 },
		{ 447, 0x30E8 }, { 448, 0x30E9 }, { 449, 0x30EA }, { 450, 0x30EB }, { 451, 0x30EC }, { 452, 0x30ED },
		{ 453, 0x30EE }, { 454, 0x30EF }, { 455, 0x30F0 }, { 456, 0x30F1 }, { 457, 0x30F2 }, { 458, 0x30F3 },
		{ 459, 0x30F4 }, { 460, 0x30F5 }, { 461, 0x30F6 }, { 470, 0x0391 }, { 471, 0x0392 }, { 472, 0x0393 },
		{ 473, 0x0394 }, { 474, 0x0395 }, { 475, 0x0396 }, { 476, 0x0397 }, { 477, 0x0398 }, { 478, 0x0399 },
		{ 479, 0x039A }, { 480, 0x039B }, { 481, 0x039C }, { 482, 0x039D }, { 483, 0x039E }, { 484, 0x039F },
		{ 485, 0x03A0 }, { 486, 0x03A1 }, { 487, 0x03A3 }, { 488, 0x03A4 }, { 489, 0x03A5 }, { 490, 0x03A6 },
		{ 491, 0x03A7 }, { 492, 0x03A8 }, { 493, 0x03A9 }, { 502, 0x03B1 }, { 503, 0x03B2 }, { 504, 0x03B3 },
		{ 505, 0x03B4 }, { 506, 0x03B5 }, { 507, 0x03B6 }, { 508, 0x03B7 }, { 509, 0x03B8 }, { 510, 0x03B9 },
		{ 511, 0x03BA }, { 512, 0x03BB }, { 513, 0x03BC }, { 514, 0x03BD }, { 515, 0x03BE }, { 516, 0x03BF },
		{ 517, 0x03C0 }, { 518, 0x03C1 }, { 519, 0x03C3 }, { 520, 0x03C4 }, { 521, 0x03C5 }, { 522, 0x03C6 },
		{ 523, 0x03C7 }, { 524, 0x03C8 }, { 525, 0x03C9 }, { 564, 0x0410 }, { 565, 0x0411 }, { 566, 0x0412 },
		{ 567, 0x0413 }, { 568, 0x0414 }, { 569, 0x0415 }, { 570, 0x0401 }, { 571, 0x0416 }, { 572, 0x0417 },
		{ 573, 0x0418 }, { 574, 0x0419 }, { 575, 0x041A }, { 576, 0x041B }, { 577, 0x041C }, { 578, 0x041D },
		{ 579, 0x041E }, { 580, 0x041F }, { 581, 0x0420 }, { 582, 0x0421 }, { 583, 0x0422 }, { 584, 0x0423 },
		{ 585, 0x0424 }, { 586, 0x0425 }, { 587, 0x0426 }, { 588, 0x0427 }, { 589, 0x0428 }, { 590, 0x0429 },
		{ 591, 0x042A }, { 592, 0x042B }, { 593, 0x042C }, { 594, 0x042D }, { 595, 0x042E }, { 596, 0x042F },
		{ 612, 0x0430 }, { 613, 0x0431 }, { 614, 0x0432 }, { 615, 0x0433 }, { 616, 0x0434 }, { 617, 0x0435 },
		{ 618, 0x0451 }, { 619, 0x0436 }, { 620, 0x0437 }, { 621, 0x0438 }, { 622, 0x0439 }, { 623, 0x043A },
		{ 624, 0x043B }, { 625, 0x043C }, { 626, 0x043D }, { 627, 0x043E }, { 628, 0x043F }, { 629, 0x0440 },
		{ 630, 0x0441 }, { 631, 0x0442 }, { 632, 0x0443 }, { 633, 0x0444 }, { 634, 0x0445 }, { 635, 0x0446 },
		{ 636, 0x0447 }, { 637, 0x0448 }, { 638, 0x0449 }, { 639, 0x044A }, { 640, 0x044B }, { 641, 0x044C },
		{ 642, 0x044D }, { 643, 0x044E }, { 644, 0x044F }, { 658, 0x2500 }, { 659, 0x2502 }, { 660, 0x250C },
		{ 661, 0x2510 }, { 662, 0x2518 }, { 663, 0x2514 }, { 664, 0x251C }, { 665, 0x252C }, { 666, 0x2524 },
		{ 667, 0x2534 }, { 668, 0x253C }, { 669, 0x2501 }, { 670, 0x2503 }, { 671, 0x250F }, { 672, 0x2513 },
		{ 673, 0x251B }, { 674, 0x2517 }, { 675, 0x2523 }, { 676, 0x2533 }, { 677, 0x252B }, { 678, 0x253B },
		{ 679, 0x254B }, { 680, 0x2520 }, { 681, 0x252F }, { 682, 0x2528 }, { 683, 0x2537 }, { 684, 0x253F },
		{ 685, 0x251D }, { 686, 0x2530 }, { 687, 0x2525 }, { 688, 0x2538 }, { 689, 0x2542 }, { 1128, 0x2460 },
		{ 1129, 0x2461 }, { 1130, 0x2462 }, { 1131, 0x2463 }, { 1132, 0x2464 }, { 1133, 0x2465 }, { 1134, 0x2466 },
		{ 1135, 0x2467 }, { 1136, 0x2468 }, { 1137, 0x2469 }, { 1138, 0x246A }, { 1139, 0x246B }, { 1140, 0x246C },
		{ 1141, 0x246D }, { 1142, 0x246E }, { 1143, 0x246F }, { 1144, 0x2470 }, { 1145, 0x2471 }, { 1146, 0x2472 },
		{ 1147, 0x2473 }, { 1148, 0x2160 }, { 1149, 0x2161 }, { 1150, 0x2162 }, { 1151, 0x2163 }, { 1152, 0x2164 },
		{ 1153, 0x2165 }, { 1154, 0x2166 }, { 1155, 0x2167 }, { 1156, 0x2168 }, { 1157, 0x2169 }, { 1159, 0x3349 },
		{ 1160, 0x3314 }, { 1161, 0x3322 }, { 1162, 0x334D }, { 1163, 0x3318 }, { 1164, 0x3327 }, { 1165, 0x3303 },
		{ 1166, 0x3336 }, { 1167, 0x3351 }, { 1168, 0x3357 }, { 1169, 0x330D }, { 1170, 0x3326 }, { 1171, 0x3323 },
		{ 1172, 0x332B }, { 1173, 0x334A }, { 1174, 0x333B }, { 1175, 0x339C }, { 1176, 0x339D }, { 1177, 0x339E },
		{ 1178, 0x338E }, { 1179, 0x338F }, { 1180, 0x33C4 }, { 1181, 0x33A1 }, { 1190, 0x337B }, { 1191, 0x301D },
		{ 1192, 0x301F }, { 1193, 0x2116 }, { 1194, 0x33CD }, { 1195, 0x2121 }, { 1196, 0x32A4 }, { 1197, 0x32A5 },
		{ 1198, 0x32A6 }, { 1199, 0x32A7 }, { 1200, 0x32A8 }, { 1201, 0x3231 }, { 1202, 0x3232 }, { 1203, 0x3239 },
		{ 1204, 0x337E }, { 1205, 0x337D }, { 1206, 0x337C }, { 1207, 0x2252 }, { 1208, 0x2261 }, { 1209, 0x222B },
		{ 1210, 0x222E }, { 1211, 0x2211 }, { 1212, 0x221A }, { 1213, 0x22A5 }, { 1214, 0x2220 }, { 1215, 0x221F },
		{ 1216, 0x22BF }, { 1217, 0x2235 }, { 1218, 0x2229 }, { 1219, 0x222A }, { 1410, 0x4E9C }, { 1411, 0x5516 },
		{ 1412, 0x5A03 }, { 1413, 0x963F }, { 1414, 0x54C0 }, { 1415, 0x611B }, { 1416, 0x6328 }, { 1417, 0x59F6 },
		{ 1418, 0x9022 }, { 1419, 0x8475 }, { 1420, 0x831C }, { 1421, 0x7A50 }, { 1422, 0x60AA }, { 1423, 0x63E1 },
		{ 1424, 0x6E25 }, { 1425, 0x65ED }, { 1426, 0x8466 }, { 1427, 0x82A6 }, { 1428, 0x9BF5 }, { 1429, 0x6893 },
		{ 1430, 0x5727 }, { 1431, 0x65A1 }, { 1432, 0x6271 }, { 1433, 0x5B9B }, { 1434, 0x59D0 }, { 1435, 0x867B },
		{ 1436, 0x98F4 }, { 1437, 0x7D62 }, { 1438, 0x7DBE }, { 1439, 0x9B8E }, { 1440, 0x6216 }, { 1441, 0x7C9F },
		{ 1442, 0x88B7 }, { 1443, 0x5B89 }, { 1444, 0x5EB5 }, { 1445, 0x6309 }, { 1446, 0x6697 }, { 1447, 0x6848 },
		{ 1448, 0x95C7 }, { 1449, 0x978D }, { 1450, 0x674F }, { 1451, 0x4EE5 }, { 1452, 0x4F0A }, { 1453, 0x4F4D },
		{ 1454, 0x4F9D }, { 1455, 0x5049 }, { 1456, 0x56F2 }, { 1457, 0x5937 }, { 1458, 0x59D4 }, { 1459, 0x5A01 },
		{ 1460, 0x5C09 }, { 1461, 0x60DF }, { 1462, 0x610F }, { 1463, 0x6170 }, { 1464, 0x6613 }, { 1465, 0x6905 },
		{ 1466, 0x70BA }, { 1467, 0x754F }, { 1468, 0x7570 }, { 1469, 0x79FB }, { 1470, 0x7DAD }, { 1471, 0x7DEF },
		{ 1472, 0x80C3 }, { 1473, 0x840E }, { 1474, 0x8863 }, { 1475, 0x8B02 }, { 1476, 0x9055 }, { 1477, 0x907A },
		{ 1478, 0x533B }, { 1479, 0x4E95 }, { 1480, 0x4EA5 }, { 1481, 0x57DF }, { 1482, 0x80B2 }, { 1483, 0x90C1 },
		{ 1484, 0x78EF }, { 1485, 0x4E00 }, { 1486, 0x58F1 }, { 1487, 0x6EA2 }, { 1488, 0x9038 }, { 1489, 0x7A32 },
		{ 1490, 0x8328 }, { 1491, 0x828B }, { 1492, 0x9C2F }, { 1493, 0x5141 }, { 1494, 0x5370 }, { 1495, 0x54BD },
		{ 1496, 0x54E1 }, { 1497, 0x56E0 }, { 1498, 0x59FB }, { 1499, 0x5F15 }, { 1500, 0x98F2 }, { 1501, 0x6DEB },
		{ 1502, 0x80E4 }, { 1503, 0x852D }, { 1504, 0x9662 }, { 1505, 0x9670 }, { 1506, 0x96A0 }, { 1507, 0x97FB },
		{ 1508, 0x540B }, { 1509, 0x53F3 }, { 1510, 0x5B87 }, { 1511, 0x70CF }, { 1512, 0x7FBD }, { 1513, 0x8FC2 },
		{ 1514, 0x96E8 }, { 1515, 0x536F }, { 1516, 0x9D5C }, { 1517, 0x7ABA }, { 1518, 0x4E11 }, { 1519, 0x7893 },
		{ 1520, 0x81FC }, { 1521, 0x6E26 }, { 1522, 0x5618 }, { 1523, 0x5504 }, { 1524, 0x6B1D }, { 1525, 0x851A },
		{ 1526, 0x9C3B }, { 1527, 0x59E5 }, { 1528, 0x53A9 }, { 1529, 0x6D66 }, { 1530, 0x74DC }, { 1531, 0x958F },
		{ 1532, 0x5642 }, { 1533, 0x4E91 }, { 1534, 0x904B }, { 1535, 0x96F2 }, { 1536, 0x834F }, { 1537, 0x990C },
		{ 1538, 0x53E1 }, { 1539, 0x55B6 }, { 1540, 0x5B30 }, { 1541, 0x5F71 }, { 1542, 0x6620 }, { 1543, 0x66F3 },
		{ 1544, 0x6804 }, { 1545, 0x6C38 }, { 1546, 0x6CF3 }, { 1547, 0x6D29 }, { 1548, 0x745B }, { 1549, 0x76C8 },
		{ 1550, 0x7A4E }, { 1551, 0x9834 }, { 1552, 0x82F1 }, { 1553, 0x885B }, { 1554, 0x8A60 }, { 1555, 0x92ED },
		{ 1556, 0x6DB2 }, { 1557, 0x75AB }, { 1558, 0x76CA }, { 1559, 0x99C5 }, { 1560, 0x60A6 }, { 1561, 0x8B01 },
		{ 1562, 0x8D8A }, { 1563, 0x95B2 }, { 1564, 0x698E }, { 1565, 0x53AD }, { 1566, 0x5186 }, { 1567, 0x5712 },
		{ 1568, 0x5830 }, { 1569, 0x5944 }, { 1570, 0x5BB4 }, { 1571, 0x5EF6 }, { 1572, 0x6028 }, { 1573, 0x63A9 },
		{ 1574, 0x63F4 }, { 1575, 0x6CBF }, { 1576, 0x6F14 }, { 1577, 0x708E }, { 1578, 0x7114 }, { 1579, 0x7159 },
		{ 1580, 0x71D5 }, { 1581, 0x733F }, { 1582, 0x7E01 }, { 1583, 0x8276 }, { 1584, 0x82D1 }, { 1585, 0x8597 },
		{ 1586, 0x9060 }, { 1587, 0x925B }, { 1588, 0x9D1B }, { 1589, 0x5869 }, { 1590, 0x65BC }, { 1591, 0x6C5A },
		{ 1592, 0x7525 }, { 1593, 0x51F9 }, { 1594, 0x592E }, { 1595, 0x5965 }, { 1596, 0x5F80 }, { 1597, 0x5FDC },
		{ 1598, 0x62BC }, { 1599, 0x65FA }, { 1600, 0x6A2A }, { 1601, 0x6B27 }, { 1602, 0x6BB4 }, { 1603, 0x738B },
		{ 1604, 0x7FC1 }, { 1605, 0x8956 }, { 1606, 0x9D2C }, { 1607, 0x9D0E }, { 1608, 0x9EC4 }, { 1609, 0x5CA1 },
		{ 1610, 0x6C96 }, { 1611, 0x837B }, { 1612, 0x5104 }, { 1613, 0x5C4B }, { 1614, 0x61B6 }, { 1615, 0x81C6 },
		{ 1616, 0x6876 }, { 1617, 0x7261 }, { 1618, 0x4E59 }, { 1619, 0x4FFA }, { 1620, 0x5378 }, { 1621, 0x6069 },
		{ 1622, 0x6E29 }, { 1623, 0x7A4F }, { 1624, 0x97F3 }, { 1625, 0x4E0B }, { 1626, 0x5316 }, { 1627, 0x4EEE },
		{ 1628, 0x4F55 }, { 1629, 0x4F3D }, { 1630, 0x4FA1 }, { 1631, 0x4F73 }, { 1632, 0x52A0 }, { 1633, 0x53EF },
		{ 1634, 0x5609 }, { 1635, 0x590F }, { 1636, 0x5AC1 }, { 1637, 0x5BB6 }, { 1638, 0x5BE1 }, { 1639, 0x79D1 },
		{ 1640, 0x6687 }, { 1641, 0x679C }, { 1642, 0x67B6 }, { 1643, 0x6B4C }, { 1644, 0x6CB3 }, { 1645, 0x706B },
		{ 1646, 0x73C2 }, { 1647, 0x798D }, { 1648, 0x79BE }, { 1649, 0x7A3C }, { 1650, 0x7B87 }, { 1651, 0x82B1 },
		{ 1652, 0x82DB }, { 1653, 0x8304 }, { 1654, 0x8377 }, { 1655, 0x83EF }, { 1656, 0x83D3 }, { 1657, 0x8766 },
		{ 1658, 0x8AB2 }, { 1659, 0x5629 }, { 1660, 0x8CA8 }, { 1661, 0x8FE6 }, { 1662, 0x904E }, { 1663, 0x971E },
		{ 1664, 0x868A }, { 1665, 0x4FC4 }, { 1666, 0x5CE8 }, { 1667, 0x6211 }, { 1668, 0x7259 }, { 1669, 0x753B },
		{ 1670, 0x81E5 }, { 1671, 0x82BD }, { 1672, 0x86FE }, { 1673, 0x8CC0 }, { 1674, 0x96C5 }, { 1675, 0x9913 },
		{ 1676, 0x99D5 }, { 1677, 0x4ECB }, { 1678, 0x4F1A }, { 1679, 0x89E3 }, { 1680, 0x56DE }, { 1681, 0x584A },
		{ 1682, 0x58CA }, { 1683, 0x5EFB }, { 1684, 0x5FEB }, { 1685, 0x602A }, { 1686, 0x6094 }, { 1687, 0x6062 },
		{ 1688, 0x61D0 }, { 1689, 0x6212 }, { 1690, 0x62D0 }, { 1691, 0x6539 }, { 1692, 0x9B41 }, { 1693, 0x6666 },
		{ 1694, 0x68B0 }, { 1695, 0x6D77 }, { 1696, 0x7070 }, { 1697, 0x754C }, { 1698, 0x7686 }, { 1699, 0x7D75 },
		{ 1700, 0x82A5 }, { 1701, 0x87F9 }, { 1702, 0x958B }, { 1703, 0x968E }, { 1704, 0x8C9D }, { 1705, 0x51F1 },
		{ 1706, 0x52BE }, { 1707, 0x5916 }, { 1708, 0x54B3 }, { 1709, 0x5BB3 }, { 1710, 0x5D16 }, { 1711, 0x6168 },
		{ 1712, 0x6982 }, { 1713, 0x6DAF }, { 1714, 0x788D }, { 1715, 0x84CB }, { 1716, 0x8857 }, { 1717, 0x8A72 },
		{ 1718, 0x93A7 }, { 1719, 0x9AB8 }, { 1720, 0x6D6C }, { 1721, 0x99A8 }, { 1722, 0x86D9 }, { 1723, 0x57A3 },
		{ 1724, 0x67FF }, { 1725, 0x86CE }, { 1726, 0x920E }, { 1727, 0x5283 }, { 1728, 0x5687 }, { 1729, 0x5404 },
		{ 1730, 0x5ED3 }, { 1731, 0x62E1 }, { 1732, 0x64B9 }, { 1733, 0x683C }, { 1734, 0x6838 }, { 1735, 0x6BBB },
		{ 1736, 0x7372 }, { 1737, 0x78BA }, { 1738, 0x7A6B }, { 1739, 0x899A }, { 1740, 0x89D2 }, { 1741, 0x8D6B },
		{ 1742, 0x8F03 }, { 1743, 0x90ED }, { 1744, 0x95A3 }, { 1745, 0x9694 }, { 1746, 0x9769 }, { 1747, 0x5B66 },
		{ 1748, 0x5CB3 }, { 1749, 0x697D }, { 1750, 0x984D }, { 1751, 0x984E }, { 1752, 0x639B }, { 1753, 0x7B20 },
		{ 1754, 0x6A2B }, { 1755, 0x6A7F }, { 1756, 0x68B6 }, { 1757, 0x9C0D }, { 1758, 0x6F5F }, { 1759, 0x5272 },
		{ 1760, 0x559D }, { 1761, 0x6070 }, { 1762, 0x62EC }, { 1763, 0x6D3B }, { 1764, 0x6E07 }, { 1765, 0x6ED1 },
		{ 1766, 0x845B }, { 1767, 0x8910 }, { 1768, 0x8F44 }, { 1769, 0x4E14 }, { 1770, 0x9C39 }, { 1771, 0x53F6 },
		{ 1772, 0x691B }, { 1773, 0x6A3A }, { 1774, 0x9784 }, { 1775, 0x682A }, { 1776, 0x515C }, { 1777, 0x7AC3 },
		{ 1778, 0x84B2 }, { 1779, 0x91DC }, { 1780, 0x938C }, { 1781, 0x565B }, { 1782, 0x9D28 }, { 1783, 0x6822 },
		{ 1784, 0x8305 }, { 1785, 0x8431 }, { 1786, 0x7CA5 }, { 1787, 0x5208 }, { 1788, 0x82C5 }, { 1789, 0x74E6 },
		{ 1790, 0x4E7E }, { 1791, 0x4F83 }, { 1792, 0x51A0 }, { 1793, 0x5BD2 }, { 1794, 0x520A }, { 1795, 0x52D8 },
		{ 1796, 0x52E7 }, { 1797, 0x5DFB }, { 1798, 0x559A }, { 1799, 0x582A }, { 1800, 0x59E6 }, { 1801, 0x5B8C },
		{ 1802, 0x5B98 }, { 1803, 0x5BDB }, { 1804, 0x5E72 }, { 1805, 0x5E79 }, { 1806, 0x60A3 }, { 1807, 0x611F },
		{ 1808, 0x6163 }, { 1809, 0x61BE }, { 1810, 0x63DB }, { 1811, 0x6562 }, { 1812, 0x67D1 }, { 1813, 0x6853 },
		{ 1814, 0x68FA }, { 1815, 0x6B3E }, { 1816, 0x6B53 }, { 1817, 0x6C57 }, { 1818, 0x6F22 }, { 1819, 0x6F97 },
		{ 1820, 0x6F45 }, { 1821, 0x74B0 }, { 1822, 0x7518 }, { 1823, 0x76E3 }, { 1824, 0x770B }, { 1825, 0x7AFF },
		{ 1826, 0x7BA1 }, { 1827, 0x7C21 }, { 1828, 0x7DE9 }, { 1829, 0x7F36 }, { 1830, 0x7FF0 }, { 1831, 0x809D },
		{ 1832, 0x8266 }, { 1833, 0x839E }, { 1834, 0x89B3 }, { 1835, 0x8ACC }, { 1836, 0x8CAB }, { 1837, 0x9084 },
		{ 1838, 0x9451 }, { 1839, 0x9593 }, { 1840, 0x9591 }, { 1841, 0x95A2 }, { 1842, 0x9665 }, { 1843, 0x97D3 },
		{ 1844, 0x9928 }, { 1845, 0x8218 }, { 1846, 0x4E38 }, { 1847, 0x542B }, { 1848, 0x5CB8 }, { 1849, 0x5DCC },
		{ 1850, 0x73A9 }, { 1851, 0x764C }, { 1852, 0x773C }, { 1853, 0x5CA9 }, { 1854, 0x7FEB }, { 1855, 0x8D0B },
		{ 1856, 0x96C1 }, { 1857, 0x9811 }, { 1858, 0x9854 }, { 1859, 0x9858 }, { 1860, 0x4F01 }, { 1861, 0x4F0E },
		{ 1862, 0x5371 }, { 1863, 0x559C }, { 1864, 0x5668 }, { 1865, 0x57FA }, { 1866, 0x5947 }, { 1867, 0x5B09 },
		{ 1868, 0x5BC4 }, { 1869, 0x5C90 }, { 1870, 0x5E0C }, { 1871, 0x5E7E }, { 1872, 0x5FCC }, { 1873, 0x63EE },
		{ 1874, 0x673A }, { 1875, 0x65D7 }, { 1876, 0x65E2 }, { 1877, 0x671F }, { 1878, 0x68CB }, { 1879, 0x68C4 },
		{ 1880, 0x6A5F }, { 1881, 0x5E30 }, { 1882, 0x6BC5 }, { 1883, 0x6C17 }, { 1884, 0x6C7D }, { 1885, 0x757F },
		{ 1886, 0x7948 }, { 1887, 0x5B63 }, { 1888, 0x7A00 }, { 1889, 0x7D00 }, { 1890, 0x5FBD }, { 1891, 0x898F },
		{ 1892, 0x8A18 }, { 1893, 0x8CB4 }, { 1894, 0x8D77 }, { 1895, 0x8ECC }, { 1896, 0x8F1D }, { 1897, 0x98E2 },
		{ 1898, 0x9A0E }, { 1899, 0x9B3C }, { 1900, 0x4E80 }, { 1901, 0x507D }, { 1902, 0x5100 }, { 1903, 0x5993 },
		{ 1904, 0x5B9C }, { 1905, 0x622F }, { 1906, 0x6280 }, { 1907, 0x64EC }, { 1908, 0x6B3A }, { 1909, 0x72A0 },
		{ 1910, 0x7591 }, { 1911, 0x7947 }, { 1912, 0x7FA9 }, { 1913, 0x87FB }, { 1914, 0x8ABC }, { 1915, 0x8B70 },
		{ 1916, 0x63AC }, { 1917, 0x83CA }, { 1918, 0x97A0 }, { 1919, 0x5409 }, { 1920, 0x5403 }, { 1921, 0x55AB },
		{ 1922, 0x6854 }, { 1923, 0x6A58 }, { 1924, 0x8A70 }, { 1925, 0x7827 }, { 1926, 0x6775 }, { 1927, 0x9ECD },
		{ 1928, 0x5374 }, { 1929, 0x5BA2 }, { 1930, 0x811A }, { 1931, 0x8650 }, { 1932, 0x9006 }, { 1933, 0x4E18 },
		{ 1934, 0x4E45 }, { 1935, 0x4EC7 }, { 1936, 0x4F11 }, { 1937, 0x53CA }, { 1938, 0x5438 }, { 1939, 0x5BAE },
		{ 1940, 0x5F13 }, { 1941, 0x6025 }, { 1942, 0x6551 }, { 1943, 0x673D }, { 1944, 0x6C42 }, { 1945, 0x6C72 },
		{ 1946, 0x6CE3 }, { 1947, 0x7078 }, { 1948, 0x7403 }, { 1949, 0x7A76 }, { 1950, 0x7AAE }, { 1951, 0x7B08 },
		{ 1952, 0x7D1A }, { 1953, 0x7CFE }, { 1954, 0x7D66 }, { 1955, 0x65E7 }, { 1956, 0x725B }, { 1957, 0x53BB },
		{ 1958, 0x5C45 }, { 1959, 0x5DE8 }, { 1960, 0x62D2 }, { 1961, 0x62E0 }, { 1962, 0x6319 }, { 1963, 0x6E20 },
		{ 1964, 0x865A }, { 1965, 0x8A31 }, { 1966, 0x8DDD }, { 1967, 0x92F8 }, { 1968, 0x6F01 }, { 1969, 0x79A6 },
		{ 1970, 0x9B5A }, { 1971, 0x4EA8 }, { 1972, 0x4EAB }, { 1973, 0x4EAC }, { 1974, 0x4F9B }, { 1975, 0x4FA0 },
		{ 1976, 0x50D1 }, { 1977, 0x5147 }, { 1978, 0x7AF6 }, { 1979, 0x5171 }, { 1980, 0x51F6 }, { 1981, 0x5354 },
		{ 1982, 0x5321 }, { 1983, 0x537F }, { 1984, 0x53EB }, { 1985, 0x55AC }, { 1986, 0x5883 }, { 1987, 0x5CE1 },
		{ 1988, 0x5F37 }, { 1989, 0x5F4A }, { 1990, 0x602F }, { 1991, 0x6050 }, { 1992, 0x606D }, { 1993, 0x631F },
		{ 1994, 0x6559 }, { 1995, 0x6A4B }, { 1996, 0x6CC1 }, { 1997, 0x72C2 }, { 1998, 0x72ED }, { 1999, 0x77EF },
		{ 2000, 0x80F8 }, { 2001, 0x8105 }, { 2002, 0x8208 }, { 2003, 0x854E }, { 2004, 0x90F7 }, { 2005, 0x93E1 },
		{ 2006, 0x97FF }, { 2007, 0x9957 }, { 2008, 0x9A5A }, { 2009, 0x4EF0 }, { 2010, 0x51DD }, { 2011, 0x5C2D },
		{ 2012, 0x6681 }, { 2013, 0x696D }, { 2014, 0x5C40 }, { 2015, 0x66F2 }, { 2016, 0x6975 }, { 2017, 0x7389 },
		{ 2018, 0x6850 }, { 2019, 0x7C81 }, { 2020, 0x50C5 }, { 2021, 0x52E4 }, { 2022, 0x5747 }, { 2023, 0x5DFE },
		{ 2024, 0x9326 }, { 2025, 0x65A4 }, { 2026, 0x6B23 }, { 2027, 0x6B3D }, { 2028, 0x7434 }, { 2029, 0x7981 },
		{ 2030, 0x79BD }, { 2031, 0x7B4B }, { 2032, 0x7DCA }, { 2033, 0x82B9 }, { 2034, 0x83CC }, { 2035, 0x887F },
		{ 2036, 0x895F }, { 2037, 0x8B39 }, { 2038, 0x8FD1 }, { 2039, 0x91D1 }, { 2040, 0x541F }, { 2041, 0x9280 },
		{ 2042, 0x4E5D }, { 2043, 0x5036 }, { 2044, 0x53E5 }, { 2045, 0x533A }, { 2046, 0x72D7 }, { 2047, 0x7396 },
		{ 2048, 0x77E9 }, { 2049, 0x82E6 }, { 2050, 0x8EAF }, { 2051, 0x99C6 }, { 2052, 0x99C8 }, { 2053, 0x99D2 },
		{ 2054, 0x5177 }, { 2055, 0x611A }, { 2056, 0x865E }, { 2057, 0x55B0 }, { 2058, 0x7A7A }, { 2059, 0x5076 },
		{ 2060, 0x5BD3 }, { 2061, 0x9047 }, { 2062, 0x9685 }, { 2063, 0x4E32 }, { 2064, 0x6ADB }, { 2065, 0x91E7 },
		{ 2066, 0x5C51 }, { 2067, 0x5C48 }, { 2068, 0x6398 }, { 2069, 0x7A9F }, { 2070, 0x6C93 }, { 2071, 0x9774 },
		{ 2072, 0x8F61 }, { 2073, 0x7AAA }, { 2074, 0x718A }, { 2075, 0x9688 }, { 2076, 0x7C82 }, { 2077, 0x6817 },
		{ 2078, 0x7E70 }, { 2079, 0x6851 }, { 2080, 0x936C }, { 2081, 0x52F2 }, { 2082, 0x541B }, { 2083, 0x85AB },
		{ 2084, 0x8A13 }, { 2085, 0x7FA4 }, { 2086, 0x8ECD }, { 2087, 0x90E1 }, { 2088, 0x5366 }, { 2089, 0x8888 },
		{ 2090, 0x7941 }, { 2091, 0x4FC2 }, { 2092, 0x50BE }, { 2093, 0x5211 }, { 2094, 0x5144 }, { 2095, 0x5553 },
		{ 2096, 0x572D }, { 2097, 0x73EA }, { 2098, 0x578B }, { 2099, 0x5951 }, { 2100, 0x5F62 }, { 2101, 0x5F84 },
		{ 2102, 0x6075 }, { 2103, 0x6176 }, { 2104, 0x6167 }, { 2105, 0x61A9 }, { 2106, 0x63B2 }, { 2107, 0x643A },
		{ 2108, 0x656C }, { 2109, 0x666F }, { 2110, 0x6842 }, { 2111, 0x6E13 }, { 2112, 0x7566 }, { 2113, 0x7A3D },
		{ 2114, 0x7CFB }, { 2115, 0x7D4C }, { 2116, 0x7D99 }, { 2117, 0x7E4B }, { 2118, 0x7F6B }, { 2119, 0x830E },
		{ 2120, 0x834A }, { 2121, 0x86CD }, { 2122, 0x8A08 }, { 2123, 0x8A63 }, { 2124, 0x8B66 }, { 2125, 0x8EFD },
		{ 2126, 0x981A }, { 2127, 0x9D8F }, { 2128, 0x82B8 }, { 2129, 0x8FCE }, { 2130, 0x9BE8 }, { 2131, 0x5287 },
		{ 2132, 0x621F }, { 2133, 0x6483 }, { 2134, 0x6FC0 }, { 2135, 0x9699 }, { 2136, 0x6841 }, { 2137, 0x5091 },
		{ 2138, 0x6B20 }, { 2139, 0x6C7A }, { 2140, 0x6F54 }, { 2141, 0x7A74 }, { 2142, 0x7D50 }, { 2143, 0x8840 },
		{ 2144, 0x8A23 }, { 2145, 0x6708 }, { 2146, 0x4EF6 }, { 2147, 0x5039 }, { 2148, 0x5026 }, { 2149, 0x5065 },
		{ 2150, 0x517C }, { 2151, 0x5238 }, { 2152, 0x5263 }, { 2153, 0x55A7 }, { 2154, 0x570F }, { 2155, 0x5805 },
		{ 2156, 0x5ACC }, { 2157, 0x5EFA }, { 2158, 0x61B2 }, { 2159, 0x61F8 }, { 2160, 0x62F3 }, { 2161, 0x6372 },
		{ 2162, 0x691C }, { 2163, 0x6A29 }, { 2164, 0x727D }, { 2165, 0x72AC }, { 2166, 0x732E }, { 2167, 0x7814 },
		{ 2168, 0x786F }, { 2169, 0x7D79 }, { 2170, 0x770C }, { 2171, 0x80A9 }, { 2172, 0x898B }, { 2173, 0x8B19 },
		{ 2174, 0x8CE2 }, { 2175, 0x8ED2 }, { 2176, 0x9063 }, { 2177, 0x9375 }, { 2178, 0x967A }, { 2179, 0x9855 },
		{ 2180, 0x9A13 }, { 2181, 0x9E78 }, { 2182, 0x5143 }, { 2183, 0x539F }, { 2184, 0x53B3 }, { 2185, 0x5E7B },
		{ 2186, 0x5F26 }, { 2187, 0x6E1B }, { 2188, 0x6E90 }, { 2189, 0x7384 }, { 2190, 0x73FE }, { 2191, 0x7D43 },
		{ 2192, 0x8237 }, { 2193, 0x8A00 }, { 2194, 0x8AFA }, { 2195, 0x9650 }, { 2196, 0x4E4E }, { 2197, 0x500B },
		{ 2198, 0x53E4 }, { 2199, 0x547C }, { 2200, 0x56FA }, { 2201, 0x59D1 }, { 2202, 0x5B64 }, { 2203, 0x5DF1 },
		{ 2204, 0x5EAB }, { 2205, 0x5F27 }, { 2206, 0x6238 }, { 2207, 0x6545 }, { 2208, 0x67AF }, { 2209, 0x6E56 },
		{ 2210, 0x72D0 }, { 2211, 0x7CCA }, { 2212, 0x88B4 }, { 2213, 0x80A1 }, { 2214, 0x80E1 }, { 2215, 0x83F0 },
		{ 2216, 0x864E }, { 2217, 0x8A87 }, { 2218, 0x8DE8 }, { 2219, 0x9237 }, { 2220, 0x96C7 }, { 2221, 0x9867 },
		{ 2222, 0x9F13 }, { 2223, 0x4E94 }, { 2224, 0x4E92 }, { 2225, 0x4F0D }, { 2226, 0x5348 }, { 2227, 0x5449 },
		{ 2228, 0x543E }, { 2229, 0x5A2F }, { 2230, 0x5F8C }, { 2231, 0x5FA1 }, { 2232, 0x609F }, { 2233, 0x68A7 },
		{ 2234, 0x6A8E }, { 2235, 0x745A }, { 2236, 0x7881 }, { 2237, 0x8A9E }, { 2238, 0x8AA4 }, { 2239, 0x8B77 },
		{ 2240, 0x9190 }, { 2241, 0x4E5E }, { 2242, 0x9BC9 }, { 2243, 0x4EA4 }, { 2244, 0x4F7C }, { 2245, 0x4FAF },
		{ 2246, 0x5019 }, { 2247, 0x5016 }, { 2248, 0x5149 }, { 2249, 0x516C }, { 2250, 0x529F }, { 2251, 0x52B9 },
		{ 2252, 0x52FE }, { 2253, 0x539A }, { 2254, 0x53E3 }, { 2255, 0x5411 }, { 2256, 0x540E }, { 2257, 0x5589 },
		{ 2258, 0x5751 }, { 2259, 0x57A2 }, { 2260, 0x597D }, { 2261, 0x5B54 }, { 2262, 0x5B5D }, { 2263, 0x5B8F },
		{ 2264, 0x5DE5 }, { 2265, 0x5DE7 }, { 2266, 0x5DF7 }, { 2267, 0x5E78 }, { 2268, 0x5E83 }, { 2269, 0x5E9A },
		{ 2270, 0x5EB7 }, { 2271, 0x5F18 }, { 2272, 0x6052 }, { 2273, 0x614C }, { 2274, 0x6297 }, { 2275, 0x62D8 },
		{ 2276, 0x63A7 }, { 2277, 0x653B }, { 2278, 0x6602 }, { 2279, 0x6643 }, { 2280, 0x66F4 }, { 2281, 0x676D },
		{ 2282, 0x6821 }, { 2283, 0x6897 }, { 2284, 0x69CB }, { 2285, 0x6C5F }, { 2286, 0x6D2A }, { 2287, 0x6D69 },
		{ 2288, 0x6E2F }, { 2289, 0x6E9D }, { 2290, 0x7532 }, { 2291, 0x7687 }, { 2292, 0x786C }, { 2293, 0x7A3F },
		{ 2294, 0x7CE0 }, { 2295, 0x7D05 }, { 2296, 0x7D18 }, { 2297, 0x7D5E }, { 2298, 0x7DB1 }, { 2299, 0x8015 },
		{ 2300, 0x8003 }, { 2301, 0x80AF }, { 2302, 0x80B1 }, { 2303, 0x8154 }, { 2304, 0x818F }, { 2305, 0x822A },
		{ 2306, 0x8352 }, { 2307, 0x884C }, { 2308, 0x8861 }, { 2309, 0x8B1B }, { 2310, 0x8CA2 }, { 2311, 0x8CFC },
		{ 2312, 0x90CA }, { 2313, 0x9175 }, { 2314, 0x9271 }, { 2315, 0x783F }, { 2316, 0x92FC }, { 2317, 0x95A4 },
		{ 2318, 0x964D }, { 2319, 0x9805 }, { 2320, 0x9999 }, { 2321, 0x9AD8 }, { 2322, 0x9D3B }, { 2323, 0x525B },
		{ 2324, 0x52AB }, { 2325, 0x53F7 }, { 2326, 0x5408 }, { 2327, 0x58D5 }, { 2328, 0x62F7 }, { 2329, 0x6FE0 },
		{ 2330, 0x8C6A }, { 2331, 0x8F5F }, { 2332, 0x9EB9 }, { 2333, 0x514B }, { 2334, 0x523B }, { 2335, 0x544A },
		{ 2336, 0x56FD }, { 2337, 0x7A40 }, { 2338, 0x9177 }, { 2339, 0x9D60 }, { 2340, 0x9ED2 }, { 2341, 0x7344 },
		{ 2342, 0x6F09 }, { 2343, 0x8170 }, { 2344, 0x7511 }, { 2345, 0x5FFD }, { 2346, 0x60DA }, { 2347, 0x9AA8 },
		{ 2348, 0x72DB }, { 2349, 0x8FBC }, { 2350, 0x6B64 }, { 2351, 0x9803 }, { 2352, 0x4ECA }, { 2353, 0x56F0 },
		{ 2354, 0x5764 }, { 2355, 0x58BE }, { 2356, 0x5A5A }, { 2357, 0x6068 }, { 2358, 0x61C7 }, { 2359, 0x660F },
		{ 2360, 0x6606 }, { 2361, 0x6839 }, { 2362, 0x68B1 }, { 2363, 0x6DF7 }, { 2364, 0x75D5 }, { 2365, 0x7D3A },
		{ 2366, 0x826E }, { 2367, 0x9B42 }, { 2368, 0x4E9B }, { 2369, 0x4F50 }, { 2370, 0x53C9 }, { 2371, 0x5506 },
		{ 2372, 0x5D6F }, { 2373, 0x5DE6 }, { 2374, 0x5DEE }, { 2375, 0x67FB }, { 2376, 0x6C99 }, { 2377, 0x7473 },
		{ 2378, 0x7802 }, { 2379, 0x8A50 }, { 2380, 0x9396 }, { 2381, 0x88DF }, { 2382, 0x5750 }, { 2383, 0x5EA7 },
		{ 2384, 0x632B }, { 2385, 0x50B5 }, { 2386, 0x50AC }, { 2387, 0x518D }, { 2388, 0x6700 }, { 2389, 0x54C9 },
		{ 2390, 0x585E }, { 2391, 0x59BB }, { 2392, 0x5BB0 }, { 2393, 0x5F69 }, { 2394, 0x624D }, { 2395, 0x63A1 },
		{ 2396, 0x683D }, { 2397, 0x6B73 }, { 2398, 0x6E08 }, { 2399, 0x707D }, { 2400, 0x91C7 }, { 2401, 0x7280 },
		{ 2402, 0x7815 }, { 2403, 0x7826 }, { 2404, 0x796D }, { 2405, 0x658E }, { 2406, 0x7D30 }, { 2407, 0x83DC },
		{ 2408, 0x88C1 }, { 2409, 0x8F09 }, { 2410, 0x969B }, { 2411, 0x5264 }, { 2412, 0x5728 }, { 2413, 0x6750 },
		{ 2414, 0x7F6A }, { 2415, 0x8CA1 }, { 2416, 0x51B4 }, { 2417, 0x5742 }, { 2418, 0x962A }, { 2419, 0x583A },
		{ 2420, 0x698A }, { 2421, 0x80B4 }, { 2422, 0x54B2 }, { 2423, 0x5D0E }, { 2424, 0x57FC }, { 2425, 0x7895 },
		{ 2426, 0x9DFA }, { 2427, 0x4F5C }, { 2428, 0x524A }, { 2429, 0x548B }, { 2430, 0x643E }, { 2431, 0x6628 },
		{ 2432, 0x6714 }, { 2433, 0x67F5 }, { 2434, 0x7A84 }, { 2435, 0x7B56 }, { 2436, 0x7D22 }, { 2437, 0x932F },
		{ 2438, 0x685C }, { 2439, 0x9BAD }, { 2440, 0x7B39 }, { 2441, 0x5319 }, { 2442, 0x518A }, { 2443, 0x5237 },
		{ 2444, 0x5BDF }, { 2445, 0x62F6 }, { 2446, 0x64AE }, { 2447, 0x64E6 }, { 2448, 0x672D }, { 2449, 0x6BBA },
		{ 2450, 0x85A9 }, { 2451, 0x96D1 }, { 2452, 0x7690 }, { 2453, 0x9BD6 }, { 2454, 0x634C }, { 2455, 0x9306 },
		{ 2456, 0x9BAB }, { 2457, 0x76BF }, { 2458, 0x6652 }, { 2459, 0x4E09 }, { 2460, 0x5098 }, { 2461, 0x53C2 },
		{ 2462, 0x5C71 }, { 2463, 0x60E8 }, { 2464, 0x6492 }, { 2465, 0x6563 }, { 2466, 0x685F }, { 2467, 0x71E6 },
		{ 2468, 0x73CA }, { 2469, 0x7523 }, { 2470, 0x7B97 }, { 2471, 0x7E82 }, { 2472, 0x8695 }, { 2473, 0x8B83 },
		{ 2474, 0x8CDB }, { 2475, 0x9178 }, { 2476, 0x9910 }, { 2477, 0x65AC }, { 2478, 0x66AB }, { 2479, 0x6B8B },
		{ 2480, 0x4ED5 }, { 2481, 0x4ED4 }, { 2482, 0x4F3A }, { 2483, 0x4F7F }, { 2484, 0x523A }, { 2485, 0x53F8 },
		{ 2486, 0x53F2 }, { 2487, 0x55E3 }, { 2488, 0x56DB }, { 2489, 0x58EB }, { 2490, 0x59CB }, { 2491, 0x59C9 },
		{ 2492, 0x59FF }, { 2493, 0x5B50 }, { 2494, 0x5C4D }, { 2495, 0x5E02 }, { 2496, 0x5E2B }, { 2497, 0x5FD7 },
		{ 2498, 0x601D }, { 2499, 0x6307 }, { 2500, 0x652F }, { 2501, 0x5B5C }, { 2502, 0x65AF }, { 2503, 0x65BD },
		{ 2504, 0x65E8 }, { 2505, 0x679D }, { 2506, 0x6B62 }, { 2507, 0x6B7B }, { 2508, 0x6C0F }, { 2509, 0x7345 },
		{ 2510, 0x7949 }, { 2511, 0x79C1 }, { 2512, 0x7CF8 }, { 2513, 0x7D19 }, { 2514, 0x7D2B }, { 2515, 0x80A2 },
		{ 2516, 0x8102 }, { 2517, 0x81F3 }, { 2518, 0x8996 }, { 2519, 0x8A5E }, { 2520, 0x8A69 }, { 2521, 0x8A66 },
		{ 2522, 0x8A8C }, { 2523, 0x8AEE }, { 2524, 0x8CC7 }, { 2525, 0x8CDC }, { 2526, 0x96CC }, { 2527, 0x98FC },
		{ 2528, 0x6B6F }, { 2529, 0x4E8B }, { 2530, 0x4F3C }, { 2531, 0x4F8D }, { 2532, 0x5150 }, { 2533, 0x5B57 },
		{ 2534, 0x5BFA }, { 2535, 0x6148 }, { 2536, 0x6301 }, { 2537, 0x6642 }, { 2538, 0x6B21 }, { 2539, 0x6ECB },
		{ 2540, 0x6CBB }, { 2541, 0x723E }, { 2542, 0x74BD }, { 2543, 0x75D4 }, { 2544, 0x78C1 }, { 2545, 0x793A },
		{ 2546, 0x800C }, { 2547, 0x8033 }, { 2548, 0x81EA }, { 2549, 0x8494 }, { 2550, 0x8F9E }, { 2551, 0x6C50 },
		{ 2552, 0x9E7F }, { 2553, 0x5F0F }, { 2554, 0x8B58 }, { 2555, 0x9D2B }, { 2556, 0x7AFA }, { 2557, 0x8EF8 },
		{ 2558, 0x5B8D }, { 2559, 0x96EB }, { 2560, 0x4E03 }, { 2561, 0x53F1 }, { 2562, 0x57F7 }, { 2563, 0x5931 },
		{ 2564, 0x5AC9 }, { 2565, 0x5BA4 }, { 2566, 0x6089 }, { 2567, 0x6E7F }, { 2568, 0x6F06 }, { 2569, 0x75BE },
		{ 2570, 0x8CEA }, { 2571, 0x5B9F }, { 2572, 0x8500 }, { 2573, 0x7BE0 }, { 2574, 0x5072 }, { 2575, 0x67F4 },
		{ 2576, 0x829D }, { 2577, 0x5C61 }, { 2578, 0x854A }, { 2579, 0x7E1E }, { 2580, 0x820E }, { 2581, 0x5199 },
		{ 2582, 0x5C04 }, { 2583, 0x6368 }, { 2584, 0x8D66 }, { 2585, 0x659C }, { 2586, 0x716E }, { 2587, 0x793E },
		{ 2588, 0x7D17 }, { 2589, 0x8005 }, { 2590, 0x8B1D }, { 2591, 0x8ECA }, { 2592, 0x906E }, { 2593, 0x86C7 },
		{ 2594, 0x90AA }, { 2595, 0x501F }, { 2596, 0x52FA }, { 2597, 0x5C3A }, { 2598, 0x6753 }, { 2599, 0x707C },
		{ 2600, 0x7235 }, { 2601, 0x914C }, { 2602, 0x91C8 }, { 2603, 0x932B }, { 2604, 0x82E5 }, { 2605, 0x5BC2 },
		{ 2606, 0x5F31 }, { 2607, 0x60F9 }, { 2608, 0x4E3B }, { 2609, 0x53D6 }, { 2610, 0x5B88 }, { 2611, 0x624B },
		{ 2612, 0x6731 }, { 2613, 0x6B8A }, { 2614, 0x72E9 }, { 2615, 0x73E0 }, { 2616, 0x7A2E }, { 2617, 0x816B },
		{ 2618, 0x8DA3 }, { 2619, 0x9152 }, { 2620, 0x9996 }, { 2621, 0x5112 }, { 2622, 0x53D7 }, { 2623, 0x546A },
		{ 2624, 0x5BFF }, { 2625, 0x6388 }, { 2626, 0x6A39 }, { 2627, 0x7DAC }, { 2628, 0x9700 }, { 2629, 0x56DA },
		{ 2630, 0x53CE }, { 2631, 0x5468 }, { 2632, 0x5B97 }, { 2633, 0x5C31 }, { 2634, 0x5DDE }, { 2635, 0x4FEE },
		{ 2636, 0x6101 }, { 2637, 0x62FE }, { 2638, 0x6D32 }, { 2639, 0x79C0 }, { 2640, 0x79CB }, { 2641, 0x7D42 },
		{ 2642, 0x7E4D }, { 2643, 0x7FD2 }, { 2644, 0x81ED }, { 2645, 0x821F }, { 2646, 0x8490 }, { 2647, 0x8846 },
		{ 2648, 0x8972 }, { 2649, 0x8B90 }, { 2650, 0x8E74 }, { 2651, 0x8F2F }, { 2652, 0x9031 }, { 2653, 0x914B },
		{ 2654, 0x916C }, { 2655, 0x96C6 }, { 2656, 0x919C }, { 2657, 0x4EC0 }, { 2658, 0x4F4F }, { 2659, 0x5145 },
		{ 2660, 0x5341 }, { 2661, 0x5F93 }, { 2662, 0x620E }, { 2663, 0x67D4 }, { 2664, 0x6C41 }, { 2665, 0x6E0B },
		{ 2666, 0x7363 }, { 2667, 0x7E26 }, { 2668, 0x91CD }, { 2669, 0x9283 }, { 2670, 0x53D4 }, { 2671, 0x5919 },
		{ 2672, 0x5BBF }, { 2673, 0x6DD1 }, { 2674, 0x795D }, { 2675, 0x7E2E }, { 2676, 0x7C9B }, { 2677, 0x587E },
		{ 2678, 0x719F }, { 2679, 0x51FA }, { 2680, 0x8853 }, { 2681, 0x8FF0 }, { 2682, 0x4FCA }, { 2683, 0x5CFB },
		{ 2684, 0x6625 }, { 2685, 0x77AC }, { 2686, 0x7AE3 }, { 2687, 0x821C }, { 2688, 0x99FF }, { 2689, 0x51C6 },
		{ 2690, 0x5FAA }, { 2691, 0x65EC }, { 2692, 0x696F }, { 2693, 0x6B89 }, { 2694, 0x6DF3 }, { 2695, 0x6E96 },
		{ 2696, 0x6F64 }, { 2697, 0x76FE }, { 2698, 0x7D14 }, { 2699, 0x5DE1 }, { 2700, 0x9075 }, { 2701, 0x9187 },
		{ 2702, 0x9806 }, { 2703, 0x51E6 }, { 2704, 0x521D }, { 2705, 0x6240 }, { 2706, 0x6691 }, { 2707, 0x66D9 },
		{ 2708, 0x6E1A }, { 2709, 0x5EB6 }, { 2710, 0x7DD2 }, { 2711, 0x7F72 }, { 2712, 0x66F8 }, { 2713, 0x85AF },
		{ 2714, 0x85F7 }, { 2715, 0x8AF8 }, { 2716, 0x52A9 }, { 2717, 0x53D9 }, { 2718, 0x5973 }, { 2719, 0x5E8F },
		{ 2720, 0x5F90 }, { 2721, 0x6055 }, { 2722, 0x92E4 }, { 2723, 0x9664 }, { 2724, 0x50B7 }, { 2725, 0x511F },
		{ 2726, 0x52DD }, { 2727, 0x5320 }, { 2728, 0x5347 }, { 2729, 0x53EC }, { 2730, 0x54E8 }, { 2731, 0x5546 },
		{ 2732, 0x5531 }, { 2733, 0x5617 }, { 2734, 0x5968 }, { 2735, 0x59BE }, { 2736, 0x5A3C }, { 2737, 0x5BB5 },
		{ 2738, 0x5C06 }, { 2739, 0x5C0F }, { 2740, 0x5C11 }, { 2741, 0x5C1A }, { 2742, 0x5E84 }, { 2743, 0x5E8A },
		{ 2744, 0x5EE0 }, { 2745, 0x5F70 }, { 2746, 0x627F }, { 2747, 0x6284 }, { 2748, 0x62DB }, { 2749, 0x638C },
		{ 2750, 0x6377 }, { 2751, 0x6607 }, { 2752, 0x660C }, { 2753, 0x662D }, { 2754, 0x6676 }, { 2755, 0x677E },
		{ 2756, 0x68A2 }, { 2757, 0x6A1F }, { 2758, 0x6A35 }, { 2759, 0x6CBC }, { 2760, 0x6D88 }, { 2761, 0x6E09 },
		{ 2762, 0x6E58 }, { 2763, 0x713C }, { 2764, 0x7126 }, { 2765, 0x7167 }, { 2766, 0x75C7 }, { 2767, 0x7701 },
		{ 2768, 0x785D }, { 2769, 0x7901 }, { 2770, 0x7965 }, { 2771, 0x79F0 }, { 2772, 0x7AE0 }, { 2773, 0x7B11 },
		{ 2774, 0x7CA7 }, { 2775, 0x7D39 }, { 2776, 0x8096 }, { 2777, 0x83D6 }, { 2778, 0x848B }, { 2779, 0x8549 },
		{ 2780, 0x885D }, { 2781, 0x88F3 }, { 2782, 0x8A1F }, { 2783, 0x8A3C }, { 2784, 0x8A54 }, { 2785, 0x8A73 },
		{ 2786, 0x8C61 }, { 2787, 0x8CDE }, { 2788, 0x91A4 }, { 2789, 0x9266 }, { 2790, 0x937E }, { 2791, 0x9418 },
		{ 2792, 0x969C }, { 2793, 0x9798 }, { 2794, 0x4E0A }, { 2795, 0x4E08 }, { 2796, 0x4E1E }, { 2797, 0x4E57 },
		{ 2798, 0x5197 }, { 2799, 0x5270 }, { 2800, 0x57CE }, { 2801, 0x5834 }, { 2802, 0x58CC }, { 2803, 0x5B22 },
		{ 2804, 0x5E38 }, { 2805, 0x60C5 }, { 2806, 0x64FE }, { 2807, 0x6761 }, { 2808, 0x6756 }, { 2809, 0x6D44 },
		{ 2810, 0x72B6 }, { 2811, 0x7573 }, { 2812, 0x7A63 }, { 2813, 0x84B8 }, { 2814, 0x8B72 }, { 2815, 0x91B8 },
		{ 2816, 0x9320 }, { 2817, 0x5631 }, { 2818, 0x57F4 }, { 2819, 0x98FE }, { 2820, 0x62ED }, { 2821, 0x690D },
		{ 2822, 0x6B96 }, { 2823, 0x71ED }, { 2824, 0x7E54 }, { 2825, 0x8077 }, { 2826, 0x8272 }, { 2827, 0x89E6 },
		{ 2828, 0x98DF }, { 2829, 0x8755 }, { 2830, 0x8FB1 }, { 2831, 0x5C3B }, { 2832, 0x4F38 }, { 2833, 0x4FE1 },
		{ 2834, 0x4FB5 }, { 2835, 0x5507 }, { 2836, 0x5A20 }, { 2837, 0x5BDD }, { 2838, 0x5BE9 }, { 2839, 0x5FC3 },
		{ 2840, 0x614E }, { 2841, 0x632F }, { 2842, 0x65B0 }, { 2843, 0x664B }, { 2844, 0x68EE }, { 2845, 0x699B },
		{ 2846, 0x6D78 }, { 2847, 0x6DF1 }, { 2848, 0x7533 }, { 2849, 0x75B9 }, { 2850, 0x771F }, { 2851, 0x795E },
		{ 2852, 0x79E6 }, { 2853, 0x7D33 }, { 2854, 0x81E3 }, { 2855, 0x82AF }, { 2856, 0x85AA }, { 2857, 0x89AA },
		{ 2858, 0x8A3A }, { 2859, 0x8EAB }, { 2860, 0x8F9B }, { 2861, 0x9032 }, { 2862, 0x91DD }, { 2863, 0x9707 },
		{ 2864, 0x4EBA }, { 2865, 0x4EC1 }, { 2866, 0x5203 }, { 2867, 0x5875 }, { 2868, 0x58EC }, { 2869, 0x5C0B },
		{ 2870, 0x751A }, { 2871, 0x5C3D }, { 2872, 0x814E }, { 2873, 0x8A0A }, { 2874, 0x8FC5 }, { 2875, 0x9663 },
		{ 2876, 0x976D }, { 2877, 0x7B25 }, { 2878, 0x8ACF }, { 2879, 0x9808 }, { 2880, 0x9162 }, { 2881, 0x56F3 },
		{ 2882, 0x53A8 }, { 2883, 0x9017 }, { 2884, 0x5439 }, { 2885, 0x5782 }, { 2886, 0x5E25 }, { 2887, 0x63A8 },
		{ 2888, 0x6C34 }, { 2889, 0x708A }, { 2890, 0x7761 }, { 2891, 0x7C8B }, { 2892, 0x7FE0 }, { 2893, 0x8870 },
		{ 2894, 0x9042 }, { 2895, 0x9154 }, { 2896, 0x9310 }, { 2897, 0x9318 }, { 2898, 0x968F }, { 2899, 0x745E },
		{ 2900, 0x9AC4 }, { 2901, 0x5D07 }, { 2902, 0x5D69 }, { 2903, 0x6570 }, { 2904, 0x67A2 }, { 2905, 0x8DA8 },
		{ 2906, 0x96DB }, { 2907, 0x636E }, { 2908, 0x6749 }, { 2909, 0x6919 }, { 2910, 0x83C5 }, { 2911, 0x9817 },
		{ 2912, 0x96C0 }, { 2913, 0x88FE }, { 2914, 0x6F84 }, { 2915, 0x647A }, { 2916, 0x5BF8 }, { 2917, 0x4E16 },
		{ 2918, 0x702C }, { 2919, 0x755D }, { 2920, 0x662F }, { 2921, 0x51C4 }, { 2922, 0x5236 }, { 2923, 0x52E2 },
		{ 2924, 0x59D3 }, { 2925, 0x5F81 }, { 2926, 0x6027 }, { 2927, 0x6210 }, { 2928, 0x653F }, { 2929, 0x6574 },
		{ 2930, 0x661F }, { 2931, 0x6674 }, { 2932, 0x68F2 }, { 2933, 0x6816 }, { 2934, 0x6B63 }, { 2935, 0x6E05 },
		{ 2936, 0x7272 }, { 2937, 0x751F }, { 2938, 0x76DB }, { 2939, 0x7CBE }, { 2940, 0x8056 }, { 2941, 0x58F0 },
		{ 2942, 0x88FD }, { 2943, 0x897F }, { 2944, 0x8AA0 }, { 2945, 0x8A93 }, { 2946, 0x8ACB }, { 2947, 0x901D },
		{ 2948, 0x9192 }, { 2949, 0x9752 }, { 2950, 0x9759 }, { 2951, 0x6589 }, { 2952, 0x7A0E }, { 2953, 0x8106 },
		{ 2954, 0x96BB }, { 2955, 0x5E2D }, { 2956, 0x60DC }, { 2957, 0x621A }, { 2958, 0x65A5 }, { 2959, 0x6614 },
		{ 2960, 0x6790 }, { 2961, 0x77F3 }, { 2962, 0x7A4D }, { 2963, 0x7C4D }, { 2964, 0x7E3E }, { 2965, 0x810A },
		{ 2966, 0x8CAC }, { 2967, 0x8D64 }, { 2968, 0x8DE1 }, { 2969, 0x8E5F }, { 2970, 0x78A9 }, { 2971, 0x5207 },
		{ 2972, 0x62D9 }, { 2973, 0x63A5 }, { 2974, 0x6442 }, { 2975, 0x6298 }, { 2976, 0x8A2D }, { 2977, 0x7A83 },
		{ 2978, 0x7BC0 }, { 2979, 0x8AAC }, { 2980, 0x96EA }, { 2981, 0x7D76 }, { 2982, 0x820C }, { 2983, 0x8749 },
		{ 2984, 0x4ED9 }, { 2985, 0x5148 }, { 2986, 0x5343 }, { 2987, 0x5360 }, { 2988, 0x5BA3 }, { 2989, 0x5C02 },
		{ 2990, 0x5C16 }, { 2991, 0x5DDD }, { 2992, 0x6226 }, { 2993, 0x6247 }, { 2994, 0x64B0 }, { 2995, 0x6813 },
		{ 2996, 0x6834 }, { 2997, 0x6CC9 }, { 2998, 0x6D45 }, { 2999, 0x6D17 }, { 3000, 0x67D3 }, { 3001, 0x6F5C },
		{ 3002, 0x714E }, { 3003, 0x717D }, { 3004, 0x65CB }, { 3005, 0x7A7F }, { 3006, 0x7BAD }, { 3007, 0x7DDA },
		{ 3008, 0x7E4A }, { 3009, 0x7FA8 }, { 3010, 0x817A }, { 3011, 0x821B }, { 3012, 0x8239 }, { 3013, 0x85A6 },
		{ 3014, 0x8A6E }, { 3015, 0x8CCE }, { 3016, 0x8DF5 }, { 3017, 0x9078 }, { 3018, 0x9077 }, { 3019, 0x92AD },
		{ 3020, 0x9291 }, { 3021, 0x9583 }, { 3022, 0x9BAE }, { 3023, 0x524D }, { 3024, 0x5584 }, { 3025, 0x6F38 },
		{ 3026, 0x7136 }, { 3027, 0x5168 }, { 3028, 0x7985 }, { 3029, 0x7E55 }, { 3030, 0x81B3 }, { 3031, 0x7CCE },
		{ 3032, 0x564C }, { 3033, 0x5851 }, { 3034, 0x5CA8 }, { 3035, 0x63AA }, { 3036, 0x66FE }, { 3037, 0x66FD },
		{ 3038, 0x695A }, { 3039, 0x72D9 }, { 3040, 0x758F }, { 3041, 0x758E }, { 3042, 0x790E }, { 3043, 0x7956 },
		{ 3044, 0x79DF }, { 3045, 0x7C97 }, { 3046, 0x7D20 }, { 3047, 0x7D44 }, { 3048, 0x8607 }, { 3049, 0x8A34 },
		{ 3050, 0x963B }, { 3051, 0x9061 }, { 3052, 0x9F20 }, { 3053, 0x50E7 }, { 3054, 0x5275 }, { 3055, 0x53CC },
		{ 3056, 0x53E2 }, { 3057, 0x5009 }, { 3058, 0x55AA }, { 3059, 0x58EE }, { 3060, 0x594F }, { 3061, 0x723D },
		{ 3062, 0x5B8B }, { 3063, 0x5C64 }, { 3064, 0x531D }, { 3065, 0x60E3 }, { 3066, 0x60F3 }, { 3067, 0x635C },
		{ 3068, 0x6383 }, { 3069, 0x633F }, { 3070, 0x63BB }, { 3071, 0x64CD }, { 3072, 0x65E9 }, { 3073, 0x66F9 },
		{ 3074, 0x5DE3 }, { 3075, 0x69CD }, { 3076, 0x69FD }, { 3077, 0x6F15 }, { 3078, 0x71E5 }, { 3079, 0x4E89 },
		{ 3080, 0x75E9 }, { 3081, 0x76F8 }, { 3082, 0x7A93 }, { 3083, 0x7CDF }, { 3084, 0x7DCF }, { 3085, 0x7D9C },
		{ 3086, 0x8061 }, { 3087, 0x8349 }, { 3088, 0x8358 }, { 3089, 0x846C }, { 3090, 0x84BC }, { 3091, 0x85FB },
		{ 3092, 0x88C5 }, { 3093, 0x8D70 }, { 3094, 0x9001 }, { 3095, 0x906D }, { 3096, 0x9397 }, { 3097, 0x971C },
		{ 3098, 0x9A12 }, { 3099, 0x50CF }, { 3100, 0x5897 }, { 3101, 0x618E }, { 3102, 0x81D3 }, { 3103, 0x8535 },
		{ 3104, 0x8D08 }, { 3105, 0x9020 }, { 3106, 0x4FC3 }, { 3107, 0x5074 }, { 3108, 0x5247 }, { 3109, 0x5373 },
		{ 3110, 0x606F }, { 3111, 0x6349 }, { 3112, 0x675F }, { 3113, 0x6E2C }, { 3114, 0x8DB3 }, { 3115, 0x901F },
		{ 3116, 0x4FD7 }, { 3117, 0x5C5E }, { 3118, 0x8CCA }, { 3119, 0x65CF }, { 3120, 0x7D9A }, { 3121, 0x5352 },
		{ 3122, 0x8896 }, { 3123, 0x5176 }, { 3124, 0x63C3 }, { 3125, 0x5B58 }, { 3126, 0x5B6B }, { 3127, 0x5C0A },
		{ 3128, 0x640D }, { 3129, 0x6751 }, { 3130, 0x905C }, { 3131, 0x4ED6 }, { 3132, 0x591A }, { 3133, 0x592A },
		{ 3134, 0x6C70 }, { 3135, 0x8A51 }, { 3136, 0x553E }, { 3137, 0x5815 }, { 3138, 0x59A5 }, { 3139, 0x60F0 },
		{ 3140, 0x6253 }, { 3141, 0x67C1 }, { 3142, 0x8235 }, { 3143, 0x6955 }, { 3144, 0x9640 }, { 3145, 0x99C4 },
		{ 3146, 0x9A28 }, { 3147, 0x4F53 }, { 3148, 0x5806 }, { 3149, 0x5BFE }, { 3150, 0x8010 }, { 3151, 0x5CB1 },
		{ 3152, 0x5E2F }, { 3153, 0x5F85 }, { 3154, 0x6020 }, { 3155, 0x614B }, { 3156, 0x6234 }, { 3157, 0x66FF },
		{ 3158, 0x6CF0 }, { 3159, 0x6EDE }, { 3160, 0x80CE }, { 3161, 0x817F }, { 3162, 0x82D4 }, { 3163, 0x888B },
		{ 3164, 0x8CB8 }, { 3165, 0x9000 }, { 3166, 0x902E }, { 3167, 0x968A }, { 3168, 0x9EDB }, { 3169, 0x9BDB },
		{ 3170, 0x4EE3 }, { 3171, 0x53F0 }, { 3172, 0x5927 }, { 3173, 0x7B2C }, { 3174, 0x918D }, { 3175, 0x984C },
		{ 3176, 0x9DF9 }, { 3177, 0x6EDD }, { 3178, 0x7027 }, { 3179, 0x5353 }, { 3180, 0x5544 }, { 3181, 0x5B85 },
		{ 3182, 0x6258 }, { 3183, 0x629E }, { 3184, 0x62D3 }, { 3185, 0x6CA2 }, { 3186, 0x6FEF }, { 3187, 0x7422 },
		{ 3188, 0x8A17 }, { 3189, 0x9438 }, { 3190, 0x6FC1 }, { 3191, 0x8AFE }, { 3192, 0x8338 }, { 3193, 0x51E7 },
		{ 3194, 0x86F8 }, { 3195, 0x53EA }, { 3196, 0x53E9 }, { 3197, 0x4F46 }, { 3198, 0x9054 }, { 3199, 0x8FB0 },
		{ 3200, 0x596A }, { 3201, 0x8131 }, { 3202, 0x5DFD }, { 3203, 0x7AEA }, { 3204, 0x8FBF }, { 3205, 0x68DA },
		{ 3206, 0x8C37 }, { 3207, 0x72F8 }, { 3208, 0x9C48 }, { 3209, 0x6A3D }, { 3210, 0x8AB0 }, { 3211, 0x4E39 },
		{ 3212, 0x5358 }, { 3213, 0x5606 }, { 3214, 0x5766 }, { 3215, 0x62C5 }, { 3216, 0x63A2 }, { 3217, 0x65E6 },
		{ 3218, 0x6B4E }, { 3219, 0x6DE1 }, { 3220, 0x6E5B }, { 3221, 0x70AD }, { 3222, 0x77ED }, { 3223, 0x7AEF },
		{ 3224, 0x7BAA }, { 3225, 0x7DBB }, { 3226, 0x803D }, { 3227, 0x80C6 }, { 3228, 0x86CB }, { 3229, 0x8A95 },
		{ 3230, 0x935B }, { 3231, 0x56E3 }, { 3232, 0x58C7 }, { 3233, 0x5F3E }, { 3234, 0x65AD }, { 3235, 0x6696 },
		{ 3236, 0x6A80 }, { 3237, 0x6BB5 }, { 3238, 0x7537 }, { 3239, 0x8AC7 }, { 3240, 0x5024 }, { 3241, 0x77E5 },
		{ 3242, 0x5730 }, { 3243, 0x5F1B }, { 3244, 0x6065 }, { 3245, 0x667A }, { 3246, 0x6C60 }, { 3247, 0x75F4 },
		{ 3248, 0x7A1A }, { 3249, 0x7F6E }, { 3250, 0x81F4 }, { 3251, 0x8718 }, { 3252, 0x9045 }, { 3253, 0x99B3 },
		{ 3254, 0x7BC9 }, { 3255, 0x755C }, { 3256, 0x7AF9 }, { 3257, 0x7B51 }, { 3258, 0x84C4 }, { 3259, 0x9010 },
		{ 3260, 0x79E9 }, { 3261, 0x7A92 }, { 3262, 0x8336 }, { 3263, 0x5AE1 }, { 3264, 0x7740 }, { 3265, 0x4E2D },
		{ 3266, 0x4EF2 }, { 3267, 0x5B99 }, { 3268, 0x5FE0 }, { 3269, 0x62BD }, { 3270, 0x663C }, { 3271, 0x67F1 },
		{ 3272, 0x6CE8 }, { 3273, 0x866B }, { 3274, 0x8877 }, { 3275, 0x8A3B }, { 3276, 0x914E }, { 3277, 0x92F3 },
		{ 3278, 0x99D0 }, { 3279, 0x6A17 }, { 3280, 0x7026 }, { 3281, 0x732A }, { 3282, 0x82E7 }, { 3283, 0x8457 },
		{ 3284, 0x8CAF }, { 3285, 0x4E01 }, { 3286, 0x5146 }, { 3287, 0x51CB }, { 3288, 0x558B }, { 3289, 0x5BF5 },
		{ 3290, 0x5E16 }, { 3291, 0x5E33 }, { 3292, 0x5E81 }, { 3293, 0x5F14 }, { 3294, 0x5F35 }, { 3295, 0x5F6B },
		{ 3296, 0x5FB4 }, { 3297, 0x61F2 }, { 3298, 0x6311 }, { 3299, 0x66A2 }, { 3300, 0x671D }, { 3301, 0x6F6E },
		{ 3302, 0x7252 }, { 3303, 0x753A }, { 3304, 0x773A }, { 3305, 0x8074 }, { 3306, 0x8139 }, { 3307, 0x8178 },
		{ 3308, 0x8776 }, { 3309, 0x8ABF }, { 3310, 0x8ADC }, { 3311, 0x8D85 }, { 3312, 0x8DF3 }, { 3313, 0x929A },
		{ 3314, 0x9577 }, { 3315, 0x9802 }, { 3316, 0x9CE5 }, { 3317, 0x52C5 }, { 3318, 0x6357 }, { 3319, 0x76F4 },
		{ 3320, 0x6715 }, { 3321, 0x6C88 }, { 3322, 0x73CD }, { 3323, 0x8CC3 }, { 3324, 0x93AE }, { 3325, 0x9673 },
		{ 3326, 0x6D25 }, { 3327, 0x589C }, { 3328, 0x690E }, { 3329, 0x69CC }, { 3330, 0x8FFD }, { 3331, 0x939A },
		{ 3332, 0x75DB }, { 3333, 0x901A }, { 3334, 0x585A }, { 3335, 0x6802 }, { 3336, 0x63B4 }, { 3337, 0x69FB },
		{ 3338, 0x4F43 }, { 3339, 0x6F2C }, { 3340, 0x67D8 }, { 3341, 0x8FBB }, { 3342, 0x8526 }, { 3343, 0x7DB4 },
		{ 3344, 0x9354 }, { 3345, 0x693F }, { 3346, 0x6F70 }, { 3347, 0x576A }, { 3348, 0x58F7 }, { 3349, 0x5B2C },
		{ 3350, 0x7D2C }, { 3351, 0x722A }, { 3352, 0x540A }, { 3353, 0x91E3 }, { 3354, 0x9DB4 }, { 3355, 0x4EAD },
		{ 3356, 0x4F4E }, { 3357, 0x505C }, { 3358, 0x5075 }, { 3359, 0x5243 }, { 3360, 0x8C9E }, { 3361, 0x5448 },
		{ 3362, 0x5824 }, { 3363, 0x5B9A }, { 3364, 0x5E1D }, { 3365, 0x5E95 }, { 3366, 0x5EAD }, { 3367, 0x5EF7 },
		{ 3368, 0x5F1F }, { 3369, 0x608C }, { 3370, 0x62B5 }, { 3371, 0x633A }, { 3372, 0x63D0 }, { 3373, 0x68AF },
		{ 3374, 0x6C40 }, { 3375, 0x7887 }, { 3376, 0x798E }, { 3377, 0x7A0B }, { 3378, 0x7DE0 }, { 3379, 0x8247 },
		{ 3380, 0x8A02 }, { 3381, 0x8AE6 }, { 3382, 0x8E44 }, { 3383, 0x9013 }, { 3384, 0x90B8 }, { 3385, 0x912D },
		{ 3386, 0x91D8 }, { 3387, 0x9F0E }, { 3388, 0x6CE5 }, { 3389, 0x6458 }, { 3390, 0x64E2 }, { 3391, 0x6575 },
		{ 3392, 0x6EF4 }, { 3393, 0x7684 }, { 3394, 0x7B1B }, { 3395, 0x9069 }, { 3396, 0x93D1 }, { 3397, 0x6EBA },
		{ 3398, 0x54F2 }, { 3399, 0x5FB9 }, { 3400, 0x64A4 }, { 3401, 0x8F4D }, { 3402, 0x8FED }, { 3403, 0x9244 },
		{ 3404, 0x5178 }, { 3405, 0x586B }, { 3406, 0x5929 }, { 3407, 0x5C55 }, { 3408, 0x5E97 }, { 3409, 0x6DFB },
		{ 3410, 0x7E8F }, { 3411, 0x751C }, { 3412, 0x8CBC }, { 3413, 0x8EE2 }, { 3414, 0x985B }, { 3415, 0x70B9 },
		{ 3416, 0x4F1D }, { 3417, 0x6BBF }, { 3418, 0x6FB1 }, { 3419, 0x7530 }, { 3420, 0x96FB }, { 3421, 0x514E },
		{ 3422, 0x5410 }, { 3423, 0x5835 }, { 3424, 0x5857 }, { 3425, 0x59AC }, { 3426, 0x5C60 }, { 3427, 0x5F92 },
		{ 3428, 0x6597 }, { 3429, 0x675C }, { 3430, 0x6E21 }, { 3431, 0x767B }, { 3432, 0x83DF }, { 3433, 0x8CED },
		{ 3434, 0x9014 }, { 3435, 0x90FD }, { 3436, 0x934D }, { 3437, 0x7825 }, { 3438, 0x783A }, { 3439, 0x52AA },
		{ 3440, 0x5EA6 }, { 3441, 0x571F }, { 3442, 0x5974 }, { 3443, 0x6012 }, { 3444, 0x5012 }, { 3445, 0x515A },
		{ 3446, 0x51AC }, { 3447, 0x51CD }, { 3448, 0x5200 }, { 3449, 0x5510 }, { 3450, 0x5854 }, { 3451, 0x5858 },
		{ 3452, 0x5957 }, { 3453, 0x5B95 }, { 3454, 0x5CF6 }, { 3455, 0x5D8B }, { 3456, 0x60BC }, { 3457, 0x6295 },
		{ 3458, 0x642D }, { 3459, 0x6771 }, { 3460, 0x6843 }, { 3461, 0x68BC }, { 3462, 0x68DF }, { 3463, 0x76D7 },
		{ 3464, 0x6DD8 }, { 3465, 0x6E6F }, { 3466, 0x6D9B }, { 3467, 0x706F }, { 3468, 0x71C8 }, { 3469, 0x5F53 },
		{ 3470, 0x75D8 }, { 3471, 0x7977 }, { 3472, 0x7B49 }, { 3473, 0x7B54 }, { 3474, 0x7B52 }, { 3475, 0x7CD6 },
		{ 3476, 0x7D71 }, { 3477, 0x5230 }, { 3478, 0x8463 }, { 3479, 0x8569 }, { 3480, 0x85E4 }, { 3481, 0x8A0E },
		{ 3482, 0x8B04 }, { 3483, 0x8C46 }, { 3484, 0x8E0F }, { 3485, 0x9003 }, { 3486, 0x900F }, { 3487, 0x9419 },
		{ 3488, 0x9676 }, { 3489, 0x982D }, { 3490, 0x9A30 }, { 3491, 0x95D8 }, { 3492, 0x50CD }, { 3493, 0x52D5 },
		{ 3494, 0x540C }, { 3495, 0x5802 }, { 3496, 0x5C0E }, { 3497, 0x61A7 }, { 3498, 0x649E }, { 3499, 0x6D1E },
		{ 3500, 0x77B3 }, { 3501, 0x7AE5 }, { 3502, 0x80F4 }, { 3503, 0x8404 }, { 3504, 0x9053 }, { 3505, 0x9285 },
		{ 3506, 0x5CE0 }, { 3507, 0x9D07 }, { 3508, 0x533F }, { 3509, 0x5F97 }, { 3510, 0x5FB3 }, { 3511, 0x6D9C },
		{ 3512, 0x7279 }, { 3513, 0x7763 }, { 3514, 0x79BF }, { 3515, 0x7BE4 }, { 3516, 0x6BD2 }, { 3517, 0x72EC },
		{ 3518, 0x8AAD }, { 3519, 0x6803 }, { 3520, 0x6A61 }, { 3521, 0x51F8 }, { 3522, 0x7A81 }, { 3523, 0x6934 },
		{ 3524, 0x5C4A }, { 3525, 0x9CF6 }, { 3526, 0x82EB }, { 3527, 0x5BC5 }, { 3528, 0x9149 }, { 3529, 0x701E },
		{ 3530, 0x5678 }, { 3531, 0x5C6F }, { 3532, 0x60C7 }, { 3533, 0x6566 }, { 3534, 0x6C8C }, { 3535, 0x8C5A },
		{ 3536, 0x9041 }, { 3537, 0x9813 }, { 3538, 0x5451 }, { 3539, 0x66C7 }, { 3540, 0x920D }, { 3541, 0x5948 },
		{ 3542, 0x90A3 }, { 3543, 0x5185 }, { 3544, 0x4E4D }, { 3545, 0x51EA }, { 3546, 0x8599 }, { 3547, 0x8B0E },
		{ 3548, 0x7058 }, { 3549, 0x637A }, { 3550, 0x934B }, { 3551, 0x6962 }, { 3552, 0x99B4 }, { 3553, 0x7E04 },
		{ 3554, 0x7577 }, { 3555, 0x5357 }, { 3556, 0x6960 }, { 3557, 0x8EDF }, { 3558, 0x96E3 }, { 3559, 0x6C5D },
		{ 3560, 0x4E8C }, { 3561, 0x5C3C }, { 3562, 0x5F10 }, { 3563, 0x8FE9 }, { 3564, 0x5302 }, { 3565, 0x8CD1 },
		{ 3566, 0x8089 }, { 3567, 0x8679 }, { 3568, 0x5EFF }, { 3569, 0x65E5 }, { 3570, 0x4E73 }, { 3571, 0x5165 },
		{ 3572, 0x5982 }, { 3573, 0x5C3F }, { 3574, 0x97EE }, { 3575, 0x4EFB }, { 3576, 0x598A }, { 3577, 0x5FCD },
		{ 3578, 0x8A8D }, { 3579, 0x6FE1 }, { 3580, 0x79B0 }, { 3581, 0x7962 }, { 3582, 0x5BE7 }, { 3583, 0x8471 },
		{ 3584, 0x732B }, { 3585, 0x71B1 }, { 3586, 0x5E74 }, { 3587, 0x5FF5 }, { 3588, 0x637B }, { 3589, 0x649A },
		{ 3590, 0x71C3 }, { 3591, 0x7C98 }, { 3592, 0x4E43 }, { 3593, 0x5EFC }, { 3594, 0x4E4B }, { 3595, 0x57DC },
		{ 3596, 0x56A2 }, { 3597, 0x60A9 }, { 3598, 0x6FC3 }, { 3599, 0x7D0D }, { 3600, 0x80FD }, { 3601, 0x8133 },
		{ 3602, 0x81BF }, { 3603, 0x8FB2 }, { 3604, 0x8997 }, { 3605, 0x86A4 }, { 3606, 0x5DF4 }, { 3607, 0x628A },
		{ 3608, 0x64AD }, { 3609, 0x8987 }, { 3610, 0x6777 }, { 3611, 0x6CE2 }, { 3612, 0x6D3E }, { 3613, 0x7436 },
		{ 3614, 0x7834 }, { 3615, 0x5A46 }, { 3616, 0x7F75 }, { 3617, 0x82AD }, { 3618, 0x99AC }, { 3619, 0x4FF3 },
		{ 3620, 0x5EC3 }, { 3621, 0x62DD }, { 3622, 0x6392 }, { 3623, 0x6557 }, { 3624, 0x676F }, { 3625, 0x76C3 },
		{ 3626, 0x724C }, { 3627, 0x80CC }, { 3628, 0x80BA }, { 3629, 0x8F29 }, { 3630, 0x914D }, { 3631, 0x500D },
		{ 3632, 0x57F9 }, { 3633, 0x5A92 }, { 3634, 0x6885 }, { 3635, 0x6973 }, { 3636, 0x7164 }, { 3637, 0x72FD },
		{ 3638, 0x8CB7 }, { 3639, 0x58F2 }, { 3640, 0x8CE0 }, { 3641, 0x966A }, { 3642, 0x9019 }, { 3643, 0x877F },
		{ 3644, 0x79E4 }, { 3645, 0x77E7 }, { 3646, 0x8429 }, { 3647, 0x4F2F }, { 3648, 0x5265 }, { 3649, 0x535A },
		{ 3650, 0x62CD }, { 3651, 0x67CF }, { 3652, 0x6CCA }, { 3653, 0x767D }, { 3654, 0x7B94 }, { 3655, 0x7C95 },
		{ 3656, 0x8236 }, { 3657, 0x8584 }, { 3658, 0x8FEB }, { 3659, 0x66DD }, { 3660, 0x6F20 }, { 3661, 0x7206 },
		{ 3662, 0x7E1B }, { 3663, 0x83AB }, { 3664, 0x99C1 }, { 3665, 0x9EA6 }, { 3666, 0x51FD }, { 3667, 0x7BB1 },
		{ 3668, 0x7872 }, { 3669, 0x7BB8 }, { 3670, 0x8087 }, { 3671, 0x7B48 }, { 3672, 0x6AE8 }, { 3673, 0x5E61 },
		{ 3674, 0x808C }, { 3675, 0x7551 }, { 3676, 0x7560 }, { 3677, 0x516B }, { 3678, 0x9262 }, { 3679, 0x6E8C },
		{ 3680, 0x767A }, { 3681, 0x9197 }, { 3682, 0x9AEA }, { 3683, 0x4F10 }, { 3684, 0x7F70 }, { 3685, 0x629C },
		{ 3686, 0x7B4F }, { 3687, 0x95A5 }, { 3688, 0x9CE9 }, { 3689, 0x567A }, { 3690, 0x5859 }, { 3691, 0x86E4 },
		{ 3692, 0x96BC }, { 3693, 0x4F34 }, { 3694, 0x5224 }, { 3695, 0x534A }, { 3696, 0x53CD }, { 3697, 0x53DB },
		{ 3698, 0x5E06 }, { 3699, 0x642C }, { 3700, 0x6591 }, { 3701, 0x677F }, { 3702, 0x6C3E }, { 3703, 0x6C4E },
		{ 3704, 0x7248 }, { 3705, 0x72AF }, { 3706, 0x73ED }, { 3707, 0x7554 }, { 3708, 0x7E41 }, { 3709, 0x822C },
		{ 3710, 0x85E9 }, { 3711, 0x8CA9 }, { 3712, 0x7BC4 }, { 3713, 0x91C6 }, { 3714, 0x7169 }, { 3715, 0x9812 },
		{ 3716, 0x98EF }, { 3717, 0x633D }, { 3718, 0x6669 }, { 3719, 0x756A }, { 3720, 0x76E4 }, { 3721, 0x78D0 },
		{ 3722, 0x8543 }, { 3723, 0x86EE }, { 3724, 0x532A }, { 3725, 0x5351 }, { 3726, 0x5426 }, { 3727, 0x5983 },
		{ 3728, 0x5E87 }, { 3729, 0x5F7C }, { 3730, 0x60B2 }, { 3731, 0x6249 }, { 3732, 0x6279 }, { 3733, 0x62AB },
		{ 3734, 0x6590 }, { 3735, 0x6BD4 }, { 3736, 0x6CCC }, { 3737, 0x75B2 }, { 3738, 0x76AE }, { 3739, 0x7891 },
		{ 3740, 0x79D8 }, { 3741, 0x7DCB }, { 3742, 0x7F77 }, { 3743, 0x80A5 }, { 3744, 0x88AB }, { 3745, 0x8AB9 },
		{ 3746, 0x8CBB }, { 3747, 0x907F }, { 3748, 0x975E }, { 3749, 0x98DB }, { 3750, 0x6A0B }, { 3751, 0x7C38 },
		{ 3752, 0x5099 }, { 3753, 0x5C3E }, { 3754, 0x5FAE }, { 3755, 0x6787 }, { 3756, 0x6BD8 }, { 3757, 0x7435 },
		{ 3758, 0x7709 }, { 3759, 0x7F8E }, { 3760, 0x9F3B }, { 3761, 0x67CA }, { 3762, 0x7A17 }, { 3763, 0x5339 },
		{ 3764, 0x758B }, { 3765, 0x9AED }, { 3766, 0x5F66 }, { 3767, 0x819D }, { 3768, 0x83F1 }, { 3769, 0x8098 },
		{ 3770, 0x5F3C }, { 3771, 0x5FC5 }, { 3772, 0x7562 }, { 3773, 0x7B46 }, { 3774, 0x903C }, { 3775, 0x6867 },
		{ 3776, 0x59EB }, { 3777, 0x5A9B }, { 3778, 0x7D10 }, { 3779, 0x767E }, { 3780, 0x8B2C }, { 3781, 0x4FF5 },
		{ 3782, 0x5F6A }, { 3783, 0x6A19 }, { 3784, 0x6C37 }, { 3785, 0x6F02 }, { 3786, 0x74E2 }, { 3787, 0x7968 },
		{ 3788, 0x8868 }, { 3789, 0x8A55 }, { 3790, 0x8C79 }, { 3791, 0x5EDF }, { 3792, 0x63CF }, { 3793, 0x75C5 },
		{ 3794, 0x79D2 }, { 3795, 0x82D7 }, { 3796, 0x9328 }, { 3797, 0x92F2 }, { 3798, 0x849C }, { 3799, 0x86ED },
		{ 3800, 0x9C2D }, { 3801, 0x54C1 }, { 3802, 0x5F6C }, { 3803, 0x658C }, { 3804, 0x6D5C }, { 3805, 0x7015 },
		{ 3806, 0x8CA7 }, { 3807, 0x8CD3 }, { 3808, 0x983B }, { 3809, 0x654F }, { 3810, 0x74F6 }, { 3811, 0x4E0D },
		{ 3812, 0x4ED8 }, { 3813, 0x57E0 }, { 3814, 0x592B }, { 3815, 0x5A66 }, { 3816, 0x5BCC }, { 3817, 0x51A8 },
		{ 3818, 0x5E03 }, { 3819, 0x5E9C }, { 3820, 0x6016 }, { 3821, 0x6276 }, { 3822, 0x6577 }, { 3823, 0x65A7 },
		{ 3824, 0x666E }, { 3825, 0x6D6E }, { 3826, 0x7236 }, { 3827, 0x7B26 }, { 3828, 0x8150 }, { 3829, 0x819A },
		{ 3830, 0x8299 }, { 3831, 0x8B5C }, { 3832, 0x8CA0 }, { 3833, 0x8CE6 }, { 3834, 0x8D74 }, { 3835, 0x961C },
		{ 3836, 0x9644 }, { 3837, 0x4FAE }, { 3838, 0x64AB }, { 3839, 0x6B66 }, { 3840, 0x821E }, { 3841, 0x8461 },
		{ 3842, 0x856A }, { 3843, 0x90E8 }, { 3844, 0x5C01 }, { 3845, 0x6953 }, { 3846, 0x98A8 }, { 3847, 0x847A },
		{ 3848, 0x8557 }, { 3849, 0x4F0F }, { 3850, 0x526F }, { 3851, 0x5FA9 }, { 3852, 0x5E45 }, { 3853, 0x670D },
		{ 3854, 0x798F }, { 3855, 0x8179 }, { 3856, 0x8907 }, { 3857, 0x8986 }, { 3858, 0x6DF5 }, { 3859, 0x5F17 },
		{ 3860, 0x6255 }, { 3861, 0x6CB8 }, { 3862, 0x4ECF }, { 3863, 0x7269 }, { 3864, 0x9B92 }, { 3865, 0x5206 },
		{ 3866, 0x543B }, { 3867, 0x5674 }, { 3868, 0x58B3 }, { 3869, 0x61A4 }, { 3870, 0x626E }, { 3871, 0x711A },
		{ 3872, 0x596E }, { 3873, 0x7C89 }, { 3874, 0x7CDE }, { 3875, 0x7D1B }, { 3876, 0x96F0 }, { 3877, 0x6587 },
		{ 3878, 0x805E }, { 3879, 0x4E19 }, { 3880, 0x4F75 }, { 3881, 0x5175 }, { 3882, 0x5840 }, { 3883, 0x5E63 },
		{ 3884, 0x5E73 }, { 3885, 0x5F0A }, { 3886, 0x67C4 }, { 3887, 0x4E26 }, { 3888, 0x853D }, { 3889, 0x9589 },
		{ 3890, 0x965B }, { 3891, 0x7C73 }, { 3892, 0x9801 }, { 3893, 0x50FB }, { 3894, 0x58C1 }, { 3895, 0x7656 },
		{ 3896, 0x78A7 }, { 3897, 0x5225 }, { 3898, 0x77A5 }, { 3899, 0x8511 }, { 3900, 0x7B86 }, { 3901, 0x504F },
		{ 3902, 0x5909 }, { 3903, 0x7247 }, { 3904, 0x7BC7 }, { 3905, 0x7DE8 }, { 3906, 0x8FBA }, { 3907, 0x8FD4 },
		{ 3908, 0x904D }, { 3909, 0x4FBF }, { 3910, 0x52C9 }, { 3911, 0x5A29 }, { 3912, 0x5F01 }, { 3913, 0x97AD },
		{ 3914, 0x4FDD }, { 3915, 0x8217 }, { 3916, 0x92EA }, { 3917, 0x5703 }, { 3918, 0x6355 }, { 3919, 0x6B69 },
		{ 3920, 0x752B }, { 3921, 0x88DC }, { 3922, 0x8F14 }, { 3923, 0x7A42 }, { 3924, 0x52DF }, { 3925, 0x5893 },
		{ 3926, 0x6155 }, { 3927, 0x620A }, { 3928, 0x66AE }, { 3929, 0x6BCD }, { 3930, 0x7C3F }, { 3931, 0x83E9 },
		{ 3932, 0x5023 }, { 3933, 0x4FF8 }, { 3934, 0x5305 }, { 3935, 0x5446 }, { 3936, 0x5831 }, { 3937, 0x5949 },
		{ 3938, 0x5B9D }, { 3939, 0x5CF0 }, { 3940, 0x5CEF }, { 3941, 0x5D29 }, { 3942, 0x5E96 }, { 3943, 0x62B1 },
		{ 3944, 0x6367 }, { 3945, 0x653E }, { 3946, 0x65B9 }, { 3947, 0x670B }, { 3948, 0x6CD5 }, { 3949, 0x6CE1 },
		{ 3950, 0x70F9 }, { 3951, 0x7832 }, { 3952, 0x7E2B }, { 3953, 0x80DE }, { 3954, 0x82B3 }, { 3955, 0x840C },
		{ 3956, 0x84EC }, { 3957, 0x8702 }, { 3958, 0x8912 }, { 3959, 0x8A2A }, { 3960, 0x8C4A }, { 3961, 0x90A6 },
		{ 3962, 0x92D2 }, { 3963, 0x98FD }, { 3964, 0x9CF3 }, { 3965, 0x9D6C }, { 3966, 0x4E4F }, { 3967, 0x4EA1 },
		{ 3968, 0x508D }, { 3969, 0x5256 }, { 3970, 0x574A }, { 3971, 0x59A8 }, { 3972, 0x5E3D }, { 3973, 0x5FD8 },
		{ 3974, 0x5FD9 }, { 3975, 0x623F }, { 3976, 0x66B4 }, { 3977, 0x671B }, { 3978, 0x67D0 }, { 3979, 0x68D2 },
		{ 3980, 0x5192 }, { 3981, 0x7D21 }, { 3982, 0x80AA }, { 3983, 0x81A8 }, { 3984, 0x8B00 }, { 3985, 0x8C8C },
		{ 3986, 0x8CBF }, { 3987, 0x927E }, { 3988, 0x9632 }, { 3989, 0x5420 }, { 3990, 0x982C }, { 3991, 0x5317 },
		{ 3992, 0x50D5 }, { 3993, 0x535C }, { 3994, 0x58A8 }, { 3995, 0x64B2 }, { 3996, 0x6734 }, { 3997, 0x7267 },
		{ 3998, 0x7766 }, { 3999, 0x7A46 }, { 4000, 0x91E6 }, { 4001, 0x52C3 }, { 4002, 0x6CA1 }, { 4003, 0x6B86 },
		{ 4004, 0x5800 }, { 4005, 0x5E4C }, { 4006, 0x5954 }, { 4007, 0x672C }, { 4008, 0x7FFB }, { 4009, 0x51E1 },
		{ 4010, 0x76C6 }, { 4011, 0x6469 }, { 4012, 0x78E8 }, { 4013, 0x9B54 }, { 4014, 0x9EBB }, { 4015, 0x57CB },
		{ 4016, 0x59B9 }, { 4017, 0x6627 }, { 4018, 0x679A }, { 4019, 0x6BCE }, { 4020, 0x54E9 }, { 4021, 0x69D9 },
		{ 4022, 0x5E55 }, { 4023, 0x819C }, { 4024, 0x6795 }, { 4025, 0x9BAA }, { 4026, 0x67FE }, { 4027, 0x9C52 },
		{ 4028, 0x685D }, { 4029, 0x4EA6 }, { 4030, 0x4FE3 }, { 4031, 0x53C8 }, { 4032, 0x62B9 }, { 4033, 0x672B },
		{ 4034, 0x6CAB }, { 4035, 0x8FC4 }, { 4036, 0x4FAD }, { 4037, 0x7E6D }, { 4038, 0x9EBF }, { 4039, 0x4E07 },
		{ 4040, 0x6162 }, { 4041, 0x6E80 }, { 4042, 0x6F2B }, { 4043, 0x8513 }, { 4044, 0x5473 }, { 4045, 0x672A },
		{ 4046, 0x9B45 }, { 4047, 0x5DF3 }, { 4048, 0x7B95 }, { 4049, 0x5CAC }, { 4050, 0x5BC6 }, { 4051, 0x871C },
		{ 4052, 0x6E4A }, { 4053, 0x84D1 }, { 4054, 0x7A14 }, { 4055, 0x8108 }, { 4056, 0x5999 }, { 4057, 0x7C8D },
		{ 4058, 0x6C11 }, { 4059, 0x7720 }, { 4060, 0x52D9 }, { 4061, 0x5922 }, { 4062, 0x7121 }, { 4063, 0x725F },
		{ 4064, 0x77DB }, { 4065, 0x9727 }, { 4066, 0x9D61 }, { 4067, 0x690B }, { 4068, 0x5A7F }, { 4069, 0x5A18 },
		{ 4070, 0x51A5 }, { 4071, 0x540D }, { 4072, 0x547D }, { 4073, 0x660E }, { 4074, 0x76DF }, { 4075, 0x8FF7 },
		{ 4076, 0x9298 }, { 4077, 0x9CF4 }, { 4078, 0x59EA }, { 4079, 0x725D }, { 4080, 0x6EC5 }, { 4081, 0x514D },
		{ 4082, 0x68C9 }, { 4083, 0x7DBF }, { 4084, 0x7DEC }, { 4085, 0x9762 }, { 4086, 0x9EBA }, { 4087, 0x6478 },
		{ 4088, 0x6A21 }, { 4089, 0x8302 }, { 4090, 0x5984 }, { 4091, 0x5B5F }, { 4092, 0x6BDB }, { 4093, 0x731B },
		{ 4094, 0x76F2 }, { 4095, 0x7DB2 }, { 4096, 0x8017 }, { 4097, 0x8499 }, { 4098, 0x5132 }, { 4099, 0x6728 },
		{ 4100, 0x9ED9 }, { 4101, 0x76EE }, { 4102, 0x6762 }, { 4103, 0x52FF }, { 4104, 0x9905 }, { 4105, 0x5C24 },
		{ 4106, 0x623B }, { 4107, 0x7C7E }, { 4108, 0x8CB0 }, { 4109, 0x554F }, { 4110, 0x60B6 }, { 4111, 0x7D0B },
		{ 4112, 0x9580 }, { 4113, 0x5301 }, { 4114, 0x4E5F }, { 4115, 0x51B6 }, { 4116, 0x591C }, { 4117, 0x723A },
		{ 4118, 0x8036 }, { 4119, 0x91CE }, { 4120, 0x5F25 }, { 4121, 0x77E2 }, { 4122, 0x5384 }, { 4123, 0x5F79 },
		{ 4124, 0x7D04 }, { 4125, 0x85AC }, { 4126, 0x8A33 }, { 4127, 0x8E8D }, { 4128, 0x9756 }, { 4129, 0x67F3 },
		{ 4130, 0x85AE }, { 4131, 0x9453 }, { 4132, 0x6109 }, { 4133, 0x6108 }, { 4134, 0x6CB9 }, { 4135, 0x7652 },
		{ 4136, 0x8AED }, { 4137, 0x8F38 }, { 4138, 0x552F }, { 4139, 0x4F51 }, { 4140, 0x512A }, { 4141, 0x52C7 },
		{ 4142, 0x53CB }, { 4143, 0x5BA5 }, { 4144, 0x5E7D }, { 4145, 0x60A0 }, { 4146, 0x6182 }, { 4147, 0x63D6 },
		{ 4148, 0x6709 }, { 4149, 0x67DA }, { 4150, 0x6E67 }, { 4151, 0x6D8C }, { 4152, 0x7336 }, { 4153, 0x7337 },
		{ 4154, 0x7531 }, { 4155, 0x7950 }, { 4156, 0x88D5 }, { 4157, 0x8A98 }, { 4158, 0x904A }, { 4159, 0x9091 },
		{ 4160, 0x90F5 }, { 4161, 0x96C4 }, { 4162, 0x878D }, { 4163, 0x5915 }, { 4164, 0x4E88 }, { 4165, 0x4F59 },
		{ 4166, 0x4E0E }, { 4167, 0x8A89 }, { 4168, 0x8F3F }, { 4169, 0x9810 }, { 4170, 0x50AD }, { 4171, 0x5E7C },
		{ 4172, 0x5996 }, { 4173, 0x5BB9 }, { 4174, 0x5EB8 }, { 4175, 0x63DA }, { 4176, 0x63FA }, { 4177, 0x64C1 },
		{ 4178, 0x66DC }, { 4179, 0x694A }, { 4180, 0x69D8 }, { 4181, 0x6D0B }, { 4182, 0x6EB6 }, { 4183, 0x7194 },
		{ 4184, 0x7528 }, { 4185, 0x7AAF }, { 4186, 0x7F8A }, { 4187, 0x8000 }, { 4188, 0x8449 }, { 4189, 0x84C9 },
		{ 4190, 0x8981 }, { 4191, 0x8B21 }, { 4192, 0x8E0A }, { 4193, 0x9065 }, { 4194, 0x967D }, { 4195, 0x990A },
		{ 4196, 0x617E }, { 4197, 0x6291 }, { 4198, 0x6B32 }, { 4199, 0x6C83 }, { 4200, 0x6D74 }, { 4201, 0x7FCC },
		{ 4202, 0x7FFC }, { 4203, 0x6DC0 }, { 4204, 0x7F85 }, { 4205, 0x87BA }, { 4206, 0x88F8 }, { 4207, 0x6765 },
		{ 4208, 0x83B1 }, { 4209, 0x983C }, { 4210, 0x96F7 }, { 4211, 0x6D1B }, { 4212, 0x7D61 }, { 4213, 0x843D },
		{ 4214, 0x916A }, { 4215, 0x4E71 }, { 4216, 0x5375 }, { 4217, 0x5D50 }, { 4218, 0x6B04 }, { 4219, 0x6FEB },
		{ 4220, 0x85CD }, { 4221, 0x862D }, { 4222, 0x89A7 }, { 4223, 0x5229 }, { 4224, 0x540F }, { 4225, 0x5C65 },
		{ 4226, 0x674E }, { 4227, 0x68A8 }, { 4228, 0x7406 }, { 4229, 0x7483 }, { 4230, 0x75E2 }, { 4231, 0x88CF },
		{ 4232, 0x88E1 }, { 4233, 0x91CC }, { 4234, 0x96E2 }, { 4235, 0x9678 }, { 4236, 0x5F8B }, { 4237, 0x7387 },
		{ 4238, 0x7ACB }, { 4239, 0x844E }, { 4240, 0x63A0 }, { 4241, 0x7565 }, { 4242, 0x5289 }, { 4243, 0x6D41 },
		{ 4244, 0x6E9C }, { 4245, 0x7409 }, { 4246, 0x7559 }, { 4247, 0x786B }, { 4248, 0x7C92 }, { 4249, 0x9686 },
		{ 4250, 0x7ADC }, { 4251, 0x9F8D }, { 4252, 0x4FB6 }, { 4253, 0x616E }, { 4254, 0x65C5 }, { 4255, 0x865C },
		{ 4256, 0x4E86 }, { 4257, 0x4EAE }, { 4258, 0x50DA }, { 4259, 0x4E21 }, { 4260, 0x51CC }, { 4261, 0x5BEE },
		{ 4262, 0x6599 }, { 4263, 0x6881 }, { 4264, 0x6DBC }, { 4265, 0x731F }, { 4266, 0x7642 }, { 4267, 0x77AD },
		{ 4268, 0x7A1C }, { 4269, 0x7CE7 }, { 4270, 0x826F }, { 4271, 0x8AD2 }, { 4272, 0x907C }, { 4273, 0x91CF },
		{ 4274, 0x9675 }, { 4275, 0x9818 }, { 4276, 0x529B }, { 4277, 0x7DD1 }, { 4278, 0x502B }, { 4279, 0x5398 },
		{ 4280, 0x6797 }, { 4281, 0x6DCB }, { 4282, 0x71D0 }, { 4283, 0x7433 }, { 4284, 0x81E8 }, { 4285, 0x8F2A },
		{ 4286, 0x96A3 }, { 4287, 0x9C57 }, { 4288, 0x9E9F }, { 4289, 0x7460 }, { 4290, 0x5841 }, { 4291, 0x6D99 },
		{ 4292, 0x7D2F }, { 4293, 0x985E }, { 4294, 0x4EE4 }, { 4295, 0x4F36 }, { 4296, 0x4F8B }, { 4297, 0x51B7 },
		{ 4298, 0x52B1 }, { 4299, 0x5DBA }, { 4300, 0x601C }, { 4301, 0x73B2 }, { 4302, 0x793C }, { 4303, 0x82D3 },
		{ 4304, 0x9234 }, { 4305, 0x96B7 }, { 4306, 0x96F6 }, { 4307, 0x970A }, { 4308, 0x9E97 }, { 4309, 0x9F62 },
		{ 4310, 0x66A6 }, { 4311, 0x6B74 }, { 4312, 0x5217 }, { 4313, 0x52A3 }, { 4314, 0x70C8 }, { 4315, 0x88C2 },
		{ 4316, 0x5EC9 }, { 4317, 0x604B }, { 4318, 0x6190 }, { 4319, 0x6F23 }, { 4320, 0x7149 }, { 4321, 0x7C3E },
		{ 4322, 0x7DF4 }, { 4323, 0x806F }, { 4324, 0x84EE }, { 4325, 0x9023 }, { 4326, 0x932C }, { 4327, 0x5442 },
		{ 4328, 0x9B6F }, { 4329, 0x6AD3 }, { 4330, 0x7089 }, { 4331, 0x8CC2 }, { 4332, 0x8DEF }, { 4333, 0x9732 },
		{ 4334, 0x52B4 }, { 4335, 0x5A41 }, { 4336, 0x5ECA }, { 4337, 0x5F04 }, { 4338, 0x6717 }, { 4339, 0x697C },
		{ 4340, 0x6994 }, { 4341, 0x6D6A }, { 4342, 0x6F0F }, { 4343, 0x7262 }, { 4344, 0x72FC }, { 4345, 0x7BED },
		{ 4346, 0x8001 }, { 4347, 0x807E }, { 4348, 0x874B }, { 4349, 0x90CE }, { 4350, 0x516D }, { 4351, 0x9E93 },
		{ 4352, 0x7984 }, { 4353, 0x808B }, { 4354, 0x9332 }, { 4355, 0x8AD6 }, { 4356, 0x502D }, { 4357, 0x548C },
		{ 4358, 0x8A71 }, { 4359, 0x6B6A }, { 4360, 0x8CC4 }, { 4361, 0x8107 }, { 4362, 0x60D1 }, { 4363, 0x67A0 },
		{ 4364, 0x9DF2 }, { 4365, 0x4E99 }, { 4366, 0x4E98 }, { 4367, 0x9C10 }, { 4368, 0x8A6B }, { 4369, 0x85C1 },
		{ 4370, 0x8568 }, { 4371, 0x6900 }, { 4372, 0x6E7E }, { 4373, 0x7897 }, { 4374, 0x8155 }, { 4418, 0x5F0C },
		{ 4419, 0x4E10 }, { 4420, 0x4E15 }, { 4421, 0x4E2A }, { 4422, 0x4E31 }, { 4423, 0x4E36 }, { 4424, 0x4E3C },
		{ 4425, 0x4E3F }, { 4426, 0x4E42 }, { 4427, 0x4E56 }, { 4428, 0x4E58 }, { 4429, 0x4E82 }, { 4430, 0x4E85 },
		{ 4431, 0x8C6B }, { 4432, 0x4E8A }, { 4433, 0x8212 }, { 4434, 0x5F0D }, { 4435, 0x4E8E }, { 4436, 0x4E9E },
		{ 4437, 0x4E9F }, { 4438, 0x4EA0 }, { 4439, 0x4EA2 }, { 4440, 0x4EB0 }, { 4441, 0x4EB3 }, { 4442, 0x4EB6 },
		{ 4443, 0x4ECE }, { 4444, 0x4ECD }, { 4445, 0x4EC4 }, { 4446, 0x4EC6 }, { 4447, 0x4EC2 }, { 4448, 0x4ED7 },
		{ 4449, 0x4EDE }, { 4450, 0x4EED }, { 4451, 0x4EDF }, { 4452, 0x4EF7 }, { 4453, 0x4F09 }, { 4454, 0x4F5A },
		{ 4455, 0x4F30 }, { 4456, 0x4F5B }, { 4457, 0x4F5D }, { 4458, 0x4F57 }, { 4459, 0x4F47 }, { 4460, 0x4F76 },
		{ 4461, 0x4F88 }, { 4462, 0x4F8F }, { 4463, 0x4F98 }, { 4464, 0x4F7B }, { 4465, 0x4F69 }, { 4466, 0x4F70 },
		{ 4467, 0x4F91 }, { 4468, 0x4F6F }, { 4469, 0x4F86 }, { 4470, 0x4F96 }, { 4471, 0x5118 }, { 4472, 0x4FD4 },
		{ 4473, 0x4FDF }, { 4474, 0x4FCE }, { 4475, 0x4FD8 }, { 4476, 0x4FDB }, { 4477, 0x4FD1 }, { 4478, 0x4FDA },
		{ 4479, 0x4FD0 }, { 4480, 0x4FE4 }, { 4481, 0x4FE5 }, { 4482, 0x501A }, { 4483, 0x5028 }, { 4484, 0x5014 },
		{ 4485, 0x502A }, { 4486, 0x5025 }, { 4487, 0x5005 }, { 4488, 0x4F1C }, { 4489, 0x4FF6 }, { 4490, 0x5021 },
		{ 4491, 0x5029 }, { 4492, 0x502C }, { 4493, 0x4FFE }, { 4494, 0x4FEF }, { 4495, 0x5011 }, { 4496, 0x5006 },
		{ 4497, 0x5043 }, { 4498, 0x5047 }, { 4499, 0x6703 }, { 4500, 0x5055 }, { 4501, 0x5050 }, { 4502, 0x5048 },
		{ 4503, 0x505A }, { 4504, 0x5056 }, { 4505, 0x506C }, { 4506, 0x5078 }, { 4507, 0x5080 }, { 4508, 0x509A },
		{ 4509, 0x5085 }, { 4510, 0x50B4 }, { 4511, 0x50B2 }, { 4512, 0x50C9 }, { 4513, 0x50CA }, { 4514, 0x50B3 },
		{ 4515, 0x50C2 }, { 4516, 0x50D6 }, { 4517, 0x50DE }, { 4518, 0x50E5 }, { 4519, 0x50ED }, { 4520, 0x50E3 },
		{ 4521, 0x50EE }, { 4522, 0x50F9 }, { 4523, 0x50F5 }, { 4524, 0x5109 }, { 4525, 0x5101 }, { 4526, 0x5102 },
		{ 4527, 0x5116 }, { 4528, 0x5115 }, { 4529, 0x5114 }, { 4530, 0x511A }, { 4531, 0x5121 }, { 4532, 0x513A },
		{ 4533, 0x5137 }, { 4534, 0x513C }, { 4535, 0x513B }, { 4536, 0x513F }, { 4537, 0x5140 }, { 4538, 0x5152 },
		{ 4539, 0x514C }, { 4540, 0x5154 }, { 4541, 0x5162 }, { 4542, 0x7AF8 }, { 4543, 0x5169 }, { 4544, 0x516A },
		{ 4545, 0x516E }, { 4546, 0x5180 }, { 4547, 0x5182 }, { 4548, 0x56D8 }, { 4549, 0x518C }, { 4550, 0x5189 },
		{ 4551, 0x518F }, { 4552, 0x5191 }, { 4553, 0x5193 }, { 4554, 0x5195 }, { 4555, 0x5196 }, { 4556, 0x51A4 },
		{ 4557, 0x51A6 }, { 4558, 0x51A2 }, { 4559, 0x51A9 }, { 4560, 0x51AA }, { 4561, 0x51AB }, { 4562, 0x51B3 },
		{ 4563, 0x51B1 }, { 4564, 0x51B2 }, { 4565, 0x51B0 }, { 4566, 0x51B5 }, { 4567, 0x51BD }, { 4568, 0x51C5 },
		{ 4569, 0x51C9 }, { 4570, 0x51DB }, { 4571, 0x51E0 }, { 4572, 0x8655 }, { 4573, 0x51E9 }, { 4574, 0x51ED },
		{ 4575, 0x51F0 }, { 4576, 0x51F5 }, { 4577, 0x51FE }, { 4578, 0x5204 }, { 4579, 0x520B }, { 4580, 0x5214 },
		{ 4581, 0x520E }, { 4582, 0x5227 }, { 4583, 0x522A }, { 4584, 0x522E }, { 4585, 0x5233 }, { 4586, 0x5239 },
		{ 4587, 0x524F }, { 4588, 0x5244 }, { 4589, 0x524B }, { 4590, 0x524C }, { 4591, 0x525E }, { 4592, 0x5254 },
		{ 4593, 0x526A }, { 4594, 0x5274 }, { 4595, 0x5269 }, { 4596, 0x5273 }, { 4597, 0x527F }, { 4598, 0x527D },
		{ 4599, 0x528D }, { 4600, 0x5294 }, { 4601, 0x5292 }, { 4602, 0x5271 }, { 4603, 0x5288 }, { 4604, 0x5291 },
		{ 4605, 0x8FA8 }, { 4606, 0x8FA7 }, { 4607, 0x52AC }, { 4608, 0x52AD }, { 4609, 0x52BC }, { 4610, 0x52B5 },
		{ 4611, 0x52C1 }, { 4612, 0x52CD }, { 4613, 0x52D7 }, { 4614, 0x52DE }, { 4615, 0x52E3 }, { 4616, 0x52E6 },
		{ 4617, 0x98ED }, { 4618, 0x52E0 }, { 4619, 0x52F3 }, { 4620, 0x52F5 }, { 4621, 0x52F8 }, { 4622, 0x52F9 },
		{ 4623, 0x5306 }, { 4624, 0x5308 }, { 4625, 0x7538 }, { 4626, 0x530D }, { 4627, 0x5310 }, { 4628, 0x530F },
		{ 4629, 0x5315 }, { 4630, 0x531A }, { 4631, 0x5323 }, { 4632, 0x532F }, { 4633, 0x5331 }, { 4634, 0x5333 },
		{ 4635, 0x5338 }, { 4636, 0x5340 }, { 4637, 0x5346 }, { 4638, 0x5345 }, { 4639, 0x4E17 }, { 4640, 0x5349 },
		{ 4641, 0x534D }, { 4642, 0x51D6 }, { 4643, 0x535E }, { 4644, 0x5369 }, { 4645, 0x536E }, { 4646, 0x5918 },
		{ 4647, 0x537B }, { 4648, 0x5377 }, { 4649, 0x5382 }, { 4650, 0x5396 }, { 4651, 0x53A0 }, { 4652, 0x53A6 },
		{ 4653, 0x53A5 }, { 4654, 0x53AE }, { 4655, 0x53B0 }, { 4656, 0x53B6 }, { 4657, 0x53C3 }, { 4658, 0x7C12 },
		{ 4659, 0x96D9 }, { 4660, 0x53DF }, { 4661, 0x66FC }, { 4662, 0x71EE }, { 4663, 0x53EE }, { 4664, 0x53E8 },
		{ 4665, 0x53ED }, { 4666, 0x53FA }, { 4667, 0x5401 }, { 4668, 0x543D }, { 4669, 0x5440 }, { 4670, 0x542C },
		{ 4671, 0x542D }, { 4672, 0x543C }, { 4673, 0x542E }, { 4674, 0x5436 }, { 4675, 0x5429 }, { 4676, 0x541D },
		{ 4677, 0x544E }, { 4678, 0x548F }, { 4679, 0x5475 }, { 4680, 0x548E }, { 4681, 0x545F }, { 4682, 0x5471 },
		{ 4683, 0x5477 }, { 4684, 0x5470 }, { 4685, 0x5492 }, { 4686, 0x547B }, { 4687, 0x5480 }, { 4688, 0x5476 },
		{ 4689, 0x5484 }, { 4690, 0x5490 }, { 4691, 0x5486 }, { 4692, 0x54C7 }, { 4693, 0x54A2 }, { 4694, 0x54B8 },
		{ 4695, 0x54A5 }, { 4696, 0x54AC }, { 4697, 0x54C4 }, { 4698, 0x54C8 }, { 4699, 0x54A8 }, { 4700, 0x54AB },
		{ 4701, 0x54C2 }, { 4702, 0x54A4 }, { 4703, 0x54BE }, { 4704, 0x54BC }, { 4705, 0x54D8 }, { 4706, 0x54E5 },
		{ 4707, 0x54E6 }, { 4708, 0x550F }, { 4709, 0x5514 }, { 4710, 0x54FD }, { 4711, 0x54EE }, { 4712, 0x54ED },
		{ 4713, 0x54FA }, { 4714, 0x54E2 }, { 4715, 0x5539 }, { 4716, 0x5540 }, { 4717, 0x5563 }, { 4718, 0x554C },
		{ 4719, 0x552E }, { 4720, 0x555C }, { 4721, 0x5545 }, { 4722, 0x5556 }, { 4723, 0x5557 }, { 4724, 0x5538 },
		{ 4725, 0x5533 }, { 4726, 0x555D }, { 4727, 0x5599 }, { 4728, 0x5580 }, { 4729, 0x54AF }, { 4730, 0x558A },
		{ 4731, 0x559F }, { 4732, 0x557B }, { 4733, 0x557E }, { 4734, 0x5598 }, { 4735, 0x559E }, { 4736, 0x55AE },
		{ 4737, 0x557C }, { 4738, 0x5583 }, { 4739, 0x55A9 }, { 4740, 0x5587 }, { 4741, 0x55A8 }, { 4742, 0x55DA },
		{ 4743, 0x55C5 }, { 4744, 0x55DF }, { 4745, 0x55C4 }, { 4746, 0x55DC }, { 4747, 0x55E4 }, { 4748, 0x55D4 },
		{ 4749, 0x5614 }, { 4750, 0x55F7 }, { 4751, 0x5616 }, { 4752, 0x55FE }, { 4753, 0x55FD }, { 4754, 0x561B },
		{ 4755, 0x55F9 }, { 4756, 0x564E }, { 4757, 0x5650 }, { 4758, 0x71DF }, { 4759, 0x5634 }, { 4760, 0x5636 },
		{ 4761, 0x5632 }, { 4762, 0x5638 }, { 4763, 0x566B }, { 4764, 0x5664 }, { 4765, 0x562F }, { 4766, 0x566C },
		{ 4767, 0x566A }, { 4768, 0x5686 }, { 4769, 0x5680 }, { 4770, 0x568A }, { 4771, 0x56A0 }, { 4772, 0x5694 },
		{ 4773, 0x568F }, { 4774, 0x56A5 }, { 4775, 0x56AE }, { 4776, 0x56B6 }, { 4777, 0x56B4 }, { 4778, 0x56C2 },
		{ 4779, 0x56BC }, { 4780, 0x56C1 }, { 4781, 0x56C3 }, { 4782, 0x56C0 }, { 4783, 0x56C8 }, { 4784, 0x56CE },
		{ 4785, 0x56D1 }, { 4786, 0x56D3 }, { 4787, 0x56D7 }, { 4788, 0x56EE }, { 4789, 0x56F9 }, { 4790, 0x5700 },
		{ 4791, 0x56FF }, { 4792, 0x5704 }, { 4793, 0x5709 }, { 4794, 0x5708 }, { 4795, 0x570B }, { 4796, 0x570D },
		{ 4797, 0x5713 }, { 4798, 0x5718 }, { 4799, 0x5716 }, { 4800, 0x55C7 }, { 4801, 0x571C }, { 4802, 0x5726 },
		{ 4803, 0x5737 }, { 4804, 0x5738 }, { 4805, 0x574E }, { 4806, 0x573B }, { 4807, 0x5740 }, { 4808, 0x574F },
		{ 4809, 0x5769 }, { 4810, 0x57C0 }, { 4811, 0x5788 }, { 4812, 0x5761 }, { 4813, 0x577F }, { 4814, 0x5789 },
		{ 4815, 0x5793 }, { 4816, 0x57A0 }, { 4817, 0x57B3 }, { 4818, 0x57A4 }, { 4819, 0x57AA }, { 4820, 0x57B0 },
		{ 4821, 0x57C3 }, { 4822, 0x57C6 }, { 4823, 0x57D4 }, { 4824, 0x57D2 }, { 4825, 0x57D3 }, { 4826, 0x580A },
		{ 4827, 0x57D6 }, { 4828, 0x57E3 }, { 4829, 0x580B }, { 4830, 0x5819 }, { 4831, 0x581D }, { 4832, 0x5872 },
		{ 4833, 0x5821 }, { 4834, 0x5862 }, { 4835, 0x584B }, { 4836, 0x5870 }, { 4837, 0x6BC0 }, { 4838, 0x5852 },
		{ 4839, 0x583D }, { 4840, 0x5879 }, { 4841, 0x5885 }, { 4842, 0x58B9 }, { 4843, 0x589F }, { 4844, 0x58AB },
		{ 4845, 0x58BA }, { 4846, 0x58DE }, { 4847, 0x58BB }, { 4848, 0x58B8 }, { 4849, 0x58AE }, { 4850, 0x58C5 },
		{ 4851, 0x58D3 }, { 4852, 0x58D1 }, { 4853, 0x58D7 }, { 4854, 0x58D9 }, { 4855, 0x58D8 }, { 4856, 0x58E5 },
		{ 4857, 0x58DC }, { 4858, 0x58E4 }, { 4859, 0x58DF }, { 4860, 0x58EF }, { 4861, 0x58FA }, { 4862, 0x58F9 },
		{ 4863, 0x58FB }, { 4864, 0x58FC }, { 4865, 0x58FD }, { 4866, 0x5902 }, { 4867, 0x590A }, { 4868, 0x5910 },
		{ 4869, 0x591B }, { 4870, 0x68A6 }, { 4871, 0x5925 }, { 4872, 0x592C }, { 4873, 0x592D }, { 4874, 0x5932 },
		{ 4875, 0x5938 }, { 4876, 0x593E }, { 4877, 0x7AD2 }, { 4878, 0x5955 }, { 4879, 0x5950 }, { 4880, 0x594E },
		{ 4881, 0x595A }, { 4882, 0x5958 }, { 4883, 0x5962 }, { 4884, 0x5960 }, { 4885, 0x5967 }, { 4886, 0x596C },
		{ 4887, 0x5969 }, { 4888, 0x5978 }, { 4889, 0x5981 }, { 4890, 0x599D }, { 4891, 0x4F5E }, { 4892, 0x4FAB },
		{ 4893, 0x59A3 }, { 4894, 0x59B2 }, { 4895, 0x59C6 }, { 4896, 0x59E8 }, { 4897, 0x59DC }, { 4898, 0x598D },
		{ 4899, 0x59D9 }, { 4900, 0x59DA }, { 4901, 0x5A25 }, { 4902, 0x5A1F }, { 4903, 0x5A11 }, { 4904, 0x5A1C },
		{ 4905, 0x5A09 }, { 4906, 0x5A1A }, { 4907, 0x5A40 }, { 4908, 0x5A6C }, { 4909, 0x5A49 }, { 4910, 0x5A35 },
		{ 4911, 0x5A36 }, { 4912, 0x5A62 }, { 4913, 0x5A6A }, { 4914, 0x5A9A }, { 4915, 0x5ABC }, { 4916, 0x5ABE },
		{ 4917, 0x5ACB }, { 4918, 0x5AC2 }, { 4919, 0x5ABD }, { 4920, 0x5AE3 }, { 4921, 0x5AD7 }, { 4922, 0x5AE6 },
		{ 4923, 0x5AE9 }, { 4924, 0x5AD6 }, { 4925, 0x5AFA }, { 4926, 0x5AFB }, { 4927, 0x5B0C }, { 4928, 0x5B0B },
		{ 4929, 0x5B16 }, { 4930, 0x5B32 }, { 4931, 0x5AD0 }, { 4932, 0x5B2A }, { 4933, 0x5B36 }, { 4934, 0x5B3E },
		{ 4935, 0x5B43 }, { 4936, 0x5B45 }, { 4937, 0x5B40 }, { 4938, 0x5B51 }, { 4939, 0x5B55 }, { 4940, 0x5B5A },
		{ 4941, 0x5B5B }, { 4942, 0x5B65 }, { 4943, 0x5B69 }, { 4944, 0x5B70 }, { 4945, 0x5B73 }, { 4946, 0x5B75 },
		{ 4947, 0x5B78 }, { 4948, 0x6588 }, { 4949, 0x5B7A }, { 4950, 0x5B80 }, { 4951, 0x5B83 }, { 4952, 0x5BA6 },
		{ 4953, 0x5BB8 }, { 4954, 0x5BC3 }, { 4955, 0x5BC7 }, { 4956, 0x5BC9 }, { 4957, 0x5BD4 }, { 4958, 0x5BD0 },
		{ 4959, 0x5BE4 }, { 4960, 0x5BE6 }, { 4961, 0x5BE2 }, { 4962, 0x5BDE }, { 4963, 0x5BE5 }, { 4964, 0x5BEB },
		{ 4965, 0x5BF0 }, { 4966, 0x5BF6 }, { 4967, 0x5BF3 }, { 4968, 0x5C05 }, { 4969, 0x5C07 }, { 4970, 0x5C08 },
		{ 4971, 0x5C0D }, { 4972, 0x5C13 }, { 4973, 0x5C20 }, { 4974, 0x5C22 }, { 4975, 0x5C28 }, { 4976, 0x5C38 },
		{ 4977, 0x5C39 }, { 4978, 0x5C41 }, { 4979, 0x5C46 }, { 4980, 0x5C4E }, { 4981, 0x5C53 }, { 4982, 0x5C50 },
		{ 4983, 0x5C4F }, { 4984, 0x5B71 }, { 4985, 0x5C6C }, { 4986, 0x5C6E }, { 4987, 0x4E62 }, { 4988, 0x5C76 },
		{ 4989, 0x5C79 }, { 4990, 0x5C8C }, { 4991, 0x5C91 }, { 4992, 0x5C94 }, { 4993, 0x599B }, { 4994, 0x5CAB },
		{ 4995, 0x5CBB }, { 4996, 0x5CB6 }, { 4997, 0x5CBC }, { 4998, 0x5CB7 }, { 4999, 0x5CC5 }, { 5000, 0x5CBE },
		{ 5001, 0x5CC7 }, { 5002, 0x5CD9 }, { 5003, 0x5CE9 }, { 5004, 0x5CFD }, { 5005, 0x5CFA }, { 5006, 0x5CED },
		{ 5007, 0x5D8C }, { 5008, 0x5CEA }, { 5009, 0x5D0B }, { 5010, 0x5D15 }, { 5011, 0x5D17 }, { 5012, 0x5D5C },
		{ 5013, 0x5D1F }, { 5014, 0x5D1B }, { 5015, 0x5D11 }, { 5016, 0x5D14 }, { 5017, 0x5D22 }, { 5018, 0x5D1A },
		{ 5019, 0x5D19 }, { 5020, 0x5D18 }, { 5021, 0x5D4C }, { 5022, 0x5D52 }, { 5023, 0x5D4E }, { 5024, 0x5D4B },
		{ 5025, 0x5D6C }, { 5026, 0x5D73 }, { 5027, 0x5D76 }, { 5028, 0x5D87 }, { 5029, 0x5D84 }, { 5030, 0x5D82 },
		{ 5031, 0x5DA2 }, { 5032, 0x5D9D }, { 5033, 0x5DAC }, { 5034, 0x5DAE }, { 5035, 0x5DBD }, { 5036, 0x5D90 },
		{ 5037, 0x5DB7 }, { 5038, 0x5DBC }, { 5039, 0x5DC9 }, { 5040, 0x5DCD }, { 5041, 0x5DD3 }, { 5042, 0x5DD2 },
		{ 5043, 0x5DD6 }, { 5044, 0x5DDB }, { 5045, 0x5DEB }, { 5046, 0x5DF2 }, { 5047, 0x5DF5 }, { 5048, 0x5E0B },
		{ 5049, 0x5E1A }, { 5050, 0x5E19 }, { 5051, 0x5E11 }, { 5052, 0x5E1B }, { 5053, 0x5E36 }, { 5054, 0x5E37 },
		{ 5055, 0x5E44 }, { 5056, 0x5E43 }, { 5057, 0x5E40 }, { 5058, 0x5E4E }, { 5059, 0x5E57 }, { 5060, 0x5E54 },
		{ 5061, 0x5E5F }, { 5062, 0x5E62 }, { 5063, 0x5E64 }, { 5064, 0x5E47 }, { 5065, 0x5E75 }, { 5066, 0x5E76 },
		{ 5067, 0x5E7A }, { 5068, 0x9EBC }, { 5069, 0x5E7F }, { 5070, 0x5EA0 }, { 5071, 0x5EC1 }, { 5072, 0x5EC2 },
		{ 5073, 0x5EC8 }, { 5074, 0x5ED0 }, { 5075, 0x5ECF }, { 5076, 0x5ED6 }, { 5077, 0x5EE3 }, { 5078, 0x5EDD },
		{ 5079, 0x5EDA }, { 5080, 0x5EDB }, { 5081, 0x5EE2 }, { 5082, 0x5EE1 }, { 5083, 0x5EE8 }, { 5084, 0x5EE9 },
		{ 5085, 0x5EEC }, { 5086, 0x5EF1 }, { 5087, 0x5EF3 }, { 5088, 0x5EF0 }, { 5089, 0x5EF4 }, { 5090, 0x5EF8 },
		{ 5091, 0x5EFE }, { 5092, 0x5F03 }, { 5093, 0x5F09 }, { 5094, 0x5F5D }, { 5095, 0x5F5C }, { 5096, 0x5F0B },
		{ 5097, 0x5F11 }, { 5098, 0x5F16 }, { 5099, 0x5F29 }, { 5100, 0x5F2D }, { 5101, 0x5F38 }, { 5102, 0x5F41 },
		{ 5103, 0x5F48 }, { 5104, 0x5F4C }, { 5105, 0x5F4E }, { 5106, 0x5F2F }, { 5107, 0x5F51 }, { 5108, 0x5F56 },
		{ 5109, 0x5F57 }, { 5110, 0x5F59 }, { 5111, 0x5F61 }, { 5112, 0x5F6D }, { 5113, 0x5F73 }, { 5114, 0x5F77 },
		{ 5115, 0x5F83 }, { 5116, 0x5F82 }, { 5117, 0x5F7F }, { 5118, 0x5F8A }, { 5119, 0x5F88 }, { 5120, 0x5F91 },
		{ 5121, 0x5F87 }, { 5122, 0x5F9E }, { 5123, 0x5F99 }, { 5124, 0x5F98 }, { 5125, 0x5FA0 }, { 5126, 0x5FA8 },
		{ 5127, 0x5FAD }, { 5128, 0x5FBC }, { 5129, 0x5FD6 }, { 5130, 0x5FFB }, { 5131, 0x5FE4 }, { 5132, 0x5FF8 },
		{ 5133, 0x5FF1 }, { 5134, 0x5FDD }, { 5135, 0x60B3 }, { 5136, 0x5FFF }, { 5137, 0x6021 }, { 5138, 0x6060 },
		{ 5139, 0x6019 }, { 5140, 0x6010 }, { 5141, 0x6029 }, { 5142, 0x600E }, { 5143, 0x6031 }, { 5144, 0x601B },
		{ 5145, 0x6015 }, { 5146, 0x602B }, { 5147, 0x6026 }, { 5148, 0x600F }, { 5149, 0x603A }, { 5150, 0x605A },
		{ 5151, 0x6041 }, { 5152, 0x606A }, { 5153, 0x6077 }, { 5154, 0x605F }, { 5155, 0x604A }, { 5156, 0x6046 },
		{ 5157, 0x604D }, { 5158, 0x6063 }, { 5159, 0x6043 }, { 5160, 0x6064 }, { 5161, 0x6042 }, { 5162, 0x606C },
		{ 5163, 0x606B }, { 5164, 0x6059 }, { 5165, 0x6081 }, { 5166, 0x608D }, { 5167, 0x60E7 }, { 5168, 0x6083 },
		{ 5169, 0x609A }, { 5170, 0x6084 }, { 5171, 0x609B }, { 5172, 0x6096 }, { 5173, 0x6097 }, { 5174, 0x6092 },
		{ 5175, 0x60A7 }, { 5176, 0x608B }, { 5177, 0x60E1 }, { 5178, 0x60B8 }, { 5179, 0x60E0 }, { 5180, 0x60D3 },
		{ 5181, 0x60B4 }, { 5182, 0x5FF0 }, { 5183, 0x60BD }, { 5184, 0x60C6 }, { 5185, 0x60B5 }, { 5186, 0x60D8 },
		{ 5187, 0x614D }, { 5188, 0x6115 }, { 5189, 0x6106 }, { 5190, 0x60F6 }, { 5191, 0x60F7 }, { 5192, 0x6100 },
		{ 5193, 0x60F4 }, { 5194, 0x60FA }, { 5195, 0x6103 }, { 5196, 0x6121 }, { 5197, 0x60FB }, { 5198, 0x60F1 },
		{ 5199, 0x610D }, { 5200, 0x610E }, { 5201, 0x6147 }, { 5202, 0x613E }, { 5203, 0x6128 }, { 5204, 0x6127 },
		{ 5205, 0x614A }, { 5206, 0x613F }, { 5207, 0x613C }, { 5208, 0x612C }, { 5209, 0x6134 }, { 5210, 0x613D },
		{ 5211, 0x6142 }, { 5212, 0x6144 }, { 5213, 0x6173 }, { 5214, 0x6177 }, { 5215, 0x6158 }, { 5216, 0x6159 },
		{ 5217, 0x615A }, { 5218, 0x616B }, { 5219, 0x6174 }, { 5220, 0x616F }, { 5221, 0x6165 }, { 5222, 0x6171 },
		{ 5223, 0x615F }, { 5224, 0x615D }, { 5225, 0x6153 }, { 5226, 0x6175 }, { 5227, 0x6199 }, { 5228, 0x6196 },
		{ 5229, 0x6187 }, { 5230, 0x61AC }, { 5231, 0x6194 }, { 5232, 0x619A }, { 5233, 0x618A }, { 5234, 0x6191 },
		{ 5235, 0x61AB }, { 5236, 0x61AE }, { 5237, 0x61CC }, { 5238, 0x61CA }, { 5239, 0x61C9 }, { 5240, 0x61F7 },
		{ 5241, 0x61C8 }, { 5242, 0x61C3 }, { 5243, 0x61C6 }, { 5244, 0x61BA }, { 5245, 0x61CB }, { 5246, 0x7F79 },
		{ 5247, 0x61CD }, { 5248, 0x61E6 }, { 5249, 0x61E3 }, { 5250, 0x61F6 }, { 5251, 0x61FA }, { 5252, 0x61F4 },
		{ 5253, 0x61FF }, { 5254, 0x61FD }, { 5255, 0x61FC }, { 5256, 0x61FE }, { 5257, 0x6200 }, { 5258, 0x6208 },
		{ 5259, 0x6209 }, { 5260, 0x620D }, { 5261, 0x620C }, { 5262, 0x6214 }, { 5263, 0x621B }, { 5264, 0x621E },
		{ 5265, 0x6221 }, { 5266, 0x622A }, { 5267, 0x622E }, { 5268, 0x6230 }, { 5269, 0x6232 }, { 5270, 0x6233 },
		{ 5271, 0x6241 }, { 5272, 0x624E }, { 5273, 0x625E }, { 5274, 0x6263 }, { 5275, 0x625B }, { 5276, 0x6260 },
		{ 5277, 0x6268 }, { 5278, 0x627C }, { 5279, 0x6282 }, { 5280, 0x6289 }, { 5281, 0x627E }, { 5282, 0x6292 },
		{ 5283, 0x6293 }, { 5284, 0x6296 }, { 5285, 0x62D4 }, { 5286, 0x6283 }, { 5287, 0x6294 }, { 5288, 0x62D7 },
		{ 5289, 0x62D1 }, { 5290, 0x62BB }, { 5291, 0x62CF }, { 5292, 0x62FF }, { 5293, 0x62C6 }, { 5294, 0x64D4 },
		{ 5295, 0x62C8 }, { 5296, 0x62DC }, { 5297, 0x62CC }, { 5298, 0x62CA }, { 5299, 0x62C2 }, { 5300, 0x62C7 },
		{ 5301, 0x629B }, { 5302, 0x62C9 }, { 5303, 0x630C }, { 5304, 0x62EE }, { 5305, 0x62F1 }, { 5306, 0x6327 },
		{ 5307, 0x6302 }, { 5308, 0x6308 }, { 5309, 0x62EF }, { 5310, 0x62F5 }, { 5311, 0x6350 }, { 5312, 0x633E },
		{ 5313, 0x634D }, { 5314, 0x641C }, { 5315, 0x634F }, { 5316, 0x6396 }, { 5317, 0x638E }, { 5318, 0x6380 },
		{ 5319, 0x63AB }, { 5320, 0x6376 }, { 5321, 0x63A3 }, { 5322, 0x638F }, { 5323, 0x6389 }, { 5324, 0x639F },
		{ 5325, 0x63B5 }, { 5326, 0x636B }, { 5327, 0x6369 }, { 5328, 0x63BE }, { 5329, 0x63E9 }, { 5330, 0x63C0 },
		{ 5331, 0x63C6 }, { 5332, 0x63E3 }, { 5333, 0x63C9 }, { 5334, 0x63D2 }, { 5335, 0x63F6 }, { 5336, 0x63C4 },
		{ 5337, 0x6416 }, { 5338, 0x6434 }, { 5339, 0x6406 }, { 5340, 0x6413 }, { 5341, 0x6426 }, { 5342, 0x6436 },
		{ 5343, 0x651D }, { 5344, 0x6417 }, { 5345, 0x6428 }, { 5346, 0x640F }, { 5347, 0x6467 }, { 5348, 0x646F },
		{ 5349, 0x6476 }, { 5350, 0x644E }, { 5351, 0x652A }, { 5352, 0x6495 }, { 5353, 0x6493 }, { 5354, 0x64A5 },
		{ 5355, 0x64A9 }, { 5356, 0x6488 }, { 5357, 0x64BC }, { 5358, 0x64DA }, { 5359, 0x64D2 }, { 5360, 0x64C5 },
		{ 5361, 0x64C7 }, { 5362, 0x64BB }, { 5363, 0x64D8 }, { 5364, 0x64C2 }, { 5365, 0x64F1 }, { 5366, 0x64E7 },
		{ 5367, 0x8209 }, { 5368, 0x64E0 }, { 5369, 0x64E1 }, { 5370, 0x62AC }, { 5371, 0x64E3 }, { 5372, 0x64EF },
		{ 5373, 0x652C }, { 5374, 0x64F6 }, { 5375, 0x64F4 }, { 5376, 0x64F2 }, { 5377, 0x64FA }, { 5378, 0x6500 },
		{ 5379, 0x64FD }, { 5380, 0x6518 }, { 5381, 0x651C }, { 5382, 0x6505 }, { 5383, 0x6524 }, { 5384, 0x6523 },
		{ 5385, 0x652B }, { 5386, 0x6534 }, { 5387, 0x6535 }, { 5388, 0x6537 }, { 5389, 0x6536 }, { 5390, 0x6538 },
		{ 5391, 0x754B }, { 5392, 0x6548 }, { 5393, 0x6556 }, { 5394, 0x6555 }, { 5395, 0x654D }, { 5396, 0x6558 },
		{ 5397, 0x655E }, { 5398, 0x655D }, { 5399, 0x6572 }, { 5400, 0x6578 }, { 5401, 0x6582 }, { 5402, 0x6583 },
		{ 5403, 0x8B8A }, { 5404, 0x659B }, { 5405, 0x659F }, { 5406, 0x65AB }, { 5407, 0x65B7 }, { 5408, 0x65C3 },
		{ 5409, 0x65C6 }, { 5410, 0x65C1 }, { 5411, 0x65C4 }, { 5412, 0x65CC }, { 5413, 0x65D2 }, { 5414, 0x65DB },
		{ 5415, 0x65D9 }, { 5416, 0x65E0 }, { 5417, 0x65E1 }, { 5418, 0x65F1 }, { 5419, 0x6772 }, { 5420, 0x660A },
		{ 5421, 0x6603 }, { 5422, 0x65FB }, { 5423, 0x6773 }, { 5424, 0x6635 }, { 5425, 0x6636 }, { 5426, 0x6634 },
		{ 5427, 0x661C }, { 5428, 0x664F }, { 5429, 0x6644 }, { 5430, 0x6649 }, { 5431, 0x6641 }, { 5432, 0x665E },
		{ 5433, 0x665D }, { 5434, 0x6664 }, { 5435, 0x6667 }, { 5436, 0x6668 }, { 5437, 0x665F }, { 5438, 0x6662 },
		{ 5439, 0x6670 }, { 5440, 0x6683 }, { 5441, 0x6688 }, { 5442, 0x668E }, { 5443, 0x6689 }, { 5444, 0x6684 },
		{ 5445, 0x6698 }, { 5446, 0x669D }, { 5447, 0x66C1 }, { 5448, 0x66B9 }, { 5449, 0x66C9 }, { 5450, 0x66BE },
		{ 5451, 0x66BC }, { 5452, 0x66C4 }, { 5453, 0x66B8 }, { 5454, 0x66D6 }, { 5455, 0x66DA }, { 5456, 0x66E0 },
		{ 5457, 0x663F }, { 5458, 0x66E6 }, { 5459, 0x66E9 }, { 5460, 0x66F0 }, { 5461, 0x66F5 }, { 5462, 0x66F7 },
		{ 5463, 0x670F }, { 5464, 0x6716 }, { 5465, 0x671E }, { 5466, 0x6726 }, { 5467, 0x6727 }, { 5468, 0x9738 },
		{ 5469, 0x672E }, { 5470, 0x673F }, { 5471, 0x6736 }, { 5472, 0x6741 }, { 5473, 0x6738 }, { 5474, 0x6737 },
		{ 5475, 0x6746 }, { 5476, 0x675E }, { 5477, 0x6760 }, { 5478, 0x6759 }, { 5479, 0x6763 }, { 5480, 0x6764 },
		{ 5481, 0x6789 }, { 5482, 0x6770 }, { 5483, 0x67A9 }, { 5484, 0x677C }, { 5485, 0x676A }, { 5486, 0x678C },
		{ 5487, 0x678B }, { 5488, 0x67A6 }, { 5489, 0x67A1 }, { 5490, 0x6785 }, { 5491, 0x67B7 }, { 5492, 0x67EF },
		{ 5493, 0x67B4 }, { 5494, 0x67EC }, { 5495, 0x67B3 }, { 5496, 0x67E9 }, { 5497, 0x67B8 }, { 5498, 0x67E4 },
		{ 5499, 0x67DE }, { 5500, 0x67DD }, { 5501, 0x67E2 }, { 5502, 0x67EE }, { 5503, 0x67B9 }, { 5504, 0x67CE },
		{ 5505, 0x67C6 }, { 5506, 0x67E7 }, { 5507, 0x6A9C }, { 5508, 0x681E }, { 5509, 0x6846 }, { 5510, 0x6829 },
		{ 5511, 0x6840 }, { 5512, 0x684D }, { 5513, 0x6832 }, { 5514, 0x684E }, { 5515, 0x68B3 }, { 5516, 0x682B },
		{ 5517, 0x6859 }, { 5518, 0x6863 }, { 5519, 0x6877 }, { 5520, 0x687F }, { 5521, 0x689F }, { 5522, 0x688F },
		{ 5523, 0x68AD }, { 5524, 0x6894 }, { 5525, 0x689D }, { 5526, 0x689B }, { 5527, 0x6883 }, { 5528, 0x6AAE },
		{ 5529, 0x68B9 }, { 5530, 0x6874 }, { 5531, 0x68B5 }, { 5532, 0x68A0 }, { 5533, 0x68BA }, { 5534, 0x690F },
		{ 5535, 0x688D }, { 5536, 0x687E }, { 5537, 0x6901 }, { 5538, 0x68CA }, { 5539, 0x6908 }, { 5540, 0x68D8 },
		{ 5541, 0x6922 }, { 5542, 0x6926 }, { 5543, 0x68E1 }, { 5544, 0x690C }, { 5545, 0x68CD }, { 5546, 0x68D4 },
		{ 5547, 0x68E7 }, { 5548, 0x68D5 }, { 5549, 0x6936 }, { 5550, 0x6912 }, { 5551, 0x6904 }, { 5552, 0x68D7 },
		{ 5553, 0x68E3 }, { 5554, 0x6925 }, { 5555, 0x68F9 }, { 5556, 0x68E0 }, { 5557, 0x68EF }, { 5558, 0x6928 },
		{ 5559, 0x692A }, { 5560, 0x691A }, { 5561, 0x6923 }, { 5562, 0x6921 }, { 5563, 0x68C6 }, { 5564, 0x6979 },
		{ 5565, 0x6977 }, { 5566, 0x695C }, { 5567, 0x6978 }, { 5568, 0x696B }, { 5569, 0x6954 }, { 5570, 0x697E },
		{ 5571, 0x696E }, { 5572, 0x6939 }, { 5573, 0x6974 }, { 5574, 0x693D }, { 5575, 0x6959 }, { 5576, 0x6930 },
		{ 5577, 0x6961 }, { 5578, 0x695E }, { 5579, 0x695D }, { 5580, 0x6981 }, { 5581, 0x696A }, { 5582, 0x69B2 },
		{ 5583, 0x69AE }, { 5584, 0x69D0 }, { 5585, 0x69BF }, { 5586, 0x69C1 }, { 5587, 0x69D3 }, { 5588, 0x69BE },
		{ 5589, 0x69CE }, { 5590, 0x5BE8 }, { 5591, 0x69CA }, { 5592, 0x69DD }, { 5593, 0x69BB }, { 5594, 0x69C3 },
		{ 5595, 0x69A7 }, { 5596, 0x6A2E }, { 5597, 0x6991 }, { 5598, 0x69A0 }, { 5599, 0x699C }, { 5600, 0x6995 },
		{ 5601, 0x69B4 }, { 5602, 0x69DE }, { 5603, 0x69E8 }, { 5604, 0x6A02 }, { 5605, 0x6A1B }, { 5606, 0x69FF },
		{ 5607, 0x6B0A }, { 5608, 0x69F9 }, { 5609, 0x69F2 }, { 5610, 0x69E7 }, { 5611, 0x6A05 }, { 5612, 0x69B1 },
		{ 5613, 0x6A1E }, { 5614, 0x69ED }, { 5615, 0x6A14 }, { 5616, 0x69EB }, { 5617, 0x6A0A }, { 5618, 0x6A12 },
		{ 5619, 0x6AC1 }, { 5620, 0x6A23 }, { 5621, 0x6A13 }, { 5622, 0x6A44 }, { 5623, 0x6A0C }, { 5624, 0x6A72 },
		{ 5625, 0x6A36 }, { 5626, 0x6A78 }, { 5627, 0x6A47 }, { 5628, 0x6A62 }, { 5629, 0x6A59 }, { 5630, 0x6A66 },
		{ 5631, 0x6A48 }, { 5632, 0x6A38 }, { 5633, 0x6A22 }, { 5634, 0x6A90 }, { 5635, 0x6A8D }, { 5636, 0x6AA0 },
		{ 5637, 0x6A84 }, { 5638, 0x6AA2 }, { 5639, 0x6AA3 }, { 5640, 0x6A97 }, { 5641, 0x8617 }, { 5642, 0x6ABB },
		{ 5643, 0x6AC3 }, { 5644, 0x6AC2 }, { 5645, 0x6AB8 }, { 5646, 0x6AB3 }, { 5647, 0x6AAC }, { 5648, 0x6ADE },
		{ 5649, 0x6AD1 }, { 5650, 0x6ADF }, { 5651, 0x6AAA }, { 5652, 0x6ADA }, { 5653, 0x6AEA }, { 5654, 0x6AFB },
		{ 5655, 0x6B05 }, { 5656, 0x8616 }, { 5657, 0x6AFA }, { 5658, 0x6B12 }, { 5659, 0x6B16 }, { 5660, 0x9B31 },
		{ 5661, 0x6B1F }, { 5662, 0x6B38 }, { 5663, 0x6B37 }, { 5664, 0x76DC }, { 5665, 0x6B39 }, { 5666, 0x98EE },
		{ 5667, 0x6B47 }, { 5668, 0x6B43 }, { 5669, 0x6B49 }, { 5670, 0x6B50 }, { 5671, 0x6B59 }, { 5672, 0x6B54 },
		{ 5673, 0x6B5B }, { 5674, 0x6B5F }, { 5675, 0x6B61 }, { 5676, 0x6B78 }, { 5677, 0x6B79 }, { 5678, 0x6B7F },
		{ 5679, 0x6B80 }, { 5680, 0x6B84 }, { 5681, 0x6B83 }, { 5682, 0x6B8D }, { 5683, 0x6B98 }, { 5684, 0x6B95 },
		{ 5685, 0x6B9E }, { 5686, 0x6BA4 }, { 5687, 0x6BAA }, { 5688, 0x6BAB }, { 5689, 0x6BAF }, { 5690, 0x6BB2 },
		{ 5691, 0x6BB1 }, { 5692, 0x6BB3 }, { 5693, 0x6BB7 }, { 5694, 0x6BBC }, { 5695, 0x6BC6 }, { 5696, 0x6BCB },
		{ 5697, 0x6BD3 }, { 5698, 0x6BDF }, { 5699, 0x6BEC }, { 5700, 0x6BEB }, { 5701, 0x6BF3 }, { 5702, 0x6BEF },
		{ 5703, 0x9EBE }, { 5704, 0x6C08 }, { 5705, 0x6C13 }, { 5706, 0x6C14 }, { 5707, 0x6C1B }, { 5708, 0x6C24 },
		{ 5709, 0x6C23 }, { 5710, 0x6C5E }, { 5711, 0x6C55 }, { 5712, 0x6C62 }, { 5713, 0x6C6A }, { 5714, 0x6C82 },
		{ 5715, 0x6C8D }, { 5716, 0x6C9A }, { 5717, 0x6C81 }, { 5718, 0x6C9B }, { 5719, 0x6C7E }, { 5720, 0x6C68 },
		{ 5721, 0x6C73 }, { 5722, 0x6C92 }, { 5723, 0x6C90 }, { 5724, 0x6CC4 }, { 5725, 0x6CF1 }, { 5726, 0x6CD3 },
		{ 5727, 0x6CBD }, { 5728, 0x6CD7 }, { 5729, 0x6CC5 }, { 5730, 0x6CDD }, { 5731, 0x6CAE }, { 5732, 0x6CB1 },
		{ 5733, 0x6CBE }, { 5734, 0x6CBA }, { 5735, 0x6CDB }, { 5736, 0x6CEF }, { 5737, 0x6CD9 }, { 5738, 0x6CEA },
		{ 5739, 0x6D1F }, { 5740, 0x884D }, { 5741, 0x6D36 }, { 5742, 0x6D2B }, { 5743, 0x6D3D }, { 5744, 0x6D38 },
		{ 5745, 0x6D19 }, { 5746, 0x6D35 }, { 5747, 0x6D33 }, { 5748, 0x6D12 }, { 5749, 0x6D0C }, { 5750, 0x6D63 },
		{ 5751, 0x6D93 }, { 5752, 0x6D64 }, { 5753, 0x6D5A }, { 5754, 0x6D79 }, { 5755, 0x6D59 }, { 5756, 0x6D8E },
		{ 5757, 0x6D95 }, { 5758, 0x6FE4 }, { 5759, 0x6D85 }, { 5760, 0x6DF9 }, { 5761, 0x6E15 }, { 5762, 0x6E0A },
		{ 5763, 0x6DB5 }, { 5764, 0x6DC7 }, { 5765, 0x6DE6 }, { 5766, 0x6DB8 }, { 5767, 0x6DC6 }, { 5768, 0x6DEC },
		{ 5769, 0x6DDE }, { 5770, 0x6DCC }, { 5771, 0x6DE8 }, { 5772, 0x6DD2 }, { 5773, 0x6DC5 }, { 5774, 0x6DFA },
		{ 5775, 0x6DD9 }, { 5776, 0x6DE4 }, { 5777, 0x6DD5 }, { 5778, 0x6DEA }, { 5779, 0x6DEE }, { 5780, 0x6E2D },
		{ 5781, 0x6E6E }, { 5782, 0x6E2E }, { 5783, 0x6E19 }, { 5784, 0x6E72 }, { 5785, 0x6E5F }, { 5786, 0x6E3E },
		{ 5787, 0x6E23 }, { 5788, 0x6E6B }, { 5789, 0x6E2B }, { 5790, 0x6E76 }, { 5791, 0x6E4D }, { 5792, 0x6E1F },
		{ 5793, 0x6E43 }, { 5794, 0x6E3A }, { 5795, 0x6E4E }, { 5796, 0x6E24 }, { 5797, 0x6EFF }, { 5798, 0x6E1D },
		{ 5799, 0x6E38 }, { 5800, 0x6E82 }, { 5801, 0x6EAA }, { 5802, 0x6E98 }, { 5803, 0x6EC9 }, { 5804, 0x6EB7 },
		{ 5805, 0x6ED3 }, { 5806, 0x6EBD }, { 5807, 0x6EAF }, { 5808, 0x6EC4 }, { 5809, 0x6EB2 }, { 5810, 0x6ED4 },
		{ 5811, 0x6ED5 }, { 5812, 0x6E8F }, { 5813, 0x6EA5 }, { 5814, 0x6EC2 }, { 5815, 0x6E9F }, { 5816, 0x6F41 },
		{ 5817, 0x6F11 }, { 5818, 0x704C }, { 5819, 0x6EEC }, { 5820, 0x6EF8 }, { 5821, 0x6EFE }, { 5822, 0x6F3F },
		{ 5823, 0x6EF2 }, { 5824, 0x6F31 }, { 5825, 0x6EEF }, { 5826, 0x6F32 }, { 5827, 0x6ECC }, { 5828, 0x6F3E },
		{ 5829, 0x6F13 }, { 5830, 0x6EF7 }, { 5831, 0x6F86 }, { 5832, 0x6F7A }, { 5833, 0x6F78 }, { 5834, 0x6F81 },
		{ 5835, 0x6F80 }, { 5836, 0x6F6F }, { 5837, 0x6F5B }, { 5838, 0x6FF3 }, { 5839, 0x6F6D }, { 5840, 0x6F82 },
		{ 5841, 0x6F7C }, { 5842, 0x6F58 }, { 5843, 0x6F8E }, { 5844, 0x6F91 }, { 5845, 0x6FC2 }, { 5846, 0x6F66 },
		{ 5847, 0x6FB3 }, { 5848, 0x6FA3 }, { 5849, 0x6FA1 }, { 5850, 0x6FA4 }, { 5851, 0x6FB9 }, { 5852, 0x6FC6 },
		{ 5853, 0x6FAA }, { 5854, 0x6FDF }, { 5855, 0x6FD5 }, { 5856, 0x6FEC }, { 5857, 0x6FD4 }, { 5858, 0x6FD8 },
		{ 5859, 0x6FF1 }, { 5860, 0x6FEE }, { 5861, 0x6FDB }, { 5862, 0x7009 }, { 5863, 0x700B }, { 5864, 0x6FFA },
		{ 5865, 0x7011 }, { 5866, 0x7001 }, { 5867, 0x700F }, { 5868, 0x6FFE }, { 5869, 0x701B }, { 5870, 0x701A },
		{ 5871, 0x6F74 }, { 5872, 0x701D }, { 5873, 0x7018 }, { 5874, 0x701F }, { 5875, 0x7030 }, { 5876, 0x703E },
		{ 5877, 0x7032 }, { 5878, 0x7051 }, { 5879, 0x7063 }, { 5880, 0x7099 }, { 5881, 0x7092 }, { 5882, 0x70AF },
		{ 5883, 0x70F1 }, { 5884, 0x70AC }, { 5885, 0x70B8 }, { 5886, 0x70B3 }, { 5887, 0x70AE }, { 5888, 0x70DF },
		{ 5889, 0x70CB }, { 5890, 0x70DD }, { 5891, 0x70D9 }, { 5892, 0x7109 }, { 5893, 0x70FD }, { 5894, 0x711C },
		{ 5895, 0x7119 }, { 5896, 0x7165 }, { 5897, 0x7155 }, { 5898, 0x7188 }, { 5899, 0x7166 }, { 5900, 0x7162 },
		{ 5901, 0x714C }, { 5902, 0x7156 }, { 5903, 0x716C }, { 5904, 0x718F }, { 5905, 0x71FB }, { 5906, 0x7184 },
		{ 5907, 0x7195 }, { 5908, 0x71A8 }, { 5909, 0x71AC }, { 5910, 0x71D7 }, { 5911, 0x71B9 }, { 5912, 0x71BE },
		{ 5913, 0x71D2 }, { 5914, 0x71C9 }, { 5915, 0x71D4 }, { 5916, 0x71CE }, { 5917, 0x71E0 }, { 5918, 0x71EC },
		{ 5919, 0x71E7 }, { 5920, 0x71F5 }, { 5921, 0x71FC }, { 5922, 0x71F9 }, { 5923, 0x71FF }, { 5924, 0x720D },
		{ 5925, 0x7210 }, { 5926, 0x721B }, { 5927, 0x7228 }, { 5928, 0x722D }, { 5929, 0x722C }, { 5930, 0x7230 },
		{ 5931, 0x7232 }, { 5932, 0x723B }, { 5933, 0x723C }, { 5934, 0x723F }, { 5935, 0x7240 }, { 5936, 0x7246 },
		{ 5937, 0x724B }, { 5938, 0x7258 }, { 5939, 0x7274 }, { 5940, 0x727E }, { 5941, 0x7282 }, { 5942, 0x7281 },
		{ 5943, 0x7287 }, { 5944, 0x7292 }, { 5945, 0x7296 }, { 5946, 0x72A2 }, { 5947, 0x72A7 }, { 5948, 0x72B9 },
		{ 5949, 0x72B2 }, { 5950, 0x72C3 }, { 5951, 0x72C6 }, { 5952, 0x72C4 }, { 5953, 0x72CE }, { 5954, 0x72D2 },
		{ 5955, 0x72E2 }, { 5956, 0x72E0 }, { 5957, 0x72E1 }, { 5958, 0x72F9 }, { 5959, 0x72F7 }, { 5960, 0x500F },
		{ 5961, 0x7317 }, { 5962, 0x730A }, { 5963, 0x731C }, { 5964, 0x7316 }, { 5965, 0x731D }, { 5966, 0x7334 },
		{ 5967, 0x732F }, { 5968, 0x7329 }, { 5969, 0x7325 }, { 5970, 0x733E }, { 5971, 0x734E }, { 5972, 0x734F },
		{ 5973, 0x9ED8 }, { 5974, 0x7357 }, { 5975, 0x736A }, { 5976, 0x7368 }, { 5977, 0x7370 }, { 5978, 0x7378 },
		{ 5979, 0x7375 }, { 5980, 0x737B }, { 5981, 0x737A }, { 5982, 0x73C8 }, { 5983, 0x73B3 }, { 5984, 0x73CE },
		{ 5985, 0x73BB }, { 5986, 0x73C0 }, { 5987, 0x73E5 }, { 5988, 0x73EE }, { 5989, 0x73DE }, { 5990, 0x74A2 },
		{ 5991, 0x7405 }, { 5992, 0x746F }, { 5993, 0x7425 }, { 5994, 0x73F8 }, { 5995, 0x7432 }, { 5996, 0x743A },
		{ 5997, 0x7455 }, { 5998, 0x743F }, { 5999, 0x745F }, { 6000, 0x7459 }, { 6001, 0x7441 }, { 6002, 0x745C },
		{ 6003, 0x7469 }, { 6004, 0x7470 }, { 6005, 0x7463 }, { 6006, 0x746A }, { 6007, 0x7476 }, { 6008, 0x747E },
		{ 6009, 0x748B }, { 6010, 0x749E }, { 6011, 0x74A7 }, { 6012, 0x74CA }, { 6013, 0x74CF }, { 6014, 0x74D4 },
		{ 6015, 0x73F1 }, { 6016, 0x74E0 }, { 6017, 0x74E3 }, { 6018, 0x74E7 }, { 6019, 0x74E9 }, { 6020, 0x74EE },
		{ 6021, 0x74F2 }, { 6022, 0x74F0 }, { 6023, 0x74F1 }, { 6024, 0x74F8 }, { 6025, 0x74F7 }, { 6026, 0x7504 },
		{ 6027, 0x7503 }, { 6028, 0x7505 }, { 6029, 0x750C }, { 6030, 0x750E }, { 6031, 0x750D }, { 6032, 0x7515 },
		{ 6033, 0x7513 }, { 6034, 0x751E }, { 6035, 0x7526 }, { 6036, 0x752C }, { 6037, 0x753C }, { 6038, 0x7544 },
		{ 6039, 0x754D }, { 6040, 0x754A }, { 6041, 0x7549 }, { 6042, 0x755B }, { 6043, 0x7546 }, { 6044, 0x755A },
		{ 6045, 0x7569 }, { 6046, 0x7564 }, { 6047, 0x7567 }, { 6048, 0x756B }, { 6049, 0x756D }, { 6050, 0x7578 },
		{ 6051, 0x7576 }, { 6052, 0x7586 }, { 6053, 0x7587 }, { 6054, 0x7574 }, { 6055, 0x758A }, { 6056, 0x7589 },
		{ 6057, 0x7582 }, { 6058, 0x7594 }, { 6059, 0x759A }, { 6060, 0x759D }, { 6061, 0x75A5 }, { 6062, 0x75A3 },
		{ 6063, 0x75C2 }, { 6064, 0x75B3 }, { 6065, 0x75C3 }, { 6066, 0x75B5 }, { 6067, 0x75BD }, { 6068, 0x75B8 },
		{ 6069, 0x75BC }, { 6070, 0x75B1 }, { 6071, 0x75CD }, { 6072, 0x75CA }, { 6073, 0x75D2 }, { 6074, 0x75D9 },
		{ 6075, 0x75E3 }, { 6076, 0x75DE }, { 6077, 0x75FE }, { 6078, 0x75FF }, { 6079, 0x75FC }, { 6080, 0x7601 },
		{ 6081, 0x75F0 }, { 6082, 0x75FA }, { 6083, 0x75F2 }, { 6084, 0x75F3 }, { 6085, 0x760B }, { 6086, 0x760D },
		{ 6087, 0x7609 }, { 6088, 0x761F }, { 6089, 0x7627 }, { 6090, 0x7620 }, { 6091, 0x7621 }, { 6092, 0x7622 },
		{ 6093, 0x7624 }, { 6094, 0x7634 }, { 6095, 0x7630 }, { 6096, 0x763B }, { 6097, 0x7647 }, { 6098, 0x7648 },
		{ 6099, 0x7646 }, { 6100, 0x765C }, { 6101, 0x7658 }, { 6102, 0x7661 }, { 6103, 0x7662 }, { 6104, 0x7668 },
		{ 6105, 0x7669 }, { 6106, 0x766A }, { 6107, 0x7667 }, { 6108, 0x766C }, { 6109, 0x7670 }, { 6110, 0x7672 },
		{ 6111, 0x7676 }, { 6112, 0x7678 }, { 6113, 0x767C }, { 6114, 0x7680 }, { 6115, 0x7683 }, { 6116, 0x7688 },
		{ 6117, 0x768B }, { 6118, 0x768E }, { 6119, 0x7696 }, { 6120, 0x7693 }, { 6121, 0x7699 }, { 6122, 0x769A },
		{ 6123, 0x76B0 }, { 6124, 0x76B4 }, { 6125, 0x76B8 }, { 6126, 0x76B9 }, { 6127, 0x76BA }, { 6128, 0x76C2 },
		{ 6129, 0x76CD }, { 6130, 0x76D6 }, { 6131, 0x76D2 }, { 6132, 0x76DE }, { 6133, 0x76E1 }, { 6134, 0x76E5 },
		{ 6135, 0x76E7 }, { 6136, 0x76EA }, { 6137, 0x862F }, { 6138, 0x76FB }, { 6139, 0x7708 }, { 6140, 0x7707 },
		{ 6141, 0x7704 }, { 6142, 0x7729 }, { 6143, 0x7724 }, { 6144, 0x771E }, { 6145, 0x7725 }, { 6146, 0x7726 },
		{ 6147, 0x771B }, { 6148, 0x7737 }, { 6149, 0x7738 }, { 6150, 0x7747 }, { 6151, 0x775A }, { 6152, 0x7768 },
		{ 6153, 0x776B }, { 6154, 0x775B }, { 6155, 0x7765 }, { 6156, 0x777F }, { 6157, 0x777E }, { 6158, 0x7779 },
		{ 6159, 0x778E }, { 6160, 0x778B }, { 6161, 0x7791 }, { 6162, 0x77A0 }, { 6163, 0x779E }, { 6164, 0x77B0 },
		{ 6165, 0x77B6 }, { 6166, 0x77B9 }, { 6167, 0x77BF }, { 6168, 0x77BC }, { 6169, 0x77BD }, { 6170, 0x77BB },
		{ 6171, 0x77C7 }, { 6172, 0x77CD }, { 6173, 0x77D7 }, { 6174, 0x77DA }, { 6175, 0x77DC }, { 6176, 0x77E3 },
		{ 6177, 0x77EE }, { 6178, 0x77FC }, { 6179, 0x780C }, { 6180, 0x7812 }, { 6181, 0x7926 }, { 6182, 0x7820 },
		{ 6183, 0x792A }, { 6184, 0x7845 }, { 6185, 0x788E }, { 6186, 0x7874 }, { 6187, 0x7886 }, { 6188, 0x787C },
		{ 6189, 0x789A }, { 6190, 0x788C }, { 6191, 0x78A3 }, { 6192, 0x78B5 }, { 6193, 0x78AA }, { 6194, 0x78AF },
		{ 6195, 0x78D1 }, { 6196, 0x78C6 }, { 6197, 0x78CB }, { 6198, 0x78D4 }, { 6199, 0x78BE }, { 6200, 0x78BC },
		{ 6201, 0x78C5 }, { 6202, 0x78CA }, { 6203, 0x78EC }, { 6204, 0x78E7 }, { 6205, 0x78DA }, { 6206, 0x78FD },
		{ 6207, 0x78F4 }, { 6208, 0x7907 }, { 6209, 0x7912 }, { 6210, 0x7911 }, { 6211, 0x7919 }, { 6212, 0x792C },
		{ 6213, 0x792B }, { 6214, 0x7940 }, { 6215, 0x7960 }, { 6216, 0x7957 }, { 6217, 0x795F }, { 6218, 0x795A },
		{ 6219, 0x7955 }, { 6220, 0x7953 }, { 6221, 0x797A }, { 6222, 0x797F }, { 6223, 0x798A }, { 6224, 0x799D },
		{ 6225, 0x79A7 }, { 6226, 0x9F4B }, { 6227, 0x79AA }, { 6228, 0x79AE }, { 6229, 0x79B3 }, { 6230, 0x79B9 },
		{ 6231, 0x79BA }, { 6232, 0x79C9 }, { 6233, 0x79D5 }, { 6234, 0x79E7 }, { 6235, 0x79EC }, { 6236, 0x79E1 },
		{ 6237, 0x79E3 }, { 6238, 0x7A08 }, { 6239, 0x7A0D }, { 6240, 0x7A18 }, { 6241, 0x7A19 }, { 6242, 0x7A20 },
		{ 6243, 0x7A1F }, { 6244, 0x7980 }, { 6245, 0x7A31 }, { 6246, 0x7A3B }, { 6247, 0x7A3E }, { 6248, 0x7A37 },
		{ 6249, 0x7A43 }, { 6250, 0x7A57 }, { 6251, 0x7A49 }, { 6252, 0x7A61 }, { 6253, 0x7A62 }, { 6254, 0x7A69 },
		{ 6255, 0x9F9D }, { 6256, 0x7A70 }, { 6257, 0x7A79 }, { 6258, 0x7A7D }, { 6259, 0x7A88 }, { 6260, 0x7A97 },
		{ 6261, 0x7A95 }, { 6262, 0x7A98 }, { 6263, 0x7A96 }, { 6264, 0x7AA9 }, { 6265, 0x7AC8 }, { 6266, 0x7AB0 },
		{ 6267, 0x7AB6 }, { 6268, 0x7AC5 }, { 6269, 0x7AC4 }, { 6270, 0x7ABF }, { 6271, 0x9083 }, { 6272, 0x7AC7 },
		{ 6273, 0x7ACA }, { 6274, 0x7ACD }, { 6275, 0x7ACF }, { 6276, 0x7AD5 }, { 6277, 0x7AD3 }, { 6278, 0x7AD9 },
		{ 6279, 0x7ADA }, { 6280, 0x7ADD }, { 6281, 0x7AE1 }, { 6282, 0x7AE2 }, { 6283, 0x7AE6 }, { 6284, 0x7AED },
		{ 6285, 0x7AF0 }, { 6286, 0x7B02 }, { 6287, 0x7B0F }, { 6288, 0x7B0A }, { 6289, 0x7B06 }, { 6290, 0x7B33 },
		{ 6291, 0x7B18 }, { 6292, 0x7B19 }, { 6293, 0x7B1E }, { 6294, 0x7B35 }, { 6295, 0x7B28 }, { 6296, 0x7B36 },
		{ 6297, 0x7B50 }, { 6298, 0x7B7A }, { 6299, 0x7B04 }, { 6300, 0x7B4D }, { 6301, 0x7B0B }, { 6302, 0x7B4C },
		{ 6303, 0x7B45 }, { 6304, 0x7B75 }, { 6305, 0x7B65 }, { 6306, 0x7B74 }, { 6307, 0x7B67 }, { 6308, 0x7B70 },
		{ 6309, 0x7B71 }, { 6310, 0x7B6C }, { 6311, 0x7B6E }, { 6312, 0x7B9D }, { 6313, 0x7B98 }, { 6314, 0x7B9F },
		{ 6315, 0x7B8D }, { 6316, 0x7B9C }, { 6317, 0x7B9A }, { 6318, 0x7B8B }, { 6319, 0x7B92 }, { 6320, 0x7B8F },
		{ 6321, 0x7B5D }, { 6322, 0x7B99 }, { 6323, 0x7BCB }, { 6324, 0x7BC1 }, { 6325, 0x7BCC }, { 6326, 0x7BCF },
		{ 6327, 0x7BB4 }, { 6328, 0x7BC6 }, { 6329, 0x7BDD }, { 6330, 0x7BE9 }, { 6331, 0x7C11 }, { 6332, 0x7C14 },
		{ 6333, 0x7BE6 }, { 6334, 0x7BE5 }, { 6335, 0x7C60 }, { 6336, 0x7C00 }, { 6337, 0x7C07 }, { 6338, 0x7C13 },
		{ 6339, 0x7BF3 }, { 6340, 0x7BF7 }, { 6341, 0x7C17 }, { 6342, 0x7C0D }, { 6343, 0x7BF6 }, { 6344, 0x7C23 },
		{ 6345, 0x7C27 }, { 6346, 0x7C2A }, { 6347, 0x7C1F }, { 6348, 0x7C37 }, { 6349, 0x7C2B }, { 6350, 0x7C3D },
		{ 6351, 0x7C4C }, { 6352, 0x7C43 }, { 6353, 0x7C54 }, { 6354, 0x7C4F }, { 6355, 0x7C40 }, { 6356, 0x7C50 },
		{ 6357, 0x7C58 }, { 6358, 0x7C5F }, { 6359, 0x7C64 }, { 6360, 0x7C56 }, { 6361, 0x7C65 }, { 6362, 0x7C6C },
		{ 6363, 0x7C75 }, { 6364, 0x7C83 }, { 6365, 0x7C90 }, { 6366, 0x7CA4 }, { 6367, 0x7CAD }, { 6368, 0x7CA2 },
		{ 6369, 0x7CAB }, { 6370, 0x7CA1 }, { 6371, 0x7CA8 }, { 6372, 0x7CB3 }, { 6373, 0x7CB2 }, { 6374, 0x7CB1 },
		{ 6375, 0x7CAE }, { 6376, 0x7CB9 }, { 6377, 0x7CBD }, { 6378, 0x7CC0 }, { 6379, 0x7CC5 }, { 6380, 0x7CC2 },
		{ 6381, 0x7CD8 }, { 6382, 0x7CD2 }, { 6383, 0x7CDC }, { 6384, 0x7CE2 }, { 6385, 0x9B3B }, { 6386, 0x7CEF },
		{ 6387, 0x7CF2 }, { 6388, 0x7CF4 }, { 6389, 0x7CF6 }, { 6390, 0x7CFA }, { 6391, 0x7D06 }, { 6392, 0x7D02 },
		{ 6393, 0x7D1C }, { 6394, 0x7D15 }, { 6395, 0x7D0A }, { 6396, 0x7D45 }, { 6397, 0x7D4B }, { 6398, 0x7D2E },
		{ 6399, 0x7D32 }, { 6400, 0x7D3F }, { 6401, 0x7D35 }, { 6402, 0x7D46 }, { 6403, 0x7D73 }, { 6404, 0x7D56 },
		{ 6405, 0x7D4E }, { 6406, 0x7D72 }, { 6407, 0x7D68 }, { 6408, 0x7D6E }, { 6409, 0x7D4F }, { 6410, 0x7D63 },
		{ 6411, 0x7D93 }, { 6412, 0x7D89 }, { 6413, 0x7D5B }, { 6414, 0x7D8F }, { 6415, 0x7D7D }, { 6416, 0x7D9B },
		{ 6417, 0x7DBA }, { 6418, 0x7DAE }, { 6419, 0x7DA3 }, { 6420, 0x7DB5 }, { 6421, 0x7DC7 }, { 6422, 0x7DBD },
		{ 6423, 0x7DAB }, { 6424, 0x7E3D }, { 6425, 0x7DA2 }, { 6426, 0x7DAF }, { 6427, 0x7DDC }, { 6428, 0x7DB8 },
		{ 6429, 0x7D9F }, { 6430, 0x7DB0 }, { 6431, 0x7DD8 }, { 6432, 0x7DDD }, { 6433, 0x7DE4 }, { 6434, 0x7DDE },
		{ 6435, 0x7DFB }, { 6436, 0x7DF2 }, { 6437, 0x7DE1 }, { 6438, 0x7E05 }, { 6439, 0x7E0A }, { 6440, 0x7E23 },
		{ 6441, 0x7E21 }, { 6442, 0x7E12 }, { 6443, 0x7E31 }, { 6444, 0x7E1F }, { 6445, 0x7E09 }, { 6446, 0x7E0B },
		{ 6447, 0x7E22 }, { 6448, 0x7E46 }, { 6449, 0x7E66 }, { 6450, 0x7E3B }, { 6451, 0x7E35 }, { 6452, 0x7E39 },
		{ 6453, 0x7E43 }, { 6454, 0x7E37 }, { 6455, 0x7E32 }, { 6456, 0x7E3A }, { 6457, 0x7E67 }, { 6458, 0x7E5D },
		{ 6459, 0x7E56 }, { 6460, 0x7E5E }, { 6461, 0x7E59 }, { 6462, 0x7E5A }, { 6463, 0x7E79 }, { 6464, 0x7E6A },
		{ 6465, 0x7E69 }, { 6466, 0x7E7C }, { 6467, 0x7E7B }, { 6468, 0x7E83 }, { 6469, 0x7DD5 }, { 6470, 0x7E7D },
		{ 6471, 0x8FAE }, { 6472, 0x7E7F }, { 6473, 0x7E88 }, { 6474, 0x7E89 }, { 6475, 0x7E8C }, { 6476, 0x7E92 },
		{ 6477, 0x7E90 }, { 6478, 0x7E93 }, { 6479, 0x7E94 }, { 6480, 0x7E96 }, { 6481, 0x7E8E }, { 6482, 0x7E9B },
		{ 6483, 0x7E9C }, { 6484, 0x7F38 }, { 6485, 0x7F3A }, { 6486, 0x7F45 }, { 6487, 0x7F4C }, { 6488, 0x7F4D },
		{ 6489, 0x7F4E }, { 6490, 0x7F50 }, { 6491, 0x7F51 }, { 6492, 0x7F55 }, { 6493, 0x7F54 }, { 6494, 0x7F58 },
		{ 6495, 0x7F5F }, { 6496, 0x7F60 }, { 6497, 0x7F68 }, { 6498, 0x7F69 }, { 6499, 0x7F67 }, { 6500, 0x7F78 },
		{ 6501, 0x7F82 }, { 6502, 0x7F86 }, { 6503, 0x7F83 }, { 6504, 0x7F88 }, { 6505, 0x7F87 }, { 6506, 0x7F8C },
		{ 6507, 0x7F94 }, { 6508, 0x7F9E }, { 6509, 0x7F9D }, { 6510, 0x7F9A }, { 6511, 0x7FA3 }, { 6512, 0x7FAF },
		{ 6513, 0x7FB2 }, { 6514, 0x7FB9 }, { 6515, 0x7FAE }, { 6516, 0x7FB6 }, { 6517, 0x7FB8 }, { 6518, 0x8B71 },
		{ 6519, 0x7FC5 }, { 6520, 0x7FC6 }, { 6521, 0x7FCA }, { 6522, 0x7FD5 }, { 6523, 0x7FD4 }, { 6524, 0x7FE1 },
		{ 6525, 0x7FE6 }, { 6526, 0x7FE9 }, { 6527, 0x7FF3 }, { 6528, 0x7FF9 }, { 6529, 0x98DC }, { 6530, 0x8006 },
		{ 6531, 0x8004 }, { 6532, 0x800B }, { 6533, 0x8012 }, { 6534, 0x8018 }, { 6535, 0x8019 }, { 6536, 0x801C },
		{ 6537, 0x8021 }, { 6538, 0x8028 }, { 6539, 0x803F }, { 6540, 0x803B }, { 6541, 0x804A }, { 6542, 0x8046 },
		{ 6543, 0x8052 }, { 6544, 0x8058 }, { 6545, 0x805A }, { 6546, 0x805F }, { 6547, 0x8062 }, { 6548, 0x8068 },
		{ 6549, 0x8073 }, { 6550, 0x8072 }, { 6551, 0x8070 }, { 6552, 0x8076 }, { 6553, 0x8079 }, { 6554, 0x807D },
		{ 6555, 0x807F }, { 6556, 0x8084 }, { 6557, 0x8086 }, { 6558, 0x8085 }, { 6559, 0x809B }, { 6560, 0x8093 },
		{ 6561, 0x809A }, { 6562, 0x80AD }, { 6563, 0x5190 }, { 6564, 0x80AC }, { 6565, 0x80DB }, { 6566, 0x80E5 },
		{ 6567, 0x80D9 }, { 6568, 0x80DD }, { 6569, 0x80C4 }, { 6570, 0x80DA }, { 6571, 0x80D6 }, { 6572, 0x8109 },
		{ 6573, 0x80EF }, { 6574, 0x80F1 }, { 6575, 0x811B }, { 6576, 0x8129 }, { 6577, 0x8123 }, { 6578, 0x812F },
		{ 6579, 0x814B }, { 6580, 0x968B }, { 6581, 0x8146 }, { 6582, 0x813E }, { 6583, 0x8153 }, { 6584, 0x8151 },
		{ 6585, 0x80FC }, { 6586, 0x8171 }, { 6587, 0x816E }, { 6588, 0x8165 }, { 6589, 0x8166 }, { 6590, 0x8174 },
		{ 6591, 0x8183 }, { 6592, 0x8188 }, { 6593, 0x818A }, { 6594, 0x8180 }, { 6595, 0x8182 }, { 6596, 0x81A0 },
		{ 6597, 0x8195 }, { 6598, 0x81A4 }, { 6599, 0x81A3 }, { 6600, 0x815F }, { 6601, 0x8193 }, { 6602, 0x81A9 },
		{ 6603, 0x81B0 }, { 6604, 0x81B5 }, { 6605, 0x81BE }, { 6606, 0x81B8 }, { 6607, 0x81BD }, { 6608, 0x81C0 },
		{ 6609, 0x81C2 }, { 6610, 0x81BA }, { 6611, 0x81C9 }, { 6612, 0x81CD }, { 6613, 0x81D1 }, { 6614, 0x81D9 },
		{ 6615, 0x81D8 }, { 6616, 0x81C8 }, { 6617, 0x81DA }, { 6618, 0x81DF }, { 6619, 0x81E0 }, { 6620, 0x81E7 },
		{ 6621, 0x81FA }, { 6622, 0x81FB }, { 6623, 0x81FE }, { 6624, 0x8201 }, { 6625, 0x8202 }, { 6626, 0x8205 },
		{ 6627, 0x8207 }, { 6628, 0x820A }, { 6629, 0x820D }, { 6630, 0x8210 }, { 6631, 0x8216 }, { 6632, 0x8229 },
		{ 6633, 0x822B }, { 6634, 0x8238 }, { 6635, 0x8233 }, { 6636, 0x8240 }, { 6637, 0x8259 }, { 6638, 0x8258 },
		{ 6639, 0x825D }, { 6640, 0x825A }, { 6641, 0x825F }, { 6642, 0x8264 }, { 6643, 0x8262 }, { 6644, 0x8268 },
		{ 6645, 0x826A }, { 6646, 0x826B }, { 6647, 0x822E }, { 6648, 0x8271 }, { 6649, 0x8277 }, { 6650, 0x8278 },
		{ 6651, 0x827E }, { 6652, 0x828D }, { 6653, 0x8292 }, { 6654, 0x82AB }, { 6655, 0x829F }, { 6656, 0x82BB },
		{ 6657, 0x82AC }, { 6658, 0x82E1 }, { 6659, 0x82E3 }, { 6660, 0x82DF }, { 6661, 0x82D2 }, { 6662, 0x82F4 },
		{ 6663, 0x82F3 }, { 6664, 0x82FA }, { 6665, 0x8393 }, { 6666, 0x8303 }, { 6667, 0x82FB }, { 6668, 0x82F9 },
		{ 6669, 0x82DE }, { 6670, 0x8306 }, { 6671, 0x82DC }, { 6672, 0x8309 }, { 6673, 0x82D9 }, { 6674, 0x8335 },
		{ 6675, 0x8334 }, { 6676, 0x8316 }, { 6677, 0x8332 }, { 6678, 0x8331 }, { 6679, 0x8340 }, { 6680, 0x8339 },
		{ 6681, 0x8350 }, { 6682, 0x8345 }, { 6683, 0x832F }, { 6684, 0x832B }, { 6685, 0x8317 }, { 6686, 0x8318 },
		{ 6687, 0x8385 }, { 6688, 0x839A }, { 6689, 0x83AA }, { 6690, 0x839F }, { 6691, 0x83A2 }, { 6692, 0x8396 },
		{ 6693, 0x8323 }, { 6694, 0x838E }, { 6695, 0x8387 }, { 6696, 0x838A }, { 6697, 0x837C }, { 6698, 0x83B5 },
		{ 6699, 0x8373 }, { 6700, 0x8375 }, { 6701, 0x83A0 }, { 6702, 0x8389 }, { 6703, 0x83A8 }, { 6704, 0x83F4 },
		{ 6705, 0x8413 }, { 6706, 0x83EB }, { 6707, 0x83CE }, { 6708, 0x83FD }, { 6709, 0x8403 }, { 6710, 0x83D8 },
		{ 6711, 0x840B }, { 6712, 0x83C1 }, { 6713, 0x83F7 }, { 6714, 0x8407 }, { 6715, 0x83E0 }, { 6716, 0x83F2 },
		{ 6717, 0x840D }, { 6718, 0x8422 }, { 6719, 0x8420 }, { 6720, 0x83BD }, { 6721, 0x8438 }, { 6722, 0x8506 },
		{ 6723, 0x83FB }, { 6724, 0x846D }, { 6725, 0x842A }, { 6726, 0x843C }, { 6727, 0x855A }, { 6728, 0x8484 },
		{ 6729, 0x8477 }, { 6730, 0x846B }, { 6731, 0x84AD }, { 6732, 0x846E }, { 6733, 0x8482 }, { 6734, 0x8469 },
		{ 6735, 0x8446 }, { 6736, 0x842C }, { 6737, 0x846F }, { 6738, 0x8479 }, { 6739, 0x8435 }, { 6740, 0x84CA },
		{ 6741, 0x8462 }, { 6742, 0x84B9 }, { 6743, 0x84BF }, { 6744, 0x849F }, { 6745, 0x84D9 }, { 6746, 0x84CD },
		{ 6747, 0x84BB }, { 6748, 0x84DA }, { 6749, 0x84D0 }, { 6750, 0x84C1 }, { 6751, 0x84C6 }, { 6752, 0x84D6 },
		{ 6753, 0x84A1 }, { 6754, 0x8521 }, { 6755, 0x84FF }, { 6756, 0x84F4 }, { 6757, 0x8517 }, { 6758, 0x8518 },
		{ 6759, 0x852C }, { 6760, 0x851F }, { 6761, 0x8515 }, { 6762, 0x8514 }, { 6763, 0x84FC }, { 6764, 0x8540 },
		{ 6765, 0x8563 }, { 6766, 0x8558 }, { 6767, 0x8548 }, { 6768, 0x8541 }, { 6769, 0x8602 }, { 6770, 0x854B },
		{ 6771, 0x8555 }, { 6772, 0x8580 }, { 6773, 0x85A4 }, { 6774, 0x8588 }, { 6775, 0x8591 }, { 6776, 0x858A },
		{ 6777, 0x85A8 }, { 6778, 0x856D }, { 6779, 0x8594 }, { 6780, 0x859B }, { 6781, 0x85EA }, { 6782, 0x8587 },
		{ 6783, 0x859C }, { 6784, 0x8577 }, { 6785, 0x857E }, { 6786, 0x8590 }, { 6787, 0x85C9 }, { 6788, 0x85BA },
		{ 6789, 0x85CF }, { 6790, 0x85B9 }, { 6791, 0x85D0 }, { 6792, 0x85D5 }, { 6793, 0x85DD }, { 6794, 0x85E5 },
		{ 6795, 0x85DC }, { 6796, 0x85F9 }, { 6797, 0x860A }, { 6798, 0x8613 }, { 6799, 0x860B }, { 6800, 0x85FE },
		{ 6801, 0x85FA }, { 6802, 0x8606 }, { 6803, 0x8622 }, { 6804, 0x861A }, { 6805, 0x8630 }, { 6806, 0x863F },
		{ 6807, 0x864D }, { 6808, 0x4E55 }, { 6809, 0x8654 }, { 6810, 0x865F }, { 6811, 0x8667 }, { 6812, 0x8671 },
		{ 6813, 0x8693 }, { 6814, 0x86A3 }, { 6815, 0x86A9 }, { 6816, 0x86AA }, { 6817, 0x868B }, { 6818, 0x868C },
		{ 6819, 0x86B6 }, { 6820, 0x86AF }, { 6821, 0x86C4 }, { 6822, 0x86C6 }, { 6823, 0x86B0 }, { 6824, 0x86C9 },
		{ 6825, 0x8823 }, { 6826, 0x86AB }, { 6827, 0x86D4 }, { 6828, 0x86DE }, { 6829, 0x86E9 }, { 6830, 0x86EC },
		{ 6831, 0x86DF }, { 6832, 0x86DB }, { 6833, 0x86EF }, { 6834, 0x8712 }, { 6835, 0x8706 }, { 6836, 0x8708 },
		{ 6837, 0x8700 }, { 6838, 0x8703 }, { 6839, 0x86FB }, { 6840, 0x8711 }, { 6841, 0x8709 }, { 6842, 0x870D },
		{ 6843, 0x86F9 }, { 6844, 0x870A }, { 6845, 0x8734 }, { 6846, 0x873F }, { 6847, 0x8737 }, { 6848, 0x873B },
		{ 6849, 0x8725 }, { 6850, 0x8729 }, { 6851, 0x871A }, { 6852, 0x8760 }, { 6853, 0x875F }, { 6854, 0x8778 },
		{ 6855, 0x874C }, { 6856, 0x874E }, { 6857, 0x8774 }, { 6858, 0x8757 }, { 6859, 0x8768 }, { 6860, 0x876E },
		{ 6861, 0x8759 }, { 6862, 0x8753 }, { 6863, 0x8763 }, { 6864, 0x876A }, { 6865, 0x8805 }, { 6866, 0x87A2 },
		{ 6867, 0x879F }, { 6868, 0x8782 }, { 6869, 0x87AF }, { 6870, 0x87CB }, { 6871, 0x87BD }, { 6872, 0x87C0 },
		{ 6873, 0x87D0 }, { 6874, 0x96D6 }, { 6875, 0x87AB }, { 6876, 0x87C4 }, { 6877, 0x87B3 }, { 6878, 0x87C7 },
		{ 6879, 0x87C6 }, { 6880, 0x87BB }, { 6881, 0x87EF }, { 6882, 0x87F2 }, { 6883, 0x87E0 }, { 6884, 0x880F },
		{ 6885, 0x880D }, { 6886, 0x87FE }, { 6887, 0x87F6 }, { 6888, 0x87F7 }, { 6889, 0x880E }, { 6890, 0x87D2 },
		{ 6891, 0x8811 }, { 6892, 0x8816 }, { 6893, 0x8815 }, { 6894, 0x8822 }, { 6895, 0x8821 }, { 6896, 0x8831 },
		{ 6897, 0x8836 }, { 6898, 0x8839 }, { 6899, 0x8827 }, { 6900, 0x883B }, { 6901, 0x8844 }, { 6902, 0x8842 },
		{ 6903, 0x8852 }, { 6904, 0x8859 }, { 6905, 0x885E }, { 6906, 0x8862 }, { 6907, 0x886B }, { 6908, 0x8881 },
		{ 6909, 0x887E }, { 6910, 0x889E }, { 6911, 0x8875 }, { 6912, 0x887D }, { 6913, 0x88B5 }, { 6914, 0x8872 },
		{ 6915, 0x8882 }, { 6916, 0x8897 }, { 6917, 0x8892 }, { 6918, 0x88AE }, { 6919, 0x8899 }, { 6920, 0x88A2 },
		{ 6921, 0x888D }, { 6922, 0x88A4 }, { 6923, 0x88B0 }, { 6924, 0x88BF }, { 6925, 0x88B1 }, { 6926, 0x88C3 },
		{ 6927, 0x88C4 }, { 6928, 0x88D4 }, { 6929, 0x88D8 }, { 6930, 0x88D9 }, { 6931, 0x88DD }, { 6932, 0x88F9 },
		{ 6933, 0x8902 }, { 6934, 0x88FC }, { 6935, 0x88F4 }, { 6936, 0x88E8 }, { 6937, 0x88F2 }, { 6938, 0x8904 },
		{ 6939, 0x890C }, { 6940, 0x890A }, { 6941, 0x8913 }, { 6942, 0x8943 }, { 6943, 0x891E }, { 6944, 0x8925 },
		{ 6945, 0x892A }, { 6946, 0x892B }, { 6947, 0x8941 }, { 6948, 0x8944 }, { 6949, 0x893B }, { 6950, 0x8936 },
		{ 6951, 0x8938 }, { 6952, 0x894C }, { 6953, 0x891D }, { 6954, 0x8960 }, { 6955, 0x895E }, { 6956, 0x8966 },
		{ 6957, 0x8964 }, { 6958, 0x896D }, { 6959, 0x896A }, { 6960, 0x896F }, { 6961, 0x8974 }, { 6962, 0x8977 },
		{ 6963, 0x897E }, { 6964, 0x8983 }, { 6965, 0x8988 }, { 6966, 0x898A }, { 6967, 0x8993 }, { 6968, 0x8998 },
		{ 6969, 0x89A1 }, { 6970, 0x89A9 }, { 6971, 0x89A6 }, { 6972, 0x89AC }, { 6973, 0x89AF }, { 6974, 0x89B2 },
		{ 6975, 0x89BA }, { 6976, 0x89BD }, { 6977, 0x89BF }, { 6978, 0x89C0 }, { 6979, 0x89DA }, { 6980, 0x89DC },
		{ 6981, 0x89DD }, { 6982, 0x89E7 }, { 6983, 0x89F4 }, { 6984, 0x89F8 }, { 6985, 0x8A03 }, { 6986, 0x8A16 },
		{ 6987, 0x8A10 }, { 6988, 0x8A0C }, { 6989, 0x8A1B }, { 6990, 0x8A1D }, { 6991, 0x8A25 }, { 6992, 0x8A36 },
		{ 6993, 0x8A41 }, { 6994, 0x8A5B }, { 6995, 0x8A52 }, { 6996, 0x8A46 }, { 6997, 0x8A48 }, { 6998, 0x8A7C },
		{ 6999, 0x8A6D }, { 7000, 0x8A6C }, { 7001, 0x8A62 }, { 7002, 0x8A85 }, { 7003, 0x8A82 }, { 7004, 0x8A84 },
		{ 7005, 0x8AA8 }, { 7006, 0x8AA1 }, { 7007, 0x8A91 }, { 7008, 0x8AA5 }, { 7009, 0x8AA6 }, { 7010, 0x8A9A },
		{ 7011, 0x8AA3 }, { 7012, 0x8AC4 }, { 7013, 0x8ACD }, { 7014, 0x8AC2 }, { 7015, 0x8ADA }, { 7016, 0x8AEB },
		{ 7017, 0x8AF3 }, { 7018, 0x8AE7 }, { 7019, 0x8AE4 }, { 7020, 0x8AF1 }, { 7021, 0x8B14 }, { 7022, 0x8AE0 },
		{ 7023, 0x8AE2 }, { 7024, 0x8AF7 }, { 7025, 0x8ADE }, { 7026, 0x8ADB }, { 7027, 0x8B0C }, { 7028, 0x8B07 },
		{ 7029, 0x8B1A }, { 7030, 0x8AE1 }, { 7031, 0x8B16 }, { 7032, 0x8B10 }, { 7033, 0x8B17 }, { 7034, 0x8B20 },
		{ 7035, 0x8B33 }, { 7036, 0x97AB }, { 7037, 0x8B26 }, { 7038, 0x8B2B }, { 7039, 0x8B3E }, { 7040, 0x8B28 },
		{ 7041, 0x8B41 }, { 7042, 0x8B4C }, { 7043, 0x8B4F }, { 7044, 0x8B4E }, { 7045, 0x8B49 }, { 7046, 0x8B56 },
		{ 7047, 0x8B5B }, { 7048, 0x8B5A }, { 7049, 0x8B6B }, { 7050, 0x8B5F }, { 7051, 0x8B6C }, { 7052, 0x8B6F },
		{ 7053, 0x8B74 }, { 7054, 0x8B7D }, { 7055, 0x8B80 }, { 7056, 0x8B8C }, { 7057, 0x8B8E }, { 7058, 0x8B92 },
		{ 7059, 0x8B93 }, { 7060, 0x8B96 }, { 7061, 0x8B99 }, { 7062, 0x8B9A }, { 7063, 0x8C3A }, { 7064, 0x8C41 },
		{ 7065, 0x8C3F }, { 7066, 0x8C48 }, { 7067, 0x8C4C }, { 7068, 0x8C4E }, { 7069, 0x8C50 }, { 7070, 0x8C55 },
		{ 7071, 0x8C62 }, { 7072, 0x8C6C }, { 7073, 0x8C78 }, { 7074, 0x8C7A }, { 7075, 0x8C82 }, { 7076, 0x8C89 },
		{ 7077, 0x8C85 }, { 7078, 0x8C8A }, { 7079, 0x8C8D }, { 7080, 0x8C8E }, { 7081, 0x8C94 }, { 7082, 0x8C7C },
		{ 7083, 0x8C98 }, { 7084, 0x621D }, { 7085, 0x8CAD }, { 7086, 0x8CAA }, { 7087, 0x8CBD }, { 7088, 0x8CB2 },
		{ 7089, 0x8CB3 }, { 7090, 0x8CAE }, { 7091, 0x8CB6 }, { 7092, 0x8CC8 }, { 7093, 0x8CC1 }, { 7094, 0x8CE4 },
		{ 7095, 0x8CE3 }, { 7096, 0x8CDA }, { 7097, 0x8CFD }, { 7098, 0x8CFA }, { 7099, 0x8CFB }, { 7100, 0x8D04 },
		{ 7101, 0x8D05 }, { 7102, 0x8D0A }, { 7103, 0x8D07 }, { 7104, 0x8D0F }, { 7105, 0x8D0D }, { 7106, 0x8D10 },
		{ 7107, 0x9F4E }, { 7108, 0x8D13 }, { 7109, 0x8CCD }, { 7110, 0x8D14 }, { 7111, 0x8D16 }, { 7112, 0x8D67 },
		{ 7113, 0x8D6D }, { 7114, 0x8D71 }, { 7115, 0x8D73 }, { 7116, 0x8D81 }, { 7117, 0x8D99 }, { 7118, 0x8DC2 },
		{ 7119, 0x8DBE }, { 7120, 0x8DBA }, { 7121, 0x8DCF }, { 7122, 0x8DDA }, { 7123, 0x8DD6 }, { 7124, 0x8DCC },
		{ 7125, 0x8DDB }, { 7126, 0x8DCB }, { 7127, 0x8DEA }, { 7128, 0x8DEB }, { 7129, 0x8DDF }, { 7130, 0x8DE3 },
		{ 7131, 0x8DFC }, { 7132, 0x8E08 }, { 7133, 0x8E09 }, { 7134, 0x8DFF }, { 7135, 0x8E1D }, { 7136, 0x8E1E },
		{ 7137, 0x8E10 }, { 7138, 0x8E1F }, { 7139, 0x8E42 }, { 7140, 0x8E35 }, { 7141, 0x8E30 }, { 7142, 0x8E34 },
		{ 7143, 0x8E4A }, { 7144, 0x8E47 }, { 7145, 0x8E49 }, { 7146, 0x8E4C }, { 7147, 0x8E50 }, { 7148, 0x8E48 },
		{ 7149, 0x8E59 }, { 7150, 0x8E64 }, { 7151, 0x8E60 }, { 7152, 0x8E2A }, { 7153, 0x8E63 }, { 7154, 0x8E55 },
		{ 7155, 0x8E76 }, { 7156, 0x8E72 }, { 7157, 0x8E7C }, { 7158, 0x8E81 }, { 7159, 0x8E87 }, { 7160, 0x8E85 },
		{ 7161, 0x8E84 }, { 7162, 0x8E8B }, { 7163, 0x8E8A }, { 7164, 0x8E93 }, { 7165, 0x8E91 }, { 7166, 0x8E94 },
		{ 7167, 0x8E99 }, { 7168, 0x8EAA }, { 7169, 0x8EA1 }, { 7170, 0x8EAC }, { 7171, 0x8EB0 }, { 7172, 0x8EC6 },
		{ 7173, 0x8EB1 }, { 7174, 0x8EBE }, { 7175, 0x8EC5 }, { 7176, 0x8EC8 }, { 7177, 0x8ECB }, { 7178, 0x8EDB },
		{ 7179, 0x8EE3 }, { 7180, 0x8EFC }, { 7181, 0x8EFB }, { 7182, 0x8EEB }, { 7183, 0x8EFE }, { 7184, 0x8F0A },
		{ 7185, 0x8F05 }, { 7186, 0x8F15 }, { 7187, 0x8F12 }, { 7188, 0x8F19 }, { 7189, 0x8F13 }, { 7190, 0x8F1C },
		{ 7191, 0x8F1F }, { 7192, 0x8F1B }, { 7193, 0x8F0C }, { 7194, 0x8F26 }, { 7195, 0x8F33 }, { 7196, 0x8F3B },
		{ 7197, 0x8F39 }, { 7198, 0x8F45 }, { 7199, 0x8F42 }, { 7200, 0x8F3E }, { 7201, 0x8F4C }, { 7202, 0x8F49 },
		{ 7203, 0x8F46 }, { 7204, 0x8F4E }, { 7205, 0x8F57 }, { 7206, 0x8F5C }, { 7207, 0x8F62 }, { 7208, 0x8F63 },
		{ 7209, 0x8F64 }, { 7210, 0x8F9C }, { 7211, 0x8F9F }, { 7212, 0x8FA3 }, { 7213, 0x8FAD }, { 7214, 0x8FAF },
		{ 7215, 0x8FB7 }, { 7216, 0x8FDA }, { 7217, 0x8FE5 }, { 7218, 0x8FE2 }, { 7219, 0x8FEA }, { 7220, 0x8FEF },
		{ 7221, 0x9087 }, { 7222, 0x8FF4 }, { 7223, 0x9005 }, { 7224, 0x8FF9 }, { 7225, 0x8FFA }, { 7226, 0x9011 },
		{ 7227, 0x9015 }, { 7228, 0x9021 }, { 7229, 0x900D }, { 7230, 0x901E }, { 7231, 0x9016 }, { 7232, 0x900B },
		{ 7233, 0x9027 }, { 7234, 0x9036 }, { 7235, 0x9035 }, { 7236, 0x9039 }, { 7237, 0x8FF8 }, { 7238, 0x904F },
		{ 7239, 0x9050 }, { 7240, 0x9051 }, { 7241, 0x9052 }, { 7242, 0x900E }, { 7243, 0x9049 }, { 7244, 0x903E },
		{ 7245, 0x9056 }, { 7246, 0x9058 }, { 7247, 0x905E }, { 7248, 0x9068 }, { 7249, 0x906F }, { 7250, 0x9076 },
		{ 7251, 0x96A8 }, { 7252, 0x9072 }, { 7253, 0x9082 }, { 7254, 0x907D }, { 7255, 0x9081 }, { 7256, 0x9080 },
		{ 7257, 0x908A }, { 7258, 0x9089 }, { 7259, 0x908F }, { 7260, 0x90A8 }, { 7261, 0x90AF }, { 7262, 0x90B1 },
		{ 7263, 0x90B5 }, { 7264, 0x90E2 }, { 7265, 0x90E4 }, { 7266, 0x6248 }, { 7267, 0x90DB }, { 7268, 0x9102 },
		{ 7269, 0x9112 }, { 7270, 0x9119 }, { 7271, 0x9132 }, { 7272, 0x9130 }, { 7273, 0x914A }, { 7274, 0x9156 },
		{ 7275, 0x9158 }, { 7276, 0x9163 }, { 7277, 0x9165 }, { 7278, 0x9169 }, { 7279, 0x9173 }, { 7280, 0x9172 },
		{ 7281, 0x918B }, { 7282, 0x9189 }, { 7283, 0x9182 }, { 7284, 0x91A2 }, { 7285, 0x91AB }, { 7286, 0x91AF },
		{ 7287, 0x91AA }, { 7288, 0x91B5 }, { 7289, 0x91B4 }, { 7290, 0x91BA }, { 7291, 0x91C0 }, { 7292, 0x91C1 },
		{ 7293, 0x91C9 }, { 7294, 0x91CB }, { 7295, 0x91D0 }, { 7296, 0x91D6 }, { 7297, 0x91DF }, { 7298, 0x91E1 },
		{ 7299, 0x91DB }, { 7300, 0x91FC }, { 7301, 0x91F5 }, { 7302, 0x91F6 }, { 7303, 0x921E }, { 7304, 0x91FF },
		{ 7305, 0x9214 }, { 7306, 0x922C }, { 7307, 0x9215 }, { 7308, 0x9211 }, { 7309, 0x925E }, { 7310, 0x9257 },
		{ 7311, 0x9245 }, { 7312, 0x9249 }, { 7313, 0x9264 }, { 7314, 0x9248 }, { 7315, 0x9295 }, { 7316, 0x923F },
		{ 7317, 0x924B }, { 7318, 0x9250 }, { 7319, 0x929C }, { 7320, 0x9296 }, { 7321, 0x9293 }, { 7322, 0x929B },
		{ 7323, 0x925A }, { 7324, 0x92CF }, { 7325, 0x92B9 }, { 7326, 0x92B7 }, { 7327, 0x92E9 }, { 7328, 0x930F },
		{ 7329, 0x92FA }, { 7330, 0x9344 }, { 7331, 0x932E }, { 7332, 0x9319 }, { 7333, 0x9322 }, { 7334, 0x931A },
		{ 7335, 0x9323 }, { 7336, 0x933A }, { 7337, 0x9335 }, { 7338, 0x933B }, { 7339, 0x935C }, { 7340, 0x9360 },
		{ 7341, 0x937C }, { 7342, 0x936E }, { 7343, 0x9356 }, { 7344, 0x93B0 }, { 7345, 0x93AC }, { 7346, 0x93AD },
		{ 7347, 0x9394 }, { 7348, 0x93B9 }, { 7349, 0x93D6 }, { 7350, 0x93D7 }, { 7351, 0x93E8 }, { 7352, 0x93E5 },
		{ 7353, 0x93D8 }, { 7354, 0x93C3 }, { 7355, 0x93DD }, { 7356, 0x93D0 }, { 7357, 0x93C8 }, { 7358, 0x93E4 },
		{ 7359, 0x941A }, { 7360, 0x9414 }, { 7361, 0x9413 }, { 7362, 0x9403 }, { 7363, 0x9407 }, { 7364, 0x9410 },
		{ 7365, 0x9436 }, { 7366, 0x942B }, { 7367, 0x9435 }, { 7368, 0x9421 }, { 7369, 0x943A }, { 7370, 0x9441 },
		{ 7371, 0x9452 }, { 7372, 0x9444 }, { 7373, 0x945B }, { 7374, 0x9460 }, { 7375, 0x9462 }, { 7376, 0x945E },
		{ 7377, 0x946A }, { 7378, 0x9229 }, { 7379, 0x9470 }, { 7380, 0x9475 }, { 7381, 0x9477 }, { 7382, 0x947D },
		{ 7383, 0x945A }, { 7384, 0x947C }, { 7385, 0x947E }, { 7386, 0x9481 }, { 7387, 0x947F }, { 7388, 0x9582 },
		{ 7389, 0x9587 }, { 7390, 0x958A }, { 7391, 0x9594 }, { 7392, 0x9596 }, { 7393, 0x9598 }, { 7394, 0x9599 },
		{ 7395, 0x95A0 }, { 7396, 0x95A8 }, { 7397, 0x95A7 }, { 7398, 0x95AD }, { 7399, 0x95BC }, { 7400, 0x95BB },
		{ 7401, 0x95B9 }, { 7402, 0x95BE }, { 7403, 0x95CA }, { 7404, 0x6FF6 }, { 7405, 0x95C3 }, { 7406, 0x95CD },
		{ 7407, 0x95CC }, { 7408, 0x95D5 }, { 7409, 0x95D4 }, { 7410, 0x95D6 }, { 7411, 0x95DC }, { 7412, 0x95E1 },
		{ 7413, 0x95E5 }, { 7414, 0x95E2 }, { 7415, 0x9621 }, { 7416, 0x9628 }, { 7417, 0x962E }, { 7418, 0x962F },
		{ 7419, 0x9642 }, { 7420, 0x964C }, { 7421, 0x964F }, { 7422, 0x964B }, { 7423, 0x9677 }, { 7424, 0x965C },
		{ 7425, 0x965E }, { 7426, 0x965D }, { 7427, 0x965F }, { 7428, 0x9666 }, { 7429, 0x9672 }, { 7430, 0x966C },
		{ 7431, 0x968D }, { 7432, 0x9698 }, { 7433, 0x9695 }, { 7434, 0x9697 }, { 7435, 0x96AA }, { 7436, 0x96A7 },
		{ 7437, 0x96B1 }, { 7438, 0x96B2 }, { 7439, 0x96B0 }, { 7440, 0x96B4 }, { 7441, 0x96B6 }, { 7442, 0x96B8 },
		{ 7443, 0x96B9 }, { 7444, 0x96CE }, { 7445, 0x96CB }, { 7446, 0x96C9 }, { 7447, 0x96CD }, { 7448, 0x894D },
		{ 7449, 0x96DC }, { 7450, 0x970D }, { 7451, 0x96D5 }, { 7452, 0x96F9 }, { 7453, 0x9704 }, { 7454, 0x9706 },
		{ 7455, 0x9708 }, { 7456, 0x9713 }, { 7457, 0x970E }, { 7458, 0x9711 }, { 7459, 0x970F }, { 7460, 0x9716 },
		{ 7461, 0x9719 }, { 7462, 0x9724 }, { 7463, 0x972A }, { 7464, 0x9730 }, { 7465, 0x9739 }, { 7466, 0x973D },
		{ 7467, 0x973E }, { 7468, 0x9744 }, { 7469, 0x9746 }, { 7470, 0x9748 }, { 7471, 0x9742 }, { 7472, 0x9749 },
		{ 7473, 0x975C }, { 7474, 0x9760 }, { 7475, 0x9764 }, { 7476, 0x9766 }, { 7477, 0x9768 }, { 7478, 0x52D2 },
		{ 7479, 0x976B }, { 7480, 0x9771 }, { 7481, 0x9779 }, { 7482, 0x9785 }, { 7483, 0x977C }, { 7484, 0x9781 },
		{ 7485, 0x977A }, { 7486, 0x9786 }, { 7487, 0x978B }, { 7488, 0x978F }, { 7489, 0x9790 }, { 7490, 0x979C },
		{ 7491, 0x97A8 }, { 7492, 0x97A6 }, { 7493, 0x97A3 }, { 7494, 0x97B3 }, { 7495, 0x97B4 }, { 7496, 0x97C3 },
		{ 7497, 0x97C6 }, { 7498, 0x97C8 }, { 7499, 0x97CB }, { 7500, 0x97DC }, { 7501, 0x97ED }, { 7502, 0x9F4F },
		{ 7503, 0x97F2 }, { 7504, 0x7ADF }, { 7505, 0x97F6 }, { 7506, 0x97F5 }, { 7507, 0x980F }, { 7508, 0x980C },
		{ 7509, 0x9838 }, { 7510, 0x9824 }, { 7511, 0x9821 }, { 7512, 0x9837 }, { 7513, 0x983D }, { 7514, 0x9846 },
		{ 7515, 0x984F }, { 7516, 0x984B }, { 7517, 0x986B }, { 7518, 0x986F }, { 7519, 0x9870 }, { 7520, 0x9871 },
		{ 7521, 0x9874 }, { 7522, 0x9873 }, { 7523, 0x98AA }, { 7524, 0x98AF }, { 7525, 0x98B1 }, { 7526, 0x98B6 },
		{ 7527, 0x98C4 }, { 7528, 0x98C3 }, { 7529, 0x98C6 }, { 7530, 0x98E9 }, { 7531, 0x98EB }, { 7532, 0x9903 },
		{ 7533, 0x9909 }, { 7534, 0x9912 }, { 7535, 0x9914 }, { 7536, 0x9918 }, { 7537, 0x9921 }, { 7538, 0x991D },
		{ 7539, 0x991E }, { 7540, 0x9924 }, { 7541, 0x9920 }, { 7542, 0x992C }, { 7543, 0x992E }, { 7544, 0x993D },
		{ 7545, 0x993E }, { 7546, 0x9942 }, { 7547, 0x9949 }, { 7548, 0x9945 }, { 7549, 0x9950 }, { 7550, 0x994B },
		{ 7551, 0x9951 }, { 7552, 0x9952 }, { 7553, 0x994C }, { 7554, 0x9955 }, { 7555, 0x9997 }, { 7556, 0x9998 },
		{ 7557, 0x99A5 }, { 7558, 0x99AD }, { 7559, 0x99AE }, { 7560, 0x99BC }, { 7561, 0x99DF }, { 7562, 0x99DB },
		{ 7563, 0x99DD }, { 7564, 0x99D8 }, { 7565, 0x99D1 }, { 7566, 0x99ED }, { 7567, 0x99EE }, { 7568, 0x99F1 },
		{ 7569, 0x99F2 }, { 7570, 0x99FB }, { 7571, 0x99F8 }, { 7572, 0x9A01 }, { 7573, 0x9A0F }, { 7574, 0x9A05 },
		{ 7575, 0x99E2 }, { 7576, 0x9A19 }, { 7577, 0x9A2B }, { 7578, 0x9A37 }, { 7579, 0x9A45 }, { 7580, 0x9A42 },
		{ 7581, 0x9A40 }, { 7582, 0x9A43 }, { 7583, 0x9A3E }, { 7584, 0x9A55 }, { 7585, 0x9A4D }, { 7586, 0x9A5B },
		{ 7587, 0x9A57 }, { 7588, 0x9A5F }, { 7589, 0x9A62 }, { 7590, 0x9A65 }, { 7591, 0x9A64 }, { 7592, 0x9A69 },
		{ 7593, 0x9A6B }, { 7594, 0x9A6A }, { 7595, 0x9AAD }, { 7596, 0x9AB0 }, { 7597, 0x9ABC }, { 7598, 0x9AC0 },
		{ 7599, 0x9ACF }, { 7600, 0x9AD1 }, { 7601, 0x9AD3 }, { 7602, 0x9AD4 }, { 7603, 0x9ADE }, { 7604, 0x9ADF },
		{ 7605, 0x9AE2 }, { 7606, 0x9AE3 }, { 7607, 0x9AE6 }, { 7608, 0x9AEF }, { 7609, 0x9AEB }, { 7610, 0x9AEE },
		{ 7611, 0x9AF4 }, { 7612, 0x9AF1 }, { 7613, 0x9AF7 }, { 7614, 0x9AFB }, { 7615, 0x9B06 }, { 7616, 0x9B18 },
		{ 7617, 0x9B1A }, { 7618, 0x9B1F }, { 7619, 0x9B22 }, { 7620, 0x9B23 }, { 7621, 0x9B25 }, { 7622, 0x9B27 },
		{ 7623, 0x9B28 }, { 7624, 0x9B29 }, { 7625, 0x9B2A }, { 7626, 0x9B2E }, { 7627, 0x9B2F }, { 7628, 0x9B32 },
		{ 7629, 0x9B44 }, { 7630, 0x9B43 }, { 7631, 0x9B4F }, { 7632, 0x9B4D }, { 7633, 0x9B4E }, { 7634, 0x9B51 },
		{ 7635, 0x9B58 }, { 7636, 0x9B74 }, { 7637, 0x9B93 }, { 7638, 0x9B83 }, { 7639, 0x9B91 }, { 7640, 0x9B96 },
		{ 7641, 0x9B97 }, { 7642, 0x9B9F }, { 7643, 0x9BA0 }, { 7644, 0x9BA8 }, { 7645, 0x9BB4 }, { 7646, 0x9BC0 },
		{ 7647, 0x9BCA }, { 7648, 0x9BB9 }, { 7649, 0x9BC6 }, { 7650, 0x9BCF }, { 7651, 0x9BD1 }, { 7652, 0x9BD2 },
		{ 7653, 0x9BE3 }, { 7654, 0x9BE2 }, { 7655, 0x9BE4 }, { 7656, 0x9BD4 }, { 7657, 0x9BE1 }, { 7658, 0x9C3A },
		{ 7659, 0x9BF2 }, { 7660, 0x9BF1 }, { 7661, 0x9BF0 }, { 7662, 0x9C15 }, { 7663, 0x9C14 }, { 7664, 0x9C09 },
		{ 7665, 0x9C13 }, { 7666, 0x9C0C }, { 7667, 0x9C06 }, { 7668, 0x9C08 }, { 7669, 0x9C12 }, { 7670, 0x9C0A },
		{ 7671, 0x9C04 }, { 7672, 0x9C2E }, { 7673, 0x9C1B }, { 7674, 0x9C25 }, { 7675, 0x9C24 }, { 7676, 0x9C21 },
		{ 7677, 0x9C30 }, { 7678, 0x9C47 }, { 7679, 0x9C32 }, { 7680, 0x9C46 }, { 7681, 0x9C3E }, { 7682, 0x9C5A },
		{ 7683, 0x9C60 }, { 7684, 0x9C67 }, { 7685, 0x9C76 }, { 7686, 0x9C78 }, { 7687, 0x9CE7 }, { 7688, 0x9CEC },
		{ 7689, 0x9CF0 }, { 7690, 0x9D09 }, { 7691, 0x9D08 }, { 7692, 0x9CEB }, { 7693, 0x9D03 }, { 7694, 0x9D06 },
		{ 7695, 0x9D2A }, { 7696, 0x9D26 }, { 7697, 0x9DAF }, { 7698, 0x9D23 }, { 7699, 0x9D1F }, { 7700, 0x9D44 },
		{ 7701, 0x9D15 }, { 7702, 0x9D12 }, { 7703, 0x9D41 }, { 7704, 0x9D3F }, { 7705, 0x9D3E }, { 7706, 0x9D46 },
		{ 7707, 0x9D48 }, { 7708, 0x9D5D }, { 7709, 0x9D5E }, { 7710, 0x9D64 }, { 7711, 0x9D51 }, { 7712, 0x9D50 },
		{ 7713, 0x9D59 }, { 7714, 0x9D72 }, { 7715, 0x9D89 }, { 7716, 0x9D87 }, { 7717, 0x9DAB }, { 7718, 0x9D6F },
		{ 7719, 0x9D7A }, { 7720, 0x9D9A }, { 7721, 0x9DA4 }, { 7722, 0x9DA9 }, { 7723, 0x9DB2 }, { 7724, 0x9DC4 },
		{ 7725, 0x9DC1 }, { 7726, 0x9DBB }, { 7727, 0x9DB8 }, { 7728, 0x9DBA }, { 7729, 0x9DC6 }, { 7730, 0x9DCF },
		{ 7731, 0x9DC2 }, { 7732, 0x9DD9 }, { 7733, 0x9DD3 }, { 7734, 0x9DF8 }, { 7735, 0x9DE6 }, { 7736, 0x9DED },
		{ 7737, 0x9DEF }, { 7738, 0x9DFD }, { 7739, 0x9E1A }, { 7740, 0x9E1B }, { 7741, 0x9E1E }, { 7742, 0x9E75 },
		{ 7743, 0x9E79 }, { 7744, 0x9E7D }, { 7745, 0x9E81 }, { 7746, 0x9E88 }, { 7747, 0x9E8B }, { 7748, 0x9E8C },
		{ 7749, 0x9E92 }, { 7750, 0x9E95 }, { 7751, 0x9E91 }, { 7752, 0x9E9D }, { 7753, 0x9EA5 }, { 7754, 0x9EA9 },
		{ 7755, 0x9EB8 }, { 7756, 0x9EAA }, { 7757, 0x9EAD }, { 7758, 0x9761 }, { 7759, 0x9ECC }, { 7760, 0x9ECE },
		{ 7761, 0x9ECF }, { 7762, 0x9ED0 }, { 7763, 0x9ED4 }, { 7764, 0x9EDC }, { 7765, 0x9EDE }, { 7766, 0x9EDD },
		{ 7767, 0x9EE0 }, { 7768, 0x9EE5 }, { 7769, 0x9EE8 }, { 7770, 0x9EEF }, { 7771, 0x9EF4 }, { 7772, 0x9EF6 },
		{ 7773, 0x9EF7 }, { 7774, 0x9EF9 }, { 7775, 0x9EFB }, { 7776, 0x9EFC }, { 7777, 0x9EFD }, { 7778, 0x9F07 },
		{ 7779, 0x9F08 }, { 7780, 0x76B7 }, { 7781, 0x9F15 }, { 7782, 0x9F21 }, { 7783, 0x9F2C }, { 7784, 0x9F3E },
		{ 7785, 0x9F4A }, { 7786, 0x9F52 }, { 7787, 0x9F54 }, { 7788, 0x9F63 }, { 7789, 0x9F5F }, { 7790, 0x9F60 },
		{ 7791, 0x9F61 }, { 7792, 0x9F66 }, { 7793, 0x9F67 }, { 7794, 0x9F6C }, { 7795, 0x9F6A }, { 7796, 0x9F77 },
		{ 7797, 0x9F72 }, { 7798, 0x9F76 }, { 7799, 0x9F95 }, { 7800, 0x9F9C }, { 7801, 0x9FA0 }, { 7802, 0x582F },
		{ 7803, 0x69C7 }, { 7804, 0x9059 }, { 7805, 0x7464 }, { 7806, 0x51DC }, { 7807, 0x7199 }, { 8272, 0x7E8A },
		{ 8273, 0x891C }, { 8274, 0x9348 }, { 8275, 0x9288 }, { 8276, 0x84DC }, { 8277, 0x4FC9 }, { 8278, 0x70BB },
		{ 8279, 0x6631 }, { 8280, 0x68C8 }, { 8281, 0x92F9 }, { 8282, 0x66FB }, { 8283, 0x5F45 }, { 8284, 0x4E28 },
		{ 8285, 0x4EE1 }, { 8286, 0x4EFC }, { 8287, 0x4F00 }, { 8288, 0x4F03 }, { 8289, 0x4F39 }, { 8290, 0x4F56 },
		{ 8291, 0x4F92 }, { 8292, 0x4F8A }, { 8293, 0x4F9A }, { 8294, 0x4F94 }, { 8295, 0x4FCD }, { 8296, 0x5040 },
		{ 8297, 0x5022 }, { 8298, 0x4FFF }, { 8299, 0x501E }, { 8300, 0x5046 }, { 8301, 0x5070 }, { 8302, 0x5042 },
		{ 8303, 0x5094 }, { 8304, 0x50F4 }, { 8305, 0x50D8 }, { 8306, 0x514A }, { 8307, 0x5164 }, { 8308, 0x519D },
		{ 8309, 0x51BE }, { 8310, 0x51EC }, { 8311, 0x5215 }, { 8312, 0x529C }, { 8313, 0x52A6 }, { 8314, 0x52C0 },
		{ 8315, 0x52DB }, { 8316, 0x5300 }, { 8317, 0x5307 }, { 8318, 0x5324 }, { 8319, 0x5372 }, { 8320, 0x5393 },
		{ 8321, 0x53B2 }, { 8322, 0x53DD }, { 8323, 0xFA0E }, { 8324, 0x549C }, { 8325, 0x548A }, { 8326, 0x54A9 },
		{ 8327, 0x54FF }, { 8328, 0x5586 }, { 8329, 0x5759 }, { 8330, 0x5765 }, { 8331, 0x57AC }, { 8332, 0x57C8 },
		{ 8333, 0x57C7 }, { 8334, 0xFA0F }, { 8335, 0xFA10 }, { 8336, 0x589E }, { 8337, 0x58B2 }, { 8338, 0x590B },
		{ 8339, 0x5953 }, { 8340, 0x595B }, { 8341, 0x595D }, { 8342, 0x5963 }, { 8343, 0x59A4 }, { 8344, 0x59BA },
		{ 8345, 0x5B56 }, { 8346, 0x5BC0 }, { 8347, 0x752F }, { 8348, 0x5BD8 }, { 8349, 0x5BEC }, { 8350, 0x5C1E },
		{ 8351, 0x5CA6 }, { 8352, 0x5CBA }, { 8353, 0x5CF5 }, { 8354, 0x5D27 }, { 8355, 0x5D53 }, { 8356, 0xFA11 },
		{ 8357, 0x5D42 }, { 8358, 0x5D6D }, { 8359, 0x5DB8 }, { 8360, 0x5DB9 }, { 8361, 0x5DD0 }, { 8362, 0x5F21 },
		{ 8363, 0x5F34 }, { 8364, 0x5F67 }, { 8365, 0x5FB7 }, { 8366, 0x5FDE }, { 8367, 0x605D }, { 8368, 0x6085 },
		{ 8369, 0x608A }, { 8370, 0x60DE }, { 8371, 0x60D5 }, { 8372, 0x6120 }, { 8373, 0x60F2 }, { 8374, 0x6111 },
		{ 8375, 0x6137 }, { 8376, 0x6130 }, { 8377, 0x6198 }, { 8378, 0x6213 }, { 8379, 0x62A6 }, { 8380, 0x63F5 },
		{ 8381, 0x6460 }, { 8382, 0x649D }, { 8383, 0x64CE }, { 8384, 0x654E }, { 8385, 0x6600 }, { 8386, 0x6615 },
		{ 8387, 0x663B }, { 8388, 0x6609 }, { 8389, 0x662E }, { 8390, 0x661E }, { 8391, 0x6624 }, { 8392, 0x6665 },
		{ 8393, 0x6657 }, { 8394, 0x6659 }, { 8395, 0xFA12 }, { 8396, 0x6673 }, { 8397, 0x6699 }, { 8398, 0x66A0 },
		{ 8399, 0x66B2 }, { 8400, 0x66BF }, { 8401, 0x66FA }, { 8402, 0x670E }, { 8403, 0xF929 }, { 8404, 0x6766 },
		{ 8405, 0x67BB }, { 8406, 0x6852 }, { 8407, 0x67C0 }, { 8408, 0x6801 }, { 8409, 0x6844 }, { 8410, 0x68CF },
		{ 8411, 0xFA13 }, { 8412, 0x6968 }, { 8413, 0xFA14 }, { 8414, 0x6998 }, { 8415, 0x69E2 }, { 8416, 0x6A30 },
		{ 8417, 0x6A6B }, { 8418, 0x6A46 }, { 8419, 0x6A73 }, { 8420, 0x6A7E }, { 8421, 0x6AE2 }, { 8422, 0x6AE4 },
		{ 8423, 0x6BD6 }, { 8424, 0x6C3F }, { 8425, 0x6C5C }, { 8426, 0x6C86 }, { 8427, 0x6C6F }, { 8428, 0x6CDA },
		{ 8429, 0x6D04 }, { 8430, 0x6D87 }, { 8431, 0x6D6F }, { 8432, 0x6D96 }, { 8433, 0x6DAC }, { 8434, 0x6DCF },
		{ 8435, 0x6DF8 }, { 8436, 0x6DF2 }, { 8437, 0x6DFC }, { 8438, 0x6E39 }, { 8439, 0x6E5C }, { 8440, 0x6E27 },
		{ 8441, 0x6E3C }, { 8442, 0x6EBF }, { 8443, 0x6F88 }, { 8444, 0x6FB5 }, { 8445, 0x6FF5 }, { 8446, 0x7005 },
		{ 8447, 0x7007 }, { 8448, 0x7028 }, { 8449, 0x7085 }, { 8450, 0x70AB }, { 8451, 0x710F }, { 8452, 0x7104 },
		{ 8453, 0x715C }, { 8454, 0x7146 }, { 8455, 0x7147 }, { 8456, 0xFA15 }, { 8457, 0x71C1 }, { 8458, 0x71FE },
		{ 8459, 0x72B1 }, { 8460, 0x72BE }, { 8461, 0x7324 }, { 8462, 0xFA16 }, { 8463, 0x7377 }, { 8464, 0x73BD },
		{ 8465, 0x73C9 }, { 8466, 0x73D6 }, { 8467, 0x73E3 }, { 8468, 0x73D2 }, { 8469, 0x7407 }, { 8470, 0x73F5 },
		{ 8471, 0x7426 }, { 8472, 0x742A }, { 8473, 0x7429 }, { 8474, 0x742E }, { 8475, 0x7462 }, { 8476, 0x7489 },
		{ 8477, 0x749F }, { 8478, 0x7501 }, { 8479, 0x756F }, { 8480, 0x7682 }, { 8481, 0x769C }, { 8482, 0x769E },
		{ 8483, 0x769B }, { 8484, 0x76A6 }, { 8485, 0xFA17 }, { 8486, 0x7746 }, { 8487, 0x52AF }, { 8488, 0x7821 },
		{ 8489, 0x784E }, { 8490, 0x7864 }, { 8491, 0x787A }, { 8492, 0x7930 }, { 8493, 0xFA18 }, { 8494, 0xFA19 },
		{ 8495, 0xFA1A }, { 8496, 0x7994 }, { 8497, 0xFA1B }, { 8498, 0x799B }, { 8499, 0x7AD1 }, { 8500, 0x7AE7 },
		{ 8501, 0xFA1C }, { 8502, 0x7AEB }, { 8503, 0x7B9E }, { 8504, 0xFA1D }, { 8505, 0x7D48 }, { 8506, 0x7D5C },
		{ 8507, 0x7DB7 }, { 8508, 0x7DA0 }, { 8509, 0x7DD6 }, { 8510, 0x7E52 }, { 8511, 0x7F47 }, { 8512, 0x7FA1 },
		{ 8513, 0xFA1E }, { 8514, 0x8301 }, { 8515, 0x8362 }, { 8516, 0x837F }, { 8517, 0x83C7 }, { 8518, 0x83F6 },
		{ 8519, 0x8448 }, { 8520, 0x84B4 }, { 8521, 0x8553 }, { 8522, 0x8559 }, { 8523, 0x856B }, { 8524, 0xFA1F },
		{ 8525, 0x85B0 }, { 8526, 0xFA20 }, { 8527, 0xFA21 }, { 8528, 0x8807 }, { 8529, 0x88F5 }, { 8530, 0x8A12 },
		{ 8531, 0x8A37 }, { 8532, 0x8A79 }, { 8533, 0x8AA7 }, { 8534, 0x8ABE }, { 8535, 0x8ADF }, { 8536, 0xFA22 },
		{ 8537, 0x8AF6 }, { 8538, 0x8B53 }, { 8539, 0x8B7F }, { 8540, 0x8CF0 }, { 8541, 0x8CF4 }, { 8542, 0x8D12 },
		{ 8543, 0x8D76 }, { 8544, 0xFA23 }, { 8545, 0x8ECF }, { 8546, 0xFA24 }, { 8547, 0xFA25 }, { 8548, 0x9067 },
		{ 8549, 0x90DE }, { 8550, 0xFA26 }, { 8551, 0x9115 }, { 8552, 0x9127 }, { 8553, 0x91DA }, { 8554, 0x91D7 },
		{ 8555, 0x91DE }, { 8556, 0x91ED }, { 8557, 0x91EE }, { 8558, 0x91E4 }, { 8559, 0x91E5 }, { 8560, 0x9206 },
		{ 8561, 0x9210 }, { 8562, 0x920A }, { 8563, 0x923A }, { 8564, 0x9240 }, { 8565, 0x923C }, { 8566, 0x924E },
		{ 8567, 0x9259 }, { 8568, 0x9251 }, { 8569, 0x9239 }, { 8570, 0x9267 }, { 8571, 0x92A7 }, { 8572, 0x9277 },
		{ 8573, 0x9278 }, { 8574, 0x92E7 }, { 8575, 0x92D7 }, { 8576, 0x92D9 }, { 8577, 0x92D0 }, { 8578, 0xFA27 },
		{ 8579, 0x92D5 }, { 8580, 0x92E0 }, { 8581, 0x92D3 }, { 8582, 0x9325 }, { 8583, 0x9321 }, { 8584, 0x92FB },
		{ 8585, 0xFA28 }, { 8586, 0x931E }, { 8587, 0x92FF }, { 8588, 0x931D }, { 8589, 0x9302 }, { 8590, 0x9370 },
		{ 8591, 0x9357 }, { 8592, 0x93A4 }, { 8593, 0x93C6 }, { 8594, 0x93DE }, { 8595, 0x93F8 }, { 8596, 0x9431 },
		{ 8597, 0x9445 }, { 8598, 0x9448 }, { 8599, 0x9592 }, { 8600, 0xF9DC }, { 8601, 0xFA29 }, { 8602, 0x969D },
		{ 8603, 0x96AF }, { 8604, 0x9733 }, { 8605, 0x973B }, { 8606, 0x9743 }, { 8607, 0x974D }, { 8608, 0x974F },
		{ 8609, 0x9751 }, { 8610, 0x9755 }, { 8611, 0x9857 }, { 8612, 0x9865 }, { 8613, 0xFA2A }, { 8614, 0xFA2B },
		{ 8615, 0x9927 }, { 8616, 0xFA2C }, { 8617, 0x999E }, { 8618, 0x9A4E }, { 8619, 0x9AD9 }, { 8620, 0x9ADC },
		{ 8621, 0x9B75 }, { 8622, 0x9B72 }, { 8623, 0x9B8F }, { 8624, 0x9BB1 }, { 8625, 0x9BBB }, { 8626, 0x9C00 },
		{ 8627, 0x9D70 }, { 8628, 0x9D6B }, { 8629, 0xFA2D }, { 8630, 0x9E19 }, { 8631, 0x9ED1 }, { 8634, 0x2170 },
		{ 8635, 0x2171 }, { 8636, 0x2172 }, { 8637, 0x2173 }, { 8638, 0x2174 }, { 8639, 0x2175 }, { 8640, 0x2176 },
		{ 8641, 0x2177 }, { 8642, 0x2178 }, { 8643, 0x2179 }, { 8644, 0xFFE2 }, { 8645, 0xFFE4 }, { 8646, 0xFF07 },
		{ 8647, 0xFF02 }, { 10716, 0x2170 }, { 10717, 0x2171 }, { 10718, 0x2172 }, { 10719, 0x2173 },
		{ 10720, 0x2174 }, { 10721, 0x2175 }, { 10722, 0x2176 }, { 10723, 0x2177 }, { 10724, 0x2178 },
		{ 10725, 0x2179 }, { 10726, 0x2160 }, { 10727, 0x2161 }, { 10728, 0x2162 }, { 10729, 0x2163 },
		{ 10730, 0x2164 }, { 10731, 0x2165 }, { 10732, 0x2166 }, { 10733, 0x2167 }, { 10734, 0x2168 },
		{ 10735, 0x2169 }, { 10736, 0xFFE2 }, { 10737, 0xFFE4 }, { 10738, 0xFF07 }, { 10739, 0xFF02 },
		{ 10740, 0x3231 }, { 10741, 0x2116 }, { 10742, 0x2121 }, { 10743, 0x2235 }, { 10744, 0x7E8A },
		{ 10745, 0x891C }, { 10746, 0x9348 }, { 10747, 0x9288 }, { 10748, 0x84DC }, { 10749, 0x4FC9 },
		{ 10750, 0x70BB }, { 10751, 0x6631 }, { 10752, 0x68C8 }, { 10753, 0x92F9 }, { 10754, 0x66FB },
		{ 10755, 0x5F45 }, { 10756, 0x4E28 }, { 10757, 0x4EE1 }, { 10758, 0x4EFC }, { 10759, 0x4F00 },
		{ 10760, 0x4F03 }, { 10761, 0x4F39 }, { 10762, 0x4F56 }, { 10763, 0x4F92 }, { 10764, 0x4F8A },
		{ 10765, 0x4F9A }, { 10766, 0x4F94 }, { 10767, 0x4FCD }, { 10768, 0x5040 }, { 10769, 0x5022 },
		{ 10770, 0x4FFF }, { 10771, 0x501E }, { 10772, 0x5046 }, { 10773, 0x5070 }, { 10774, 0x5042 },
		{ 10775, 0x5094 }, { 10776, 0x50F4 }, { 10777, 0x50D8 }, { 10778, 0x514A }, { 10779, 0x5164 },
		{ 10780, 0x519D }, { 10781, 0x51BE }, { 10782, 0x51EC }, { 10783, 0x5215 }, { 10784, 0x529C },
		{ 10785, 0x52A6 }, { 10786, 0x52C0 }, { 10787, 0x52DB }, { 10788, 0x5300 }, { 10789, 0x5307 },
		{ 10790, 0x5324 }, { 10791, 0x5372 }, { 10792, 0x5393 }, { 10793, 0x53B2 }, { 10794, 0x53DD },
		{ 10795, 0xFA0E }, { 10796, 0x549C }, { 10797, 0x548A }, { 10798, 0x54A9 }, { 10799, 0x54FF },
		{ 10800, 0x5586 }, { 10801, 0x5759 }, { 10802, 0x5765 }, { 10803, 0x57AC }, { 10804, 0x57C8 },
		{ 10805, 0x57C7 }, { 10806, 0xFA0F }, { 10807, 0xFA10 }, { 10808, 0x589E }, { 10809, 0x58B2 },
		{ 10810, 0x590B }, { 10811, 0x5953 }, { 10812, 0x595B }, { 10813, 0x595D }, { 10814, 0x5963 },
		{ 10815, 0x59A4 }, { 10816, 0x59BA }, { 10817, 0x5B56 }, { 10818, 0x5BC0 }, { 10819, 0x752F },
		{ 10820, 0x5BD8 }, { 10821, 0x5BEC }, { 10822, 0x5C1E }, { 10823, 0x5CA6 }, { 10824, 0x5CBA },
		{ 10825, 0x5CF5 }, { 10826, 0x5D27 }, { 10827, 0x5D53 }, { 10828, 0xFA11 }, { 10829, 0x5D42 },
		{ 10830, 0x5D6D }, { 10831, 0x5DB8 }, { 10832, 0x5DB9 }, { 10833, 0x5DD0 }, { 10834, 0x5F21 },
		{ 10835, 0x5F34 }, { 10836, 0x5F67 }, { 10837, 0x5FB7 }, { 10838, 0x5FDE }, { 10839, 0x605D },
		{ 10840, 0x6085 }, { 10841, 0x608A }, { 10842, 0x60DE }, { 10843, 0x60D5 }, { 10844, 0x6120 },
		{ 10845, 0x60F2 }, { 10846, 0x6111 }, { 10847, 0x6137 }, { 10848, 0x6130 }, { 10849, 0x6198 },
		{ 10850, 0x6213 }, { 10851, 0x62A6 }, { 10852, 0x63F5 }, { 10853, 0x6460 }, { 10854, 0x649D },
		{ 10855, 0x64CE }, { 10856, 0x654E }, { 10857, 0x6600 }, { 10858, 0x6615 }, { 10859, 0x663B },
		{ 10860, 0x6609 }, { 10861, 0x662E }, { 10862, 0x661E }, { 10863, 0x6624 }, { 10864, 0x6665 },
		{ 10865, 0x6657 }, { 10866, 0x6659 }, { 10867, 0xFA12 }, { 10868, 0x6673 }, { 10869, 0x6699 },
		{ 10870, 0x66A0 }, { 10871, 0x66B2 }, { 10872, 0x66BF }, { 10873, 0x66FA }, { 10874, 0x670E },
		{ 10875, 0xF929 }, { 10876, 0x6766 }, { 10877, 0x67BB }, { 10878, 0x6852 }, { 10879, 0x67C0 },
		{ 10880, 0x6801 }, { 10881, 0x6844 }, { 10882, 0x68CF }, { 10883, 0xFA13 }, { 10884, 0x6968 },
		{ 10885, 0xFA14 }, { 10886, 0x6998 }, { 10887, 0x69E2 }, { 10888, 0x6A30 }, { 10889, 0x6A6B },
		{ 10890, 0x6A46 }, { 10891, 0x6A73 }, { 10892, 0x6A7E }, { 10893, 0x6AE2 }, { 10894, 0x6AE4 },
		{ 10895, 0x6BD6 }, { 10896, 0x6C3F }, { 10897, 0x6C5C }, { 10898, 0x6C86 }, { 10899, 0x6C6F },
		{ 10900, 0x6CDA }, { 10901, 0x6D04 }, { 10902, 0x6D87 }, { 10903, 0x6D6F }, { 10904, 0x6D96 },
		{ 10905, 0x6DAC }, { 10906, 0x6DCF }, { 10907, 0x6DF8 }, { 10908, 0x6DF2 }, { 10909, 0x6DFC },
		{ 10910, 0x6E39 }, { 10911, 0x6E5C }, { 10912, 0x6E27 }, { 10913, 0x6E3C }, { 10914, 0x6EBF },
		{ 10915, 0x6F88 }, { 10916, 0x6FB5 }, { 10917, 0x6FF5 }, { 10918, 0x7005 }, { 10919, 0x7007 },
		{ 10920, 0x7028 }, { 10921, 0x7085 }, { 10922, 0x70AB }, { 10923, 0x710F }, { 10924, 0x7104 },
		{ 10925, 0x715C }, { 10926, 0x7146 }, { 10927, 0x7147 }, { 10928, 0xFA15 }, { 10929, 0x71C1 },
		{ 10930, 0x71FE }, { 10931, 0x72B1 }, { 10932, 0x72BE }, { 10933, 0x7324 }, { 10934, 0xFA16 },
		{ 10935, 0x7377 }, { 10936, 0x73BD }, { 10937, 0x73C9 }, { 10938, 0x73D6 }, { 10939, 0x73E3 },
		{ 10940, 0x73D2 }, { 10941, 0x7407 }, { 10942, 0x73F5 }, { 10943, 0x7426 }, { 10944, 0x742A },
		{ 10945, 0x7429 }, { 10946, 0x742E }, { 10947, 0x7462 }, { 10948, 0x7489 }, { 10949, 0x749F },
		{ 10950, 0x7501 }, { 10951, 0x756F }, { 10952, 0x7682 }, { 10953, 0x769C }, { 10954, 0x769E },
		{ 10955, 0x769B }, { 10956, 0x76A6 }, { 10957, 0xFA17 }, { 10958, 0x7746 }, { 10959, 0x52AF },
		{ 10960, 0x7821 }, { 10961, 0x784E }, { 10962, 0x7864 }, { 10963, 0x787A }, { 10964, 0x7930 },
		{ 10965, 0xFA18 }, { 10966, 0xFA19 }, { 10967, 0xFA1A }, { 10968, 0x7994 }, { 10969, 0xFA1B },
		{ 10970, 0x799B }, { 10971, 0x7AD1 }, { 10972, 0x7AE7 }, { 10973, 0xFA1C }, { 10974, 0x7AEB },
		{ 10975, 0x7B9E }, { 10976, 0xFA1D }, { 10977, 0x7D48 }, { 10978, 0x7D5C }, { 10979, 0x7DB7 },
		{ 10980, 0x7DA0 }, { 10981, 0x7DD6 }, { 10982, 0x7E52 }, { 10983, 0x7F47 }, { 10984, 0x7FA1 },
		{ 10985, 0xFA1E }, { 10986, 0x8301 }, { 10987, 0x8362 }, { 10988, 0x837F }, { 10989, 0x83C7 },
		{ 10990, 0x83F6 }, { 10991, 0x8448 }, { 10992, 0x84B4 }, { 10993, 0x8553 }, { 10994, 0x8559 },
		{ 10995, 0x856B }, { 10996, 0xFA1F }, { 10997, 0x85B0 }, { 10998, 0xFA20 }, { 10999, 0xFA21 },
		{ 11000, 0x8807 }, { 11001, 0x88F5 }, { 11002, 0x8A12 }, { 11003, 0x8A37 }, { 11004, 0x8A79 },
		{ 11005, 0x8AA7 }, { 11006, 0x8ABE }, { 11007, 0x8ADF }, { 11008, 0xFA22 }, { 11009, 0x8AF6 },
		{ 11010, 0x8B53 }, { 11011, 0x8B7F }, { 11012, 0x8CF0 }, { 11013, 0x8CF4 }, { 11014, 0x8D12 },
		{ 11015, 0x8D76 }, { 11016, 0xFA23 }, { 11017, 0x8ECF }, { 11018, 0xFA24 }, { 11019, 0xFA25 },
		{ 11020, 0x9067 }, { 11021, 0x90DE }, { 11022, 0xFA26 }, { 11023, 0x9115 }, { 11024, 0x9127 },
		{ 11025, 0x91DA }, { 11026, 0x91D7 }, { 11027, 0x91DE }, { 11028, 0x91ED }, { 11029, 0x91EE },
		{ 11030, 0x91E4 }, { 11031, 0x91E5 }, { 11032, 0x9206 }, { 11033, 0x9210 }, { 11034, 0x920A },
		{ 11035, 0x923A }, { 11036, 0x9240 }, { 11037, 0x923C }, { 11038, 0x924E }, { 11039, 0x9259 },
		{ 11040, 0x9251 }, { 11041, 0x9239 }, { 11042, 0x9267 }, { 11043, 0x92A7 }, { 11044, 0x9277 },
		{ 11045, 0x9278 }, { 11046, 0x92E7 }, { 11047, 0x92D7 }, { 11048, 0x92D9 }, { 11049, 0x92D0 },
		{ 11050, 0xFA27 }, { 11051, 0x92D5 }, { 11052, 0x92E0 }, { 11053, 0x92D3 }, { 11054, 0x9325 },
		{ 11055, 0x9321 }, { 11056, 0x92FB }, { 11057, 0xFA28 }, { 11058, 0x931E }, { 11059, 0x92FF },
		{ 11060, 0x931D }, { 11061, 0x9302 }, { 11062, 0x9370 }, { 11063, 0x9357 }, { 11064, 0x93A4 },
		{ 11065, 0x93C6 }, { 11066, 0x93DE }, { 11067, 0x93F8 }, { 11068, 0x9431 }, { 11069, 0x9445 },
		{ 11070, 0x9448 }, { 11071, 0x9592 }, { 11072, 0xF9DC }, { 11073, 0xFA29 }, { 11074, 0x969D },
		{ 11075, 0x96AF }, { 11076, 0x9733 }, { 11077, 0x973B }, { 11078, 0x9743 }, { 11079, 0x974D },
		{ 11080, 0x974F }, { 11081, 0x9751 }, { 11082, 0x9755 }, { 11083, 0x9857 }, { 11084, 0x9865 },
		{ 11085, 0xFA2A }, { 11086, 0xFA2B }, { 11087, 0x9927 }, { 11088, 0xFA2C }, { 11089, 0x999E },
		{ 11090, 0x9A4E }, { 11091, 0x9AD9 }, { 11092, 0x9ADC }, { 11093, 0x9B75 }, { 11094, 0x9B72 },
		{ 11095, 0x9B8F }, { 11096, 0x9BB1 }, { 11097, 0x9BBB }, { 11098, 0x9C00 }, { 11099, 0x9D70 },
		{ 11100, 0x9D6B }, { 11101, 0xFA2D }, { 11102, 0x9E19 }, { 11103, 0x9ED1 } } };

	static bool less_than_index_target(
	     std::pair<std::uint_least32_t, std::uint_least32_t> value, std::uint_least32_t target) {
		return value.first < target;
	}

} // namespace

namespace detail {
	std::optional<char32_t> shift_jis_index_to_code_point(std::size_t lookup_index_pointer) noexcept {
		std::uint_least32_t lookup_index = static_cast<std::uint_least32_t>(lookup_index_pointer);
		auto it                          = std::lower_bound(
               index_codepoint_map.cbegin(), index_codepoint_map.cend(), lookup_index, &less_than_index_target);
		if (it == index_codepoint_map.cend()) {
			return std::nullopt;
		}
		const std::pair<std::uint_least32_t, std::uint_least32_t> index_and_codepoint = *it;
		if (index_and_codepoint.first != lookup_index) {
			return std::nullopt;
		}
		return static_cast<char32_t>(index_and_codepoint.second);
	}

	std::optional<std::size_t> shift_jis_code_point_to_index(char32_t code) noexcept {
		auto predicate = [&code](const std::pair<std::uint_least32_t, std::uint_least32_t>& value) {
			return code == value.second && !(value.first > 8272 && value.first < 8835);
		};
		auto it = std::find_if(index_codepoint_map.cbegin(), index_codepoint_map.cend(), predicate);
		if (it == index_codepoint_map.cend()) {
			return std::nullopt;
		}
		const std::pair<std::uint_least32_t, std::uint_least32_t> index_and_codepoint = *it;
		return static_cast<std::size_t>(index_and_codepoint.first);
	}
} // namespace detail
